#############################################################################
# blankcell.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# Functions related to Blank VOB Cell creation.  Now entierely written in Tcl.
# (was: Interface to jsoto's BlankCell.exe)
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::blankcell {

namespace export AppendBlankCell

## Return a blank VOB Cell (matching the video attributes).
## Arg 1: video attributes (from the V*_MAT table of the IFO)
## Arg 2-4: VOB ID, Cell ID and LBA.
proc blankCell {vattr vid cid lba} {
# read the VOB cell from disk, based on standard and resolution options
if {[expr {$vattr & 0x3000}] == 0} {
set standard "ntsc"
} else {
set standard "pal"
}
switch [expr {($vattr & 0x0018) / 8}] {
0 {set resolution 720}
1 {set resolution 704}
2 {set resolution 352}
3 {set resolution 352v}
}
set data [readFile [file join $::config(apppath) "lib/data/blank_${standard}_${resolution}.vob"] binary]

# patch the Aspect code in first Video pack
if {[expr {$vattr & 0x0C00}] == 0} {
set aspect 0x20	;# 4:3
} else {
set aspect 0x30	;# 16:9
}
set_int data 0x82B [expr {([get_int data 0x82B 1] & 0x0F) | $aspect}] 1

# patch the LBA numbers in the NAV pack
set_int data 0x002D $lba 4
set_int data 0x040B $lba 4

# patch the VOB and Cell IDs in the NAV pack
set_int data 0x041F $vid 2
set_int data 0x0422 $cid 1

return $data
}

# Arg: VMGM_C_ADT, VTSM_C_ADT or VTS_C_ADT table
# Return a list with the last used VOB_ID
# and the number of sectors used (ie. greatest End Sector +1)
proc lastvid {adt} {
if {$adt == ""} {
return [list 0 0]
}
set lastbyte [get_int adt 4 4]
set lastvid 0
set lastend 0
set idx 8
while {$idx < $lastbyte} {
set vid [get_int adt $idx 2]
set end [get_int adt [expr {$idx+8}] 4]
if {$vid > $lastvid} {set lastvid $vid}
if {$end > $lastend} {set lastend $end}
incr idx 12
}
if {$lastend > 0} {incr lastend}
return [list $lastvid $lastend]
}

# Args: menu|pgc, vts_number
# Return the number of sectors in the VOB file(s) of the specified domain
proc numSectorsInVOB {type vts} {
if {$vts == 0 && $type == "pgc"} {return -1}

set totalsectors 0

if {$type == "menu"} {
set loopstart 0
set loopend 0
} else {
set loopstart 1
set loopend 9
}
if {$vts == 0} {
set filename [file join $::dvddir "VIDEO_TS.VOB"]
} else {
set filename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]
}
for {set v $loopstart} {$v <= $loopend} {incr v} {
if {$vts > 0 && $type == "pgc"} {
set filename [file join $::dvddir "VTS_[format %02d $vts]_${v}.VOB"]
}
if {[file exists $filename]} {
incr totalsectors [expr {[file size $filename]/0x800}]
} else break
}
return $totalsectors
}

# Args: menu|pgc, vts_number
# Return the filename of the last VOB file for the specified domain
proc lastVOBfilename {type vts} {
if {$vts == 0 && $type == "pgc"} {return ""}

if {$type == "menu"} {
if {$vts == 0} {
return [file join $::dvddir "VIDEO_TS.VOB"]
} else {
return [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]
}
} else {
set filename ""
for {set v 1} {$v <= 9} {incr v} {
set tmp [file join $::dvddir "VTS_[format %02d $vts]_${v}.VOB"]
if {[file exists $tmp]} {
set filename $tmp
} else break
}
return $filename
}
}

# Add a blank cell to the specified domain VOBs.
# Args: menu|pgc, vts_number, bool_backup_VOB
# The VOBID to use is computed from the MAT table.  CellID is 1.
# The MAT and VOBU_ADMAP tables are updated.
# Return a list with the VOB ID and cell ID (always 1) writen
# and a binary Cell Playback Information Table entry
# (initialized with default values).
# If the user cancel the requester, return {0 ""}
# Or return {<negative error code> 0 "Error message"}
proc AppendBlankCell {type vts backupvob {quiet false} {vid 0} {cid 1}} {
set utilname "Add Blank VOB Cell"

if {$vts == 0 && $type == "pgc"} {
return [list -2 0 "No VOB for First-Play PGC!"]
}

# compute table names for specified domain
if {$vts == 0} {
set mattablename "VMGM_MAT"
set adttablename "VMGM_C_ADT"
set admaptablename "VMGM_VOBU_ADMAP"
} else {
set mattablename "VTSI_MAT"
if {$type == "menu"} {
set adttablename "VTSM_C_ADT"
set admaptablename "VTSM_VOBU_ADMAP"
} else {
set adttablename "VTS_C_ADT"
set admaptablename "VTS_VOBU_ADMAP"
}
}
if {$type == "menu"} {
set offset [x2d 100]
} else {
set offset [x2d 200]
}
# get current video attributes from the IFOs.
# TODO: Patch it to reflect the real contents of the file (only if file is created)?
set vattr [get_int ::raw_sectors($vts,$mattablename) $offset 2]
if {$vattr==0} {
set attr_present false
if {$vts == 0} {
set vattr 0
} else {
set vattr [get_int ::raw_sectors($vts,$mattablename) 0x200 2]
}
} else {
set attr_present true
}
if {$vattr == 0} {
set standard [::streams::get_video_standard_from_fp_pgc]
if {$standard == "NTSC"} {
set vattr [expr 0x4000]
} else {
set vattr [expr 0x5000]
}
} else {
if {[expr {$vattr & 0X1000}] == 0} {
set standard "NTSC"
} else {
set standard "PAL"
}
}

if {!$attr_present} {
set txt "A new $standard VOB cell will be created."
set txt "$txt\n\nPlease select its aspect ratio:"
if {$::pgcs(numvts) > 0} {
if {[expr [get_int ::raw_sectors(1,VTSI_MAT) 0x200 2] & 0x0C00] == 0x0C00} {
set default "16:9"
} else {
set default "4:3"
}
} else {
set default "4:3"
}
set r [tk_messageBox -message $txt -type "4:3|16:9|cancel" -default $default \
-icon question -title $utilname -parent .]
if {$r == "cancel"} {return [list 0 0 ""]}
if {$r == "16:9"} {
set vattr [expr $vattr | 0x0E00]
} else {
set vattr [expr $vattr & 0xF0FF | 0x0300]
}
}

# retrieve number of VOB IDs and number of sectors used
set adtinfo [lastvid $::raw_sectors($vts,$adttablename)]
set lastvid [lindex $adtinfo 0]
set numsecs [lindex $adtinfo 1]
set createnewvob false
if {$lastvid == 0} {
set txt "No VOB ID found in the $adttablename table!"
set txt "$txt\nCreate a new VOB file?"
set r [tk_messageBox -title $utilname -icon question \
-message $txt -parent . -type okcancel -default ok]
if {$r == "cancel"} {
return [list 0 0 ""]
}
set createnewvob true
} else {
if {$numsecs == 0} {
return [list -4 0 "Last sector found in the $adttablename table is 0!"]
}
}
if {$vid == 0} {
set newvid [expr {$lastvid+1}]
} else {
set newvid $vid
}

# check if the VOB filesize match the number of sectors
set vobsecs [numSectorsInVOB $type $vts]
if {$vobsecs == 0 && !$createnewvob} {
return [list -6 0 "VOB file not found, or empty!"]
}
if {$vobsecs != $numsecs} {
return [list -7 0 "The number of sectors in the VOB file(s) ($vobsecs sectors) don't match the $adttablename table ($numsecs sectors)!\n\nUse FixVTS to remove the additional sectors from the VOB, and retry..."]
}

# retrieve the filename of the last VOB
if {$createnewvob} {
set mode -write
if {$type == "menu"} {
if {$vts == 0} {
set vobfilename [file join $::dvddir "VIDEO_TS.VOB"]
} else {
set vobfilename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]
}
} else {
set vobfilename [file join $::dvddir "VTS_[format %02d $vts]_1.VOB"]
}

# verify the coherence of the video attributes (PAL/NTSC mismatch)
set standard2 [::streams::get_video_standard_from_fp_pgc]
if {$vts == 0 && $::pgcs(numvts) > 0} {
set vattr2 [::streams::get_video_attribute_value pgc 1]
} elseif {$vts > 0} {
set vattr2 [::streams::get_video_attribute_value pgc $vts]
}
if {$standard != $standard2} {
set txt "WARNING: The video standard of the domain ($standard)\n"
set txt "${txt}doesn't match the standard of your DVD ($standard2)."
if {$vattr2 != 0} {
set txt "${txt}\n\nForce the $standard2 standard ?"
set r [tk_messageBox -title $utilname -icon warning \
-message $txt -parent . -type yesnocancel -default yes]
if {$r == "cancel"} {
return [list 0 0 ""]
}
if {$r == "yes"} {
set vattr [expr ($vattr & 0x0F1E) | ($vattr2 & 0x10C1) | 0x4000]
set standard $standard2
::streams::set_video_attribute_value $vattr2 $type $vts
set attr_present true
}
} else {
tk_messageBox -title $utilname -icon error -message $txt -parent .
}
}
} else {
set vobfilename [lastVOBfilename $type $vts]
if {$vobfilename == ""} {
return [list -8 0 "VOB file not found!"]
}

# if resulting VOB will be > 1GB, then create a new VOB.
set mode -append
if {[expr {[file size $vobfilename]+(0x800*6)}] > [x2d 40000000]} {
if {$type == "menu"} {
return [list -9 0 "Menu VOB file size limit of 1GB reached!"]
}
set idx [expr {[string length $vobfilename] -5}]
set vobnum [string range $vobfilename $idx $idx]
if {$vobnum == 9} {
return [list -10 0 "Limit of 9 VTS VOB files reached!"]
}
set vobfilename [string range $vobfilename 0 [expr {$idx-1}]][expr {$vobnum+1}][string range $vobfilename [expr {$idx+1}] end]
set mode -write
}

# check if file is writable
if {$mode == "-append" && ! [file writable $vobfilename]} {
return [list -11 0 "Read-only VOB file!"]
}
}

# Confirm action!
if {! $quiet} {
if {$mode == "-append"} {
set txt "Append blank cell (VOB ID $newvid, Cell ID $cid)\nto \"$vobfilename\"?"
} else {
set txt "Create blank cell (VOB ID $newvid, Cell ID $cid)\nas \"$vobfilename\"?"
}
set r [tk_messageBox -title $utilname -icon question \
-message $txt -parent . -type okcancel -default ok -skip append_blank_cell_to_vob]
if {$r == "cancel"} {
return [list 0 0 ""]
}
}

if {!$backupvob && $mode == "-append" && ![file exists [file join $::backupdir [file tail $vobfilename]]]} {
set txt "A new cell will be added to \"[file tail $vobfilename]\"."
set txt "$txt\nThere is no backup of this VOB file yet."
set txt "$txt\n\nDo you want to back it up now?  (Recommended!)"
set r [tk_messageBox -title $utilname -icon question \
-message $txt -parent . -type yesnocancel -default yes -skip append_blank_cell_backup]
if {$r == "cancel"} {
return [list 0 0 ""]
} elseif {$r == "yes"} {
set backupvob true
}
}

# Backup VOB
if {$backupvob} {
set back true
set vobtailname [file tail $vobfilename]
if {! [file exists $::backupdir]} {
makedir $::backupdir "PgcEdit: Create Backup" .
} else {
if {[file exists [file join $::backupdir $vobtailname]]} {
set r [tk_messageBox -title "PgcEdit: Backup" -icon warning \
-message "Backup of VOB file already exists!\nOverwrite?" -parent . \
-type yesnocancel -default no]
if {$r == "cancel"} {return [list 0 0 ""]}
if {$r == "no"} {set back false}
}
}
if {$back} {
show_msg "Copying to backup: \"$vobtailname\""
update idletasks
after 50 {set ::pause false}
tkwait variable ::pause
catch {file delete [file join $::backupdir $vobtailname]}
#blocking_msg "Backing up VOB file..."
CopyFile $vobfilename $::backupdir -force "PgcEdit: Append Blank Cell" .
#close_blocking_msg
}
}

set blankcelldata [blankCell $vattr $newvid $cid $numsecs]
set rc [saveFile $vobfilename binary $blankcelldata $mode]
if {$rc != 0} {
return [list -13 0 "Error writing to file!"]
}

# Retrieve the number of sectors of the modified VOB file
set newnumsecs [numSectorsInVOB $type $vts]
if {$newnumsecs <= $numsecs} {
return [list -14 0 "UNKNOW ERROR: VOB file size wrong after writing!"]
}

# Update ADT table
# append new cell informations at end of table
if {$createnewvob} {
set ::raw_sectors($vts,$adttablename) [string repeat \000 8]
}
set tmp [string repeat \000 12]
set_int tmp 0 $newvid 2
set_int tmp 2 $cid 1
set_int tmp 4 $numsecs 4
set_int tmp 8 [expr {$newnumsecs-1}] 4
set ::raw_sectors($vts,$adttablename) "$::raw_sectors($vts,$adttablename)$tmp"
# increment number of cells at offset 0
set_int ::raw_sectors($vts,$adttablename) 0 [expr {1+[get_int ::raw_sectors($vts,$adttablename) 0 2]}] 2

# Update VOBU_ADMAP table
if {$createnewvob} {
set ::raw_sectors($vts,$admaptablename) [string repeat \000 4]
}
set tmp [string repeat \000 4]
set_int tmp 0 $numsecs 4
set ::raw_sectors($vts,$admaptablename) "$::raw_sectors($vts,$admaptablename)$tmp"

# Create a Cell PB Information Table entry, ready to be used in the PGC.
# (It will be returned as the 2nd argument in the return list.)
set pbentry [string repeat \000 24]
set_int pbentry 0 2 1	;# Cell Type flags default to STC Discontinuity
if {$standard == "PAL"} {
set_int pbentry 4 [x2d 52] 4
} else {
set_int pbentry 4 [x2d D2] 4
}
set_int pbentry 8  $numsecs 4
set_int pbentry 16 $numsecs 4
set_int pbentry 20 [expr {$newnumsecs-1}] 4

# Done.
show_msg "$utilname: Blank cell (VOBID $newvid, CellID $cid) appended OK."
::main::dvd_modified true false

# one subpic stream must be present in the menu domain
if {$type == "menu" && [get_int ::raw_sectors($vts,$mattablename) 0x155 1] < 1} {
set_int ::raw_sectors($vts,$mattablename) 0x155 1 1
}
if {!$attr_present} {
::streams::set_video_attribute_value $vattr $type $vts
set txt "Since you have just created the first VOB cell"
set txt "$txt\nof the current Menu domain, the Video Attributes"
set txt "$txt\nof the domain have been forced, and a new Subpic"
set txt "$txt\nstream has been defined."
set txt "$txt\nPlease verify them with Domain Streams Attributes."
set txt "$txt\n\n(I'll save the DVD right now.)"
tk_messageBox -title $utilname -icon info -message $txt -parent .pgceditor \
-skip create_first_blank_cell
}
return [list $newvid 1 $pbentry]
}

# Verify if the last cell of the last VOB of the specified domain is a tiny black cell.
# If it's a tiny cell, return a list with his VOB and Cell ID, and a binary Cell Playback
# Information Table entry (initialized with default values).
# Else, return {0 0 ""}
# Example usage:  ::blankcell::LastVOBCellIsBlankCell pgc 5
proc LastVOBCellIsBlankCell {type vts} {
if {$vts == 0 && $type == "pgc"} {
return {0 0 ""}
}

set vobfilename [lastVOBfilename $type $vts]
if {$vobfilename == ""} {
return {0 0 ""}
}
if {! [file exists $vobfilename]} {
return {0 0 ""}
}

# compute table names for specified domain
if {$vts == 0} {
set mattablename "VMGM_MAT"
set adttablename "VMGM_C_ADT"
set admaptablename "VMGM_VOBU_ADMAP"
} else {
set mattablename "VTSI_MAT"
if {$type == "menu"} {
set adttablename "VTSM_C_ADT"
set admaptablename "VTSM_VOBU_ADMAP"
} else {
set adttablename "VTS_C_ADT"
set admaptablename "VTS_VOBU_ADMAP"
}
}
if {$type == "menu"} {
set offset [x2d 100]
} else {
set offset [x2d 200]
}
# get current video attributes from the IFOs.
# TODO: Patch it to reflect the real contents of the file (only if file is created)?
set vattr [get_int ::raw_sectors($vts,$mattablename) $offset 2]
if {[expr {$vattr & 0X1000}] == 0} {
set standard "NTSC"
} else {
set standard "PAL"
}
set blankcelltemplate [blankCell $vattr 1 1 0]
set blankcelllength [string length $blankcelltemplate]
set origstart [expr [file size $vobfilename] - $blankcelllength]
if {$origstart < 0} {
return {0 0 ""}
}
set blankcellorig [readFile $vobfilename binary $origstart]
if {$blankcellorig == ""} {
return {0 0 ""}
}

# patch the LBA numbers in the NAV pack
set orig_lba [get_int blankcellorig 0x002D 4]
set_int blankcellorig 0x002D 0 4
set_int blankcellorig 0x040B 0 4
# patch the VOB and Cell IDs in the NAV pack
set orig_vid [get_int blankcellorig 0x041F 2]
set_int blankcellorig 0x041F 1 2
set orig_cid [get_int blankcellorig 0x0422 1]
set_int blankcellorig 0x0422 1 1

# If last cell is not equal to template, then the cell is not a blank cell,
# (or at least it is not a blank cell created by PgcEdit!)
if {[string compare $blankcelltemplate $blankcellorig] != 0} {
return {0 0 ""}
}

# Create a Cell PB Information Table entry, ready to be used in the PGC.
# (It will be returned as the 2nd argument in the return list.)
set pbentry [string repeat \000 24]
set_int pbentry 0 2 1	;# Cell Type flags default to STC Discontinuity
if {$standard == "PAL"} {
set_int pbentry 4 [x2d 52] 4
} else {
set_int pbentry 4 [x2d D2] 4
}
set_int pbentry 8  $orig_lba 4
set_int pbentry 16 $orig_lba 4
set_int pbentry 20 [expr {[numSectorsInVOB $type $vts]-1}] 4

# Done.
return [list $orig_vid $orig_cid $pbentry]
}

# save as LastVOBCellIsBlankCell, but ask the user if he want to use the last cell
# Return an empty string if he answer no, or if the last cell is not a tiny blank cell.
# Else, return same list than LastVOBCellIsBlankCell.
# Example:  ::blankcell::askUseLastVOBCell pgc 5
proc askUseLastVOBCell {type vts} {
set returnlist [LastVOBCellIsBlankCell $type $vts]
if {[lindex $returnlist 0] != 0} {
set txt "The last cell in the VOB file \"[file tail [lastVOBfilename $type $vts]]\""
set txt "$txt\nis already a blank cell (with VOB/Cell ID [lindex $returnlist 0]/[lindex $returnlist 1].)"
set txt "$txt\n\nIt is recommended to reuse that cell instead"
set txt "$txt\nof creating a new one if you don't want to"
set txt "$txt\nreplace the cell later with FixVTS or VobBlanker."
set txt "$txt\n\nReuse last tiny blank cell?"
set r [tk_messageBox -title "Blank Cell" -message $txt -icon question \
-type yesno -default yes]
if {$r == "no"} {
set returnlist {0 0 ""}
}
}
return $returnlist
}

# If there is a blank cell ready to be reused, ask the user if he wavts to reuse it.
# If there is no blank cell or if the user answer no, create a new blank cell.
proc ReuseOrAppendBlankCell {type vts backupvob {quiet false} {vid 0} {cid 1}} {
set returnlist [askUseLastVOBCell $type $vts]
if {[lindex $returnlist 0] == 0} {
set returnlist [AppendBlankCell $type $vts $backupvob true $vid $cid]
}
return $returnlist
}

# return the number of buttons in the cell
# or -1 if the title cell has not been scanned for BOVs
proc NumButtonsInCell {type index cell} {
set numbuts 0
if {$type == "pgc"} {
set vts [lindex [split $index ,] 0]
if {$vts == 0} {return 0}
set cellpostable $::pgcs($index,cellpostable)
set idx [expr ($cell-1) * 4]
set vid [get_int cellpostable $idx 2]
set cid [get_int cellpostable [expr $idx+3] 1]
if {[info exists ::pgcs($index,numbuttons)] && \
$::pgcs($index,numbuttons)>0} {
if {[info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)]} {
set numbuts $::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)
}
} else {
set numbuts -1
}
} else {
set vts [lindex [split $index ,] 0]
set numcells $::menus($index,numcells)
set cellpostable $::menus($index,cellpostable)
set idx [expr ($cell-1) * 4]
set vid [get_int cellpostable $idx 2]
set cid [get_int cellpostable [expr $idx+3] 1]
if {$::menus($index,numbuttons) > 0} {
if {[info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)]} {
set numbuts $::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)
}
}
}
return $numbuts
}

# Replace the cell by a tiny black cell
# If force is true, do it also if the cehh has menu buttons
# Return 1 if cell is blanked, 0 if not blanked (cancelled),
# or -1 if cell has buttons and is not blanked because force==false
# Examples: ::blankcell::blankThisCell pgc 1,1 1
#           ::blankcell::blankThisCell menu 0,1,5 1
proc blankThisCell {type index cell {force false}} {
set backupvob false	;# TODO
set indexslist [split $index ,]
if {$type == "pgc"} {
set vts [lindex $indexslist 0]
if {$vts == 0} {bell; return 0}
set lu  0
set pgc [lindex $indexslist 1]
set numcells $::pgcs($vts,$pgc,numcells)
set cellpbtable  $::pgcs($vts,$pgc,cellpbtable)
set cellpostable $::pgcs($vts,$pgc,cellpostable)
set header $::pgcs($vts,$pgc,header)
set label [::main::pgc_label "" pgc $vts $pgc]
} else {
set vts [lindex $indexslist 0]
set lu  [lindex $indexslist 1]
set pgc [lindex $indexslist 2]
set numcells $::menus($vts,$lu,$pgc,numcells)
set cellpbtable  $::menus($vts,$lu,$pgc,cellpbtable)
set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
set header $::menus($vts,$lu,$pgc,header)
set label [::main::pgc_label "" menu $vts $lu $pgc]
}
set cellpostableidx [expr ($cell-1) * 4]
set cell_vid [get_int cellpostable $cellpostableidx 2]
set cell_cid [get_int cellpostable [expr $cellpostableidx+3] 1]
set cell_lba [get_int cellpbtable [expr (($cell-1)*24)+8] 4]

if {$numcells == 0 || $cell > $numcells} {bell; return 0}
set numbuts [NumButtonsInCell $type $index $cell]
if {$numbuts < 0} {set numbuts 0}
if {$numbuts > 0 && ! $force} {
set txt "Blanking Cell $cell of $label."
set txt "$txt\n\nWARNING:  There are $numbuts buttons in this cell!"
set txt "$txt\nBlanking this cell could (and probably will) break the navigation!"
set txt "$txt\n\nAre you sure you want to blank this cell?  (NOT recommended!)"
set r [tk_messageBox -title "Blanking cell" -icon warning \
-message $txt -parent . -type yesno -default no]
if {$r == "no"} {
show_msg "Blanking Cell $cell of $label: Aborted"
return -1
}
}

set safekillpb [::macros::check_safe_kill_pb $type $vts $lu $pgc]
if {$numcells == 1 && $type == "menu" && ! $force && $safekillpb==1} {
set txt "Blanking Cell $cell of $label."
set txt "$txt\n\nAs there is only one Cell in this Menu PGC, it is easier and safer"
set txt "$txt\nto use the Kill Playback method to remove the cell from the PGC."
set txt "$txt\n\nKill Playback of this PGC?"
set txt "$txt\n(YES to Kill PB, NO to blank cell, CANCEL to do nothing)"
set r [tk_messageBox -title "Blanking cell" -icon question \
-message $txt -parent . -type yesnocancel -default yes]
if {$r == "yes"} {
if {[::macros::kill_pgc_playback ::${type}s($index) safely always] == 1} {
show_msg "Blanking Cell $cell of $label: Playback killed"
return 1
} else {
set safekillpb 0
set txt "Blanking Cell $cell of $label."
set txt "$txt\n\nAs it is not safe to Kill the Playback of this PGC,"
set txt "$txt\nit is recommended to blank the cell.  Sorry!"
set txt "$txt\n\nBlank this cell?"
set txt "$txt\n(OK to blank the Cell, CANCEL to do nothing)"
set r [tk_messageBox -title "Blanking cell" -icon question \
-message $txt -parent . -type okcancel -default ok]
if {$r == "cancel"} {
show_msg "Blanking Cell $cell of $label: Aborted"
return 0
}
}
}
if {$r == "cancel"} {
show_msg "Blanking Cell $cell of $label: Aborted"
return 0
}
}

# Append a blank cell to the VOB (or reuse last tiny cell)
set resultlist [LastVOBCellIsBlankCell $type $vts]
if {[lindex $resultlist 0] == 0} {
show_msg "Blanking Cell $cell of $label: Appending new cell"
set resultlist [AppendBlankCell $type $vts $backupvob true]
}

set new_vobid  [lindex $resultlist 0]
set new_cellid [lindex $resultlist 1]
set src_cellpbentry [lindex $resultlist 2]
if {$new_vobid == 0} {
return 0
}
if {$new_vobid < 0} {
set txt "Error appending new Cell to VOB file:\n"
set txt "$txt\n$src_cellpbentry"
tk_messageBox -title "Append Blank Cell to VOB" -icon error \
-message $txt -parent .
show_msg "Blanking Cell $cell of $label: $src_cellpbentry"
return 0
}
set new_lba [get_int src_cellpbentry 8 4]
if {$cell_lba == $new_lba} {
show_msg "Blanking Cell $cell of $label: Cell is already blanked!"
return 1
}

if {$type == "menu"} {
incr ::menus($index,numbuttons) [expr {$numbuts * -1}]
} else {
if {[info exists ::pgcs($index,numbuttons)]} {
incr ::pgcs($index,numbuttons) [expr {$numbuts * -1}]
}
}
set ::buttons::butsinfo($type,$vts,numvobids) $new_vobid
set ::buttons::butsinfo($type,$vts,$::buttons::butsinfo($type,$vts,numvobids),numcellids) 1
set ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,navpack) $new_lba
set ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,numbuts) 0
set ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,numgroups) 0
unset -nocomplain ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,forceselectbutn)
unset -nocomplain ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,forceactionbutn)
unset -nocomplain ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,1,grouptype)
unset -nocomplain ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,2,grouptype)
unset -nocomplain ::buttons::butsinfo($type,$vts,$new_vobid,$new_cellid,3,grouptype)

# copy the cell type and the cell command number from the original cell
set cellpbidx [expr {($cell - 1) * 24}]
set celltype [expr {([get_int cellpbtable $cellpbidx 1] & 0xF7) | 2}] ;# add SRC Discontinuity flag, and remove Seamless Joint
set cmdnum [get_int cellpbtable [expr {$cellpbidx + 3}] 1]
set_int src_cellpbentry 0 $celltype 1
set_int src_cellpbentry 3 $cmdnum 1

# replace the new entry in the current cell playback table
set cellpbtable "[string range $cellpbtable 0 [expr {$cellpbidx-1}]]$src_cellpbentry[string range $cellpbtable [expr {$cellpbidx+24}] end]"

# set the VOB ID and Cell ID in the Cell Position table
set_int cellpostable $cellpostableidx $new_vobid 2
set_int cellpostable [expr {$cellpostableidx + 3}] $new_cellid 1

# Set the STC Discontinuity flag in the modified Cell entry, and clear the Seamless flag
set_int cellpbtable $cellpbidx [expr {([get_int cellpbtable $cellpbidx 1] | 2) & 0xF7}] 1
# Set the STC Discontinuity flag in the NEXT Cell entry, and clear the Seamless flag
if {$numcells > $cell} {
set idx [expr {$cell * 24}]
set_int cellpbtable $idx [expr {([get_int cellpbtable $idx 1] | 2) & 0xF7}] 1
}

# Fix PGC total playback time, and check if it is recommended to kill the PGC playback
set totaltime "\000\000\000\000"
set_int totaltime 3 [expr [get_int cellpbtable 7 1] & 0XC0] 1
set iskillable true
for {set c 1} {$c <= $numcells} {incr c} {
set cellidx1 [expr {($c - 1) * 24}]
set celltype [get_int cellpbtable $cellidx1 1]
# Add cell time to total PGC time, but only if angle 1.
if {[expr {$celltype & 0x10}] == 0 || [expr {$celltype & 0xC0}] == 0x40} {
set totaltime [::pgced::AddBCDtime [string range $cellpbtable [expr {$cellidx1 + 4}] \
[expr {$cellidx1 + 7}]] $totaltime]
}
if {[get_int cellpbtable [expr $cellidx1+8] 4] != $new_lba} {
if {! $::config(trace_skip_small_cells_preview) || \
[get_int cellpbtable [expr $cellidx1+4] 4] > 256 || \
[get_int cellpbtable [expr $cellidx1+2] 1] > 0} {
set iskillable false
}
}
}
set header "[string range $header 0 3]$totaltime[string range $header 8 end]"

# Remember the new PGC tables
if {$type == "pgc"} {
set ::pgcs($vts,$pgc,header) $header
set ::pgcs($vts,$pgc,cellpbtable)  $cellpbtable
set ::pgcs($vts,$pgc,cellpostable) $cellpostable
} else {
set ::menus($vts,$lu,$pgc,header) $header
set ::menus($vts,$lu,$pgc,cellpbtable)  $cellpbtable
set ::menus($vts,$lu,$pgc,cellpostable) $cellpostable
}
show_msg "Blanking Cell $cell of $label: Cell blanked OK"

if {$safekillpb==1 && $iskillable} {
set txt "Blanking Cell $cell of $label."
set txt "$txt\n\nAs all Cells in this"
if {$type == "menu"} {set txt "$txt Menu"} else {set txt "$txt Title"}
set txt "$txt PGC are now blanked,"
set txt "$txt\nyou probably want to Kill the PGC Playback, so"
set txt "$txt\nthat you will not notice the tiny blank cells at all."
set txt "$txt\n\nKill Playback of this PGC?"
set r [tk_messageBox -title "Blanking cell" -icon question \
-message $txt -parent . -type yesno -default yes]
if {$r == "yes"} {
::macros::kill_pgc_playback ::${type}s($index) safely always
show_msg "Blanking Cell $cell of $label: Playback killed"
}
}
# refresh PGC selector to show the new number of menubuttons and playback time
::main::init_tree_menu_no_move
::main::dvd_modified true false
# rebuild the time map of the current PGC
if {$type == "pgc" && $vts>0} {
::tmap::need_new_tmap $vts $pgc
}
update idletasks
if {![info exists ::blankcell::fixvts_usage_message]} {
set ::blankcell::fixvts_usage_message true
set txt "Blanking Cell $cell of $label."
set txt "$txt\n\nYou have successfully blanked the cell."
set txt "$txt\n\nThe original cell is NOT removed from the VOB file."
set txt "$txt\nIf you want to get the space used by the old cell back, you"
set txt "$txt\nshould use FixVTS or VobBlanker to process the current domain"
set txt "$txt\n(or the whole DVD.)  If you use VobBlanker, it is sufficient"
set txt "$txt\nto process the domain or DVD. You don't need (and you will"
set txt "$txt\nprobably not be able) to select the original cells to blank"
set txt "$txt\nthem again!  The unreferenced cells will be dropped anyway."
set txt "$txt\n\nWarning: Don't forget to quit PgcEdit or to reload the DVD"
set txt "$txt\nin PgcEdit after the FixVTS/VobBlanker pass, or the IFOs will"
set txt "$txt\nnot be synchronized with the VOBs!"
set txt "$txt\n\nOf course, if you want to blank more cells, do it now, and"
set txt "$txt\napply the FixVTS/VobBlanker pass when all cells are blanked."
set txt "$txt\n\n(Note: This message is shown only once per session.)"
tk_messageBox -title "Blanking cell" -icon info -message $txt -parent . \
-skip "BlankCell/FixVTS usage"
}
return 1
}


}	;# End namespace ::blankcell::
