#############################################################################
# dvdtrace.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# Trace DVD functions.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

# TODO:
# - Use SPRMs 15 to 19 to determine the current audio and subtitle.
#   (if sprm(1:audio stream) and sprm(2:subpic stream) are not specified)
# - Simulation for SetNVTMR and SetGPRMMD in counter mode. (difficult!)
# - Cleanup the code
# - Re-check everything :(
# - Documentation. (pfff...)


namespace eval ::tr {


proc kill_trace {} {
	if {$::tr::tracemode} {
		cleanexit
		tkwait variable ::tr::tr_killed
		set ::tr::tracemode 0
	}
	unset -nocomplain ::tr::snapshot_bank
	destroy .mf.cf.trace
}



################################ Trace entry proc (GUI) ###########################
proc dvd_trace {} {
	set utiltitle "Trace"

	if {! [info exists ::dvddir]} {
		set ::tr::tracemode 0
		bell
		return
	}
	if {$::scratchbook_active} {
		set ::scratchbook_active false
		::main::scratchbook
	}

	set w .mf.cf.trace
	set ::tr::wm_main_window $w
	if {[winfo exists $w] && [winfo ismapped $w]} {
		cleanexit
		return
	}

	set ::tr::tracemode 1

	# disable unwanted parts of the interface
	set ::tr::toolbar_opened $::toolbar::toolbar_opened
	set ::tr::buttons_opened $::main::buttons_opened
	if {$::tr::hide_gui} {
		::toolbar::init_toolbar false
		set ::toolbar::toolbar_opened false
		::main::action_buttons hide
		pack forget .mf.st
	}
	.mf.f1.sb configure -state disabled

	if {![winfo exists $w]} {
		# init all GPRMs and SPRMs 'Changed' flags
		if {! [info exists ::tr::gprmbreak(0)]} {
			for {set g 0} {$g < 16} {incr g} {
				set ::tr::gprm_changed($g) 0
				set ::tr::gprmbreak($g) 0
			}
		}
		for {set s 0} {$s < 24} {incr s} {
			set ::tr::sprm_changed($s) 0
			set ::tr::sprmbreak($s) 0
		}

		init_gui_buttons

		catch {destroy .tracemenu}
		menu .tracemenu

		frame $w

		############# Left part
		pack [frame $w.fl] -side left

		############# First row: simulate a menu

		pack [frame $w.fl.mnu -bd 1 -relief sunken] -fill x -padx 2
		pack [menubutton $w.fl.mnu.watch -text "Watch" -menu $w.fl.mnu.watch.m_watch \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.watch.m_watch
			$w.fl.mnu.watch.m_watch add checkbutton -label "Input" -variable ::tr::view_input -command ::tr::configure_info_window
			$w.fl.mnu.watch.m_watch add checkbutton -label "GPRMs" -variable ::tr::view_gprms -command ::tr::configure_info_window
			$w.fl.mnu.watch.m_watch add checkbutton -label "GPRM labels" -variable ::tr::view_glabs -command "::tr::info_window_gprms .trinfo"
			$w.fl.mnu.watch.m_watch add checkbutton -label "SPRMs" -variable ::tr::view_sprms -command ::tr::configure_info_window
			$w.fl.mnu.watch.m_watch add checkbutton -label "SPRM labels" -variable ::tr::view_slabs -command "::tr::info_window_sprms .trinfo"
			$w.fl.mnu.watch.m_watch add checkbutton -label "PUOs" -variable  ::tr::view_puos  -command ::tr::configure_info_window
			$w.fl.mnu.watch.m_watch add separator
			$w.fl.mnu.watch.m_watch add command -label "Registers" \
					-command {set ::tr::view_input 1; set ::tr::view_glabs 1; \
							set ::tr::view_gprms 1; set ::tr::view_sprms 1; \
							set ::tr::view_puos 0; ::tr::configure_info_window}
			$w.fl.mnu.watch.m_watch add command -label "All" \
					-command {set ::tr::view_input 1; set ::tr::view_puos 1; \
							set ::tr::view_gprms 1; set ::tr::view_glabs 1; \
							set ::tr::view_sprms 1; set ::tr::view_slabs 1; \
							::tr::configure_info_window}
			$w.fl.mnu.watch.m_watch add separator
			$w.fl.mnu.watch.m_watch add command -label "Define GPRM labels" -command {::utils::set_gprm_labels}

		.tracemenu add cascade -label "Watch" -menu $w.fl.mnu.watch.m_watch

		pack [menubutton $w.fl.mnu.bp -text "Breakpoints" -menu $w.fl.mnu.bp.m_bp \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.bp.m_bp
			$w.fl.mnu.bp.m_bp add cascade -label "GPRM write breakpoints" -menu $w.fl.mnu.bp.m_bp.gprm
			$w.fl.mnu.bp.m_bp add cascade -label "SPRM write breakpoints" -menu $w.fl.mnu.bp.m_bp.sprm
			menu $w.fl.mnu.bp.m_bp.gprm
			for {set i 0} {$i < 16} {incr i} {
				set b false
				if {$i == 8} {set b true}
				$w.fl.mnu.bp.m_bp.gprm add checkbutton -label "GPRM $i" -variable ::tr::gprmbreak($i) -columnbreak $b
				trace add variable ::tr::gprmbreak($i) write ::tr::highlight_gprm_break_changes
			}
			$w.fl.mnu.bp.m_bp.gprm add command -label "Clear all" -command {for {set i 0} {$i < 16} {incr i} {set ::tr::gprmbreak($i) 0}}
			menu $w.fl.mnu.bp.m_bp.sprm
			for {set i 0} {$i < 24} {incr i} {
				set b false
				if {$i == 7} {set b true}
				if {[lsearch {1 2 3 4 5 6 7 8 9 10 11 14} $i] >= 0} {
					$w.fl.mnu.bp.m_bp.sprm add checkbutton -label "SPRM $i" -variable ::tr::sprmbreak($i) -columnbreak $b
				}
				trace add variable ::tr::sprmbreak($i) write ::tr::highlight_sprm_break_changes
			}
			$w.fl.mnu.bp.m_bp.sprm add command -label "Clear all" -command {for {set i 0} {$i < 24} {incr i} {set ::tr::sprmbreak($i) 0}}

			$w.fl.mnu.bp.m_bp add separator
			$w.fl.mnu.bp.m_bp add command -label "Toggle PGC pre/post breakpoint" -command {::tr::toggle_breakpoint prepost}  -accelerator "Ctrl+Shift+T"
			$w.fl.mnu.bp.m_bp add command -label "Toggle PGC playback breakpoint" -command {::tr::toggle_breakpoint playback} -accelerator "Ctrl+Alt+T"
			$w.fl.mnu.bp.m_bp add command -label "Clear all PGC breakpoints" -command ::tr::clear_all_breakpoints
			$w.fl.mnu.bp.m_bp add separator
			$w.fl.mnu.bp.m_bp add checkbutton -label "Break at all Menus"  -variable ::tr::break_at_all_menus
			$w.fl.mnu.bp.m_bp add checkbutton -label "Break at all Titles" -variable ::tr::break_at_all_titles
			$w.fl.mnu.bp.m_bp add separator
			$w.fl.mnu.bp.m_bp add checkbutton -label "Break at log output" -variable ::tr::break_at_log_output \
					-command "if \$::tr::break_at_log_output {if {! \[winfo ismapped $w.fr.b\]} {pack  $w.fr.b -side bottom -fill x}; focus $w.fr.b.s.e; set ::tr::search_visible 1}"
		set ::tr::break_at_all_menus  0
		set ::tr::break_at_all_titles 0

		.tracemenu add cascade -label "Breakpoints" -menu $w.fl.mnu.bp.m_bp

		pack [menubutton $w.fl.mnu.state -text "State" -menu $w.fl.mnu.state.m_state \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.state.m_state
			$w.fl.mnu.state.m_state add cascade -label "Return to"  -menu $w.fl.mnu.state.m_state.auto
			menu $w.fl.mnu.state.m_state.auto
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(pgc) \
						-label "Current PGC"	-command {::tr::restore_state pgc}
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(last_pgc) \
						-label "Last PGC"		-command {::tr::restore_state last_pgc}
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(domain) \
						-label "Current Domain"	-command {::tr::restore_state domain}
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(last_domain) \
						-label "Last Domain"	-command {::tr::restore_state last_domain}
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(pb) \
						-label "Last Playback"	-command {::tr::restore_state pb}
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(menu) \
						-label "Last Menu"		-command {::tr::restore_state menu}
				$w.fl.mnu.state.m_state.auto add separator
				$w.fl.mnu.state.m_state.auto add checkbutton -variable ::tr::snapshot_bank(auto) \
						-label "Trace Mode closed" -command {::tr::restore_state auto}

			$w.fl.mnu.state.m_state add cascade -label "Snapshot State" -menu $w.fl.mnu.state.m_state.rem
			menu $w.fl.mnu.state.m_state.rem
				for {set i 1} {$i <= 4} {incr i} {
					$w.fl.mnu.state.m_state.rem add checkbutton -variable ::tr::snapshot_bank($i) \
							-label "User bank $i" -command "::tr::snapshot_state $i"
				}
				$w.fl.mnu.state.m_state.rem add separator
				$w.fl.mnu.state.m_state.rem add checkbutton -variable ::tr::snapshot_bank(pos) \
						-label "Only Position"	-command {::tr::snapshot_pos pos}
				$w.fl.mnu.state.m_state.rem add checkbutton -variable ::tr::snapshot_bank(gprm) \
						-label "Only GPRMs"		-command {::tr::snapshot_gprms gprm}
				$w.fl.mnu.state.m_state.rem add checkbutton -variable ::tr::snapshot_bank(sprm) \
						-label "Only SPRMs"		-command {::tr::snapshot_sprms sprm}

			$w.fl.mnu.state.m_state add cascade -label "Restore State"  -menu $w.fl.mnu.state.m_state.res
			menu $w.fl.mnu.state.m_state.res
				for {set i 1} {$i <= 4} {incr i} {
					$w.fl.mnu.state.m_state.res add checkbutton -variable ::tr::snapshot_bank($i) \
							-label "User bank $i" -command "::tr::restore_state $i"
				}
				$w.fl.mnu.state.m_state.res add separator
				$w.fl.mnu.state.m_state.res add checkbutton -variable ::tr::snapshot_bank(pos) \
						-label "Only Position"	-command {::tr::restore_pos pos}
				$w.fl.mnu.state.m_state.res add checkbutton -variable ::tr::snapshot_bank(gprm) \
						-label "Only GPRMs"		-command {::tr::restore_gprms gprm}
				$w.fl.mnu.state.m_state.res add checkbutton -variable ::tr::snapshot_bank(sprm) \
						-label "Only SPRMs"		-command {::tr::restore_sprms sprm}

			$w.fl.mnu.state.m_state add command -label "Clear All Snapshots" \
					-command {::tr::clear_all_state_banks; ::tr::snapshot_state pgc}

			$w.fl.mnu.state.m_state add separator
			$w.fl.mnu.state.m_state add command -label "Clear 'Register Changed' flags" \
					-command ::tr::clear_changed_flags
			$w.fl.mnu.state.m_state add command -label "Clear 'PGC played' highlights" \
					-command ::main::init_tree_menu_no_move

			$w.fl.mnu.state.m_state add separator
			$w.fl.mnu.state.m_state add command -label "Restart playback from Exit domain" \
					-command ::tr::exit2play

		.tracemenu add cascade -label "State" -menu $w.fl.mnu.state.m_state

		pack [menubutton $w.fl.mnu.log -text "Log" -menu $w.fl.mnu.log.m_log \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.log.m_log
			$w.fl.mnu.log.m_log add cascade -label "Log Filter" -menu $w.fl.mnu.log.m_log.filters
			menu $w.fl.mnu.log.m_log.filters
				$w.fl.mnu.log.m_log.filters add checkbutton -label Commands -variable ::tr::log_filter(cmd)
				set ::tr::log_filter(cmd) 1
				$w.fl.mnu.log.m_log.filters add checkbutton -label "If conditions" -variable ::tr::log_filter(if)
				set ::tr::log_filter(if) 1
				$w.fl.mnu.log.m_log.filters add checkbutton -label GPRMs -variable ::tr::log_filter(gprm)
				set ::tr::log_filter(gprm) 1
				$w.fl.mnu.log.m_log.filters add checkbutton -label SPRMs -variable ::tr::log_filter(sprm)
				set ::tr::log_filter(sprm) 1
				$w.fl.mnu.log.m_log.filters add checkbutton -label "PUOs value" -variable ::tr::log_filter(puo)
				set ::tr::log_filter(puo) 1
				$w.fl.mnu.log.m_log.filters add checkbutton -label "PUOs details" -variable ::tr::log_filter(puo_details)
				set ::tr::log_filter(puo_details) 0

				set ::tr::log_filter(always) 1
				$w.fl.mnu.log.m_log.filters add separator
				$w.fl.mnu.log.m_log.filters add command -label "Log all" \
						-command {foreach i {cmd if gprm sprm puo puo_details} {set ::tr::log_filter($i) 1}}
				$w.fl.mnu.log.m_log.filters add command -label "Log none" \
						-command {foreach i {cmd if gprm sprm puo puo_details} {set ::tr::log_filter($i) 0}}

			$w.fl.mnu.log.m_log add separator
			$w.fl.mnu.log.m_log add command -label "Log all GPRMs now" -command ::tr::log_gprms_now
			$w.fl.mnu.log.m_log add command -label "Log all SPRMs now" -command ::tr::log_sprms_now

		.tracemenu add cascade -label "Log" -menu $w.fl.mnu.log.m_log

		pack [menubutton $w.fl.mnu.clip -text "Clipboard" -menu $w.fl.mnu.clip.m_clip \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.clip.m_clip
			$w.fl.mnu.clip.m_clip add command -label "Copy Used Registers as commands" -command {::tr::copy_regs_as_cmds used clip}
			$w.fl.mnu.clip.m_clip add command -label "Copy Changed Registers as commands" -command {::tr::copy_regs_as_cmds changed clip}
			$w.fl.mnu.clip.m_clip add separator
			$w.fl.mnu.clip.m_clip add command -label "Append Used Registers as cmds to Scratchbook" -command {::tr::copy_regs_as_cmds used scratch}
			$w.fl.mnu.clip.m_clip add command -label "Append Changed Registers as cmds to Scratchbook" -command {::tr::copy_regs_as_cmds changed scratch}

		.tracemenu add cascade -label "Clipboard" -menu $w.fl.mnu.clip.m_clip

		pack [menubutton $w.fl.mnu.setup -text "Setup" -menu $w.fl.mnu.setup.m_setup \
				-activebackground gray -bd 0 -relief flat -indicatoron false] -side left
		menu $w.fl.mnu.setup.m_setup
			$w.fl.mnu.setup.m_setup add command -label "Virtual Player Setup" -command ::tr::player_setup
			if {$::config(use_preview)} {
				$w.fl.mnu.setup.m_setup add cascade -menu $w.fl.mnu.setup.m_setup.preview -label "Preview options"
				menu $w.fl.mnu.setup.m_setup.preview
				$w.fl.mnu.setup.m_setup.preview add checkbutton -variable ::config(trace_auto_preview) \
						-label "Open Preview automatically when playing a cell ?" -onvalue true -offvalue false
				$w.fl.mnu.setup.m_setup.preview add checkbutton -label "Automatically start preview Playback?" \
						-variable ::config(preview_autoplay)
				$w.fl.mnu.setup.m_setup.preview add checkbutton -label "Real Time Playback?" \
						-variable ::config(preview_realtime)
				$w.fl.mnu.setup.m_setup.preview add checkbutton -variable ::config(trace_auto_close) \
						-label "Close preview automatically at the end of the cell playback ?"
				$w.fl.mnu.setup.m_setup.preview add command -label "Close preview automatically after a short delay ?" \
						-command {tk_messageBox -title "PgcEdit: Trace" -icon info -parent . -type ok \
							-message "To close the preview window automatically after a delay,\nset the Cell Playback Duration slider to one second or more.\nTo disable this feature, set it to less than one second."
						}
				$w.fl.mnu.setup.m_setup.preview add checkbutton -variable ::config(trace_skip_small_cells_preview) \
						-label "Skip Preview of short cells ( < 1sec., without still time) ?" -onvalue true -offvalue false
			}
			$w.fl.mnu.setup.m_setup add checkbutton -variable ::tr::hide_gui \
					-label "Hide toolbars in Trace mode ?" -onvalue true -offvalue false
			$w.fl.mnu.setup.m_setup add checkbutton -variable ::tr::watch_always_on_top \
					-label "Watch window on top ?" -onvalue "." -offvalue {} -command \
					{catch {wm transient .trinfo $::tr::watch_always_on_top; if {$::tr::watch_always_on_top==""} {wm deiconify .}}}
			$w.fl.mnu.setup.m_setup add checkbutton -variable ::tr::highlight_log_step \
					-label "Highlight last step in log ?" -onvalue true -offvalue false \
					-command {$::tr::wm_main_window.fr.a.txt tag remove highlight 0.0 end}

		.tracemenu add cascade -label "Setup" -menu $w.fl.mnu.setup.m_setup

		pack [button $w.fl.mnu.quit -image ::tr::bitmapx -width 12 -height 12 -command ::tr::cleanexit] -side right -padx 4
		::balloon::set_balloon $w.fl.mnu.quit "Quit Trace mode"

		if {$::config(use_preview)} {
			checkbutton $w.fl.mnu.pvc -image ::preview::bitmapplay -height 13 \
					-variable ::config(trace_auto_preview) -onvalue true -offvalue false
			pack $w.fl.mnu.pvc -side right -pady 0
			::balloon::set_balloon $w.fl.mnu.pvc "Open Preview automatically\nwhen playing a cell"
		}

		.tracemenu add separator
		.tracemenu add checkbutton -variable ::tr::tracemode -label "Trace mode" -command ::tr::dvd_trace

		####### part 1: player
		pack [frame $w.fl.fp -bd 2 -relief sunken] -padx 2 -pady 4 -fill x

		frame $w.fl.fp.f2
		#pack [label $w.fl.fp.f2.lab -text "Menus:" -anchor w -width 14] -side left
		pack [button $w.fl.fp.f2.m1 -text "Title"   -width 6 -command {::tr::menucall Title}]   -side left -padx 1 -fill both -expand 1
		pack [button $w.fl.fp.f2.m2 -text "Root"    -width 6 -command {::tr::menucall Root}]    -side left -padx 1 -fill both -expand 1
		pack [button $w.fl.fp.f2.m5 -text "SubPic"  -width 6 -command {::tr::menucall SubPic}]  -side left -padx 1 -fill both -expand 1
		pack [button $w.fl.fp.f2.m4 -text "Audio"   -width 6 -command {::tr::menucall Audio}]   -side left -padx 1 -fill both -expand 1
		pack [button $w.fl.fp.f2.m6 -text "Angle"   -width 6 -command {::tr::menucall Angle}]   -side left -padx 1 -fill both -expand 1
		pack [button $w.fl.fp.f2.m3 -text "Chapter" -width 6 -command {::tr::menucall Chapter}] -side left -padx 1 -fill both -expand 1

		pack [button $w.fl.fp.f2.m7 -image ::tr::bitmapup    -height 17 -width 16 -command ::tr::goup_button]         -side left -fill both -expand 1 -padx 1
		::balloon::set_balloon $w.fl.fp.f2.m7 "Return (GoUpPGCN)"
		pack [button $w.fl.fp.f2.b1 -image ::tr::bitmapfirst -height 17 -width 16 -command {::tr::chapter_button -1}] -side left -fill both -expand 1 -padx 1
		::balloon::set_balloon $w.fl.fp.f2.b1 "Prev Chapter (PrevPGCN)"
		pack [button $w.fl.fp.f2.b8 -image ::tr::bitmaplast  -height 17 -width 16 -command {::tr::chapter_button 1}]  -side left -fill both -expand 1 -padx 1
		::balloon::set_balloon $w.fl.fp.f2.b8 "Next Chapter (NextPGCN)"


		frame $w.fl.fp.f3
		set col 0
		foreach item {Title SubPic Audio Angle PTT} {
			label $w.fl.fp.f3.l$col -text $item -pady 0 -width 1 -anchor center
			grid $w.fl.fp.f3.l$col -column $col -row 0 -sticky ew
			entry $w.fl.fp.f3.mb$col -textvariable ::tr::display($item) -width 2 \
					-justify center -state readonly
			grid $w.fl.fp.f3.mb$col -column $col -row 1 -sticky nsew
			grid columnconfigure $w.fl.fp.f3 $col -weight 1
			incr col
		}

		$w.fl.fp.f3.mb1 configure -width 12
		$w.fl.fp.f3.mb2 configure -width 12
		grid columnconfigure $w.fl.fp.f3 1 -weight 4
		grid columnconfigure $w.fl.fp.f3 2 -weight 4

		pack $w.fl.fp.f3 -padx 2 -pady 2 -fill x -expand 1
		pack $w.fl.fp.f2 -padx 1 -pady 3 -fill x

		######### part 2: trace
		pack [frame $w.fl.ft] -fill x

		pack [frame  $w.fl.ft.f1] -fill x -padx 1

		pack [button $w.fl.ft.f1.restart -image ::tr::bitmapeject \
				-command ::tr::reset_trace -width 16] -side left -padx 1 -fill y
		bind $w.fl.ft.f1.restart <3> {::tr::reset_trace true}
		::balloon::set_balloon $w.fl.ft.f1.restart "Insert DVD (Restart Trace)\nRight-click to simulate restarting the\nplayback from the Exit domain."

		button $w.fl.ft.f1.preview -image ::preview::bitmapplay -width 16 -command \
				{::preview::preview2 [lindex $::editor_cmds_from 0] \
						[lindex $::editor_cmds_from 1] $::tr::current(Cell) false false 0}
		if {$::config(use_preview)} {
			pack $w.fl.ft.f1.preview -side left -padx 1 -fill y
			::balloon::set_balloon $w.fl.ft.f1.preview "Preview current cell"
		}
		pack [button $w.fl.ft.f1.bstep -width 10 -text "-> Step" \
				-command {set ::tr::wait_step 1; set ::tr::wait_for_step_pause 0}] \
				-side left -padx 1 -fill x ;#-expand 1
		::balloon::set_balloon $w.fl.ft.f1.bstep "Next single step"
		pack [button $w.fl.ft.f1.bend  -width 10 -text "->> Next PB" \
				-command {set ::tr::break_at_next_playback 1; set ::tr::wait_step 2}] \
				-padx 1 -side left -fill x ;#-expand 1
		::balloon::set_balloon $w.fl.ft.f1.bend "Run to next video playback\nor breakpoint"
		set ::tr::break_at_next_playback 0
		pack [button $w.fl.ft.f1.brun  -width 10 -text "->>> Run " \
				-command {if {$::tr::wait_step == 2} {set ::tr::wait_step 0; set ::tr::break_at_next_playback 0} else {set ::tr::wait_step 2}}] \
				-side left -padx 1 -fill x ;#-expand 1
		::balloon::set_balloon $w.fl.ft.f1.brun "Run to next breakpoint\n(or stop Run mode)"
		bind $w.fl.ft.f1.brun <3> {if {$::tr::delay > 0.0} {set ::tr::delay 0.0} else {set ::tr::delay 5.0}}
		pack [scale  $w.fl.ft.f1.scale -from 0.0 -to 30.0 -resolution 0.1 -bigincrement 1.0 -showvalue false -variable ::tr::delay] -side left -fill x -expand 1
		::balloon::set_balloon $w.fl.ft.f1.scale "Cells playback duration\nfor Run mode, in seconds."
		pack [label  $w.fl.ft.f1.secs -textvariable ::tr::delay -width 3 -anchor e] -side left -padx 0
		trace add variable ::tr::wait_step write ::tr::set_run_button

		bind . <Control-Right> {set ::tr::wait_step 1; set ::tr::wait_for_step_pause 0}
		bind . <Control-Down>  {if {$::tr::wait_step == 2} {set ::tr::wait_step 0; set ::tr::break_at_next_playback 0} else {set ::tr::wait_step 2}}
		bind . <Control-Up>    {set ::tr::break_at_next_playback 1; set ::tr::wait_step 2}
		bind . <Control-Left>  {set ::tr::wait_step 0; set ::tr::break_at_next_playback 0}
		bind . <Escape>        {set ::tr::wait_step 0; set ::tr::break_at_next_playback 0}


		############# Right part: Log window
		pack [frame $w.fr] -side left -fill both -expand 1 -padx 1
		frame $w.fr.a
		pack  $w.fr.a -side top -fill both -expand 1
		text  $w.fr.a.txt -xscrollcommand "$w.fr.a.fxs.xs set" \
				-yscrollcommand "$w.fr.a.ys set" -width 10 -height 4 \
				-font $::config(fixed_width_font) -wrap none
		$w.fr.a.txt tag configure highlight -background [$w.fr.a.txt cget -highlightbackground] \
				-foreground [$w.fr.a.txt cget -highlightcolor] -borderwidth 1
		scrollbar $w.fr.a.ys -command "$w.fr.a.txt yview"
		frame     $w.fr.a.fxs
		scrollbar $w.fr.a.fxs.xs -command "$w.fr.a.txt xview" -orient horizontal
		pack      $w.fr.a.fxs.xs -side left -expand 1 -fill x
		button $w.fr.a.more -image ::tr::bitmapmore -bd 1 \
				-command "if {\[$w.fr.a.txt cget -height\]!=20} {$w.fr.a.txt configure -height 20} else {$w.fr.a.txt configure -height 0}"
		grid  $w.fr.a.more -column 0 -row 1 -sticky nesw
		grid  $w.fr.a.txt -column 1 -row 0 -sticky nesw
		grid  $w.fr.a.ys  -column 0 -row 0 -sticky ns
		grid  $w.fr.a.fxs  -column 1 -row 1 -sticky ew
		grid columnconf $w.fr.a 1 -weight 1
		grid rowconf    $w.fr.a 0 -weight 1

		# search
		frame $w.fr.b
		#pack  $w.fr.b -side bottom -fill x
		frame  $w.fr.b.s -bd 0
		pack   $w.fr.b.s -side left -fill x -expand 1

		checkbutton $w.fr.b.s.cb -bd $::config(min_bd) -text "Break" \
				-variable ::tr::break_at_log_output -pady 0 -padx 0 -command "focus $w.fr.b.s.e"
		pack   $w.fr.b.s.cb -side left
		entry  $w.fr.b.s.e -textvariable ::main::trace_searchstring
		pack   $w.fr.b.s.e -side left -fill x -expand 1
		bind   $w.fr.b.s.e <Return> "::main::searchtext $w.fr -forwards trace"
		bind   $w.fr.b.s.e $::config(popups_button) "tk_popup $w.fr.searcheditmenu \[winfo pointerx .\] \[winfo pointery .\] 3"
		if {! [winfo exists $w.fr.searcheditmenu]} {
			menu $w.fr.searcheditmenu -tearoff 0
			$w.fr.searcheditmenu add command -label "Copy"  -accelerator "Ctrl+C" -command "clipboard clear; clipboard append \$::main::trace_searchstring"
			$w.fr.searcheditmenu add command -label "Paste" -accelerator "Ctrl+V" -command "if \[catch {set ::main::trace_searchstring \[clipboard get\]}\] {bell} else {$w.fr.b.s.e icursor end}"
			$w.fr.searcheditmenu add separator
			$w.fr.searcheditmenu add command -label "Clear" -accelerator "Ctrl+K" -command "set ::main::trace_searchstring {}"
			$w.fr.searcheditmenu add separator
			$w.fr.searcheditmenu add command -label "Close" -command "pack forget $w.fr.b; set ::tr::search_visible 0"
		}
		if {$::tcl_platform(platform) == "windows"} {set minwidth 2} else {set minwidth 1}
		button $w.fr.b.s.b1 -text "<" -width $minwidth -command "::main::searchtext $w.fr -backwards trace" -pady 0
		pack   $w.fr.b.s.b1 -side left -padx 1
		button $w.fr.b.s.b2 -text ">" -width $minwidth -command "::main::searchtext $w.fr -forwards  trace" -pady 0
		pack   $w.fr.b.s.b2 -side left -padx 1
		checkbutton $w.fr.b.s.c1 -bd $::config(min_bd) -text "case" \
				-variable ::main::trace_searchcase -pady 0
		pack   $w.fr.b.s.c1 -side left
		checkbutton $w.fr.b.s.c2 -bd $::config(min_bd) -text "reg expr" \
				-variable ::main::trace_searchregexp -pady 0 \
				-onvalue "-regexp" -offvalue "-exact"
		pack   $w.fr.b.s.c2 -side left

		menu $w.fr.editmenu -tearoff 0
		$w.fr.editmenu add command -label "Show in text viewer" -command "::main::showtext .tracelog {Trace Log} \[string trim \[$w.fr.a.txt get 0.0 end\] \n\]\\n"
		$w.fr.editmenu add command -label "Save" -command "::main::savetext $w.fr"
		$w.fr.editmenu add separator
		$w.fr.editmenu add checkbutton -label "Search"  -variable ::tr::search_visible \
			-command "if \[winfo ismapped $w.fr.b\] {pack forget $w.fr.b} else {pack  $w.fr.b -side bottom -fill x; focus $w.fr.b.s.e}"
		$w.fr.editmenu add separator
		$w.fr.editmenu add command -label "Copy" -accelerator "Ctrl+C" -command "::main::showtext_copy $w.fr"
		$w.fr.editmenu add command -label "Select all" -accelerator "Ctrl+Shift+A" -command "::main::showtext_select_all $w.fr"
		$w.fr.editmenu add separator
		$w.fr.editmenu add command -label "Clear all" -command "::main::showtext_clear_all $w.fr"

		bind $w.fr.a.txt $::config(popups_button) "tk_popup $w.fr.editmenu \[winfo pointerx .\] \[winfo pointery .\] 5"
		bind $w.fr.a.txt <Control-A> "::main::showtext_select_all $w.fr"

		pack $w -fill x -pady 2 -expand 1

		init_trace_config
		build_info_window .trinfo
		configure_info_window
		#wm deiconify $w
		reset_trace
	} else {
		pack $w -fill x -pady 2 -expand 1

		trace_main_window_state disabled
		configure_info_window

		::main::set_listbox_selection [lindex [::main::get_selection_indexes] 0]
		if {[info exists ::tr::snapshot_bank(auto)] && $::tr::snapshot_bank(auto)} {
			restore_state auto
			if {$::tr::num_active_wait_step <= 0} {
				step
			}
		} else {
			reset_trace
		}
	}
	if {$::config(platform) == "mac"} {
		if {$::tr::hide_gui} {
			frame $w.fr.a.fxs.macpad
			pack  $w.fr.a.fxs.macpad -side left -padx 5
		} else {
			destroy $w.fr.a.fxs.macpad
		}
	}
	set ::tr::wait_step 0
	set ::tr::wait_for_step_pause 0
	update idletasks
	#step	;# launch step loop

	return
}
# update labels in State menu
proc update_state_menu_labels {} {
	set w .mf.cf.trace
	if {[winfo exists $w.fl.mnu.state.m_state.auto]} {
		foreach set {1 2 3 4 pos gprm sprm pgc last_pgc domain last_domain pb menu auto} {
			if {[info exists ::tr::bank_pgc_label($set)]} {
				set label($set) ":  $::tr::bank_pgc_label($set)"
			} else {
				set label($set) ""
			}
		}
		$w.fl.mnu.state.m_state.auto entryconfigure "Current PGC*" -label "Current PGC$label(pgc)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Last PGC*" -label "Last PGC$label(last_pgc)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Current Domain*" -label "Current Domain$label(domain)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Last Domain*" -label "Last Domain$label(last_domain)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Last Playback*" -label "Last Playback$label(pb)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Last Menu*" -label "Last Menu$label(menu)"
		$w.fl.mnu.state.m_state.auto entryconfigure "Trace Mode closed*" -label "Trace Mode closed$label(auto)"
		for {set i 1} {$i <= 4} {incr i} {
			$w.fl.mnu.state.m_state.rem entryconfigure "User bank $i*" -label "User bank $i$label($i)"
			$w.fl.mnu.state.m_state.res entryconfigure "User bank $i*" -label "User bank $i$label($i)"
		}
		$w.fl.mnu.state.m_state.rem entryconfigure "Only Position*" -label "Only Position$label(pos)"
		$w.fl.mnu.state.m_state.res entryconfigure "Only Position*" -label "Only Position$label(pos)"
		$w.fl.mnu.state.m_state.rem entryconfigure "Only GPRMs*" -label "Only GPRMs$label(gprm)"
		$w.fl.mnu.state.m_state.res entryconfigure "Only GPRMs*" -label "Only GPRMs$label(gprm)"
		$w.fl.mnu.state.m_state.rem entryconfigure "Only SPRMs*" -label "Only SPRMs$label(sprm)"
		$w.fl.mnu.state.m_state.res entryconfigure "Only SPRMs*" -label "Only SPRMs$label(sprm)"
	}
}
proc cleanexit {} {
	snapshot_state auto
	set ::tr::tracemode 0
	set delay 100
	if {[winfo exists .menubuttons]} {
		set ::menubuttons::menubuttons_done -1
		set delay 600
	}
	after $delay {
		set ::tr::now_playing false
		set ::tr::wait_step -1
		set ::tr::wait_for_step_pause 0
		set ::tr::wait_for_step_pause2 0
		after 200 {
			.mf.f1.sb configure -state normal
			pack forget $::tr::wm_main_window
			wm withdraw .trinfo
			set ::toolbar::toolbar_opened $::tr::toolbar_opened
			if {$::toolbar::toolbar_opened} {
				::toolbar::init_toolbar true
			}
			set ::main::buttons_opened $::tr::buttons_opened
			if {$::main::buttons_opened} {
				::main::action_buttons show
			} else {
				.mf.cf configure -height 1
				update
			}
			if {! [winfo ismapped .mf.st]} {
				pack .mf.st -fill x -side bottom
			}

			catch {destroy .menubuttons}
			catch {image delete ::menubuttons::preview_ppm}
			for {set i 1} {$i <= 32} {incr i} {
				if {[catch {image delete ::menubuttons::preview_but($i)}]} break
			}
			set ::tr::tr_killed 1
		}
	}
}
################################ GUI procs ###########################

proc init_gui_buttons {} {
	set butfirst {
		#define first_width 13
		#define first_height 13
		static unsigned char first_bits[] = {
		   0x1e, 0x08, 0x1e, 0x0c, 0x1e, 0x0e, 0x1e, 0x0f, 0x9e, 0x0f, 0xde, 0x0f, 0xfe,
		   0x0f, 0xde, 0x0f, 0x9e, 0x0f, 0x1e, 0x0f, 0x1e, 0x0e, 0x1e, 0x0c, 0x1e, 0x08};
	}
	image create bitmap ::tr::bitmapfirst -background {} -data $butfirst

	set butlast {
		#define last_width 13
		#define last_height 13
		static unsigned char last_bits[] = {
		   0x02, 0x0f, 0x06, 0x0f, 0x0e, 0x0f, 0x1e, 0x0f, 0x3e, 0x0f, 0x7e, 0x0f, 0xfe,
		   0x0f, 0x7e, 0x0f, 0x3e, 0x0f, 0x1e, 0x0f, 0x0e, 0x0f, 0x06, 0x0f, 0x02, 0x0f};
	}
	image create bitmap ::tr::bitmaplast -background {} -data $butlast

	set butup {
		#define up_width 13
		#define up_height 13
		static unsigned char up_bits[] = {
		   0x00, 0x01,
		   0x80, 0x03,
		   0xC0, 0x07,
		   0xE0, 0x0F,
		   0xF0, 0x1F,
		   0x80, 0x03,
		   0x80, 0x03,
		   0x80, 0x03,
		   0x80, 0x03,
		   0x80, 0x03,
		   0xff, 0x03,
		   0xff, 0x03,
		   0xff, 0x03};
	}
	image create bitmap ::tr::bitmapup -background {} -data $butup

	set buteject {
		#define eject_width 13
		#define eject_height 13
		static unsigned char eject_bits[] = {
		   0x40, 0x00, 0xe0, 0x00, 0xf0, 0x01, 0xf8, 0x03, 0xfc, 0x07, 0xfe, 0x0f, 0xff,
		   0x1f, 0x00, 0x00, 0x00, 0x00, 0xff, 0x1f, 0xff, 0x1f, 0xff, 0x1f, 0xff, 0x1f};
	}
	image create bitmap ::tr::bitmapeject -background {} -data $buteject

	set butx {
		#define x_width 8
		#define x_height 8
		static unsigned char x_bits[] = {
		   0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81};
	}
	image create bitmap ::tr::bitmapx -background {} -data $butx

	set butmore {
		#define more_width 7
		#define more_height 7
		static unsigned char more_bits[] = {
		   0x08, 0x08, 0x08, 0x7F, 0x08, 0x08, 0x08, 0x00};
	}
	image create bitmap ::tr::bitmapmore -background {} -data $butmore
}

proc trace_main_window_state {state} {
	set w $::tr::wm_main_window
	if {$state == "disabled"} {
		set relief "groove"
		set color [$w.fl.fp.f2.b1 cget -disabledforeground]
	} else {
		set relief "raised"
		set color black
	}

	$w.fl.fp.f2.b1 configure -state $state -relief $relief
	$w.fl.fp.f2.b8 configure -state $state -relief $relief

	$w.fl.fp.f2.m1 configure -state $state -relief $relief
	$w.fl.fp.f2.m2 configure -state $state -relief $relief
	$w.fl.fp.f2.m3 configure -state $state -relief $relief
	$w.fl.fp.f2.m4 configure -state $state -relief $relief
	$w.fl.fp.f2.m5 configure -state $state -relief $relief
	$w.fl.fp.f2.m6 configure -state $state -relief $relief
	$w.fl.fp.f2.m7 configure -state $state -relief $relief

	$w.fl.ft.f1.preview configure -state $state

	set col 0
	foreach item {Title PTT Audio SubPic Angle} {
		$w.fl.fp.f3.mb$col configure -foreground $color
		incr col
	}
}

proc trace_main_window_state_puos {puos} {
	set w $::tr::wm_main_window
	set titlepuos 0
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	if {[lindex $::editor_cmds_from 0] == "pgc" && $vts != 0} {
		set ttn $::pgcs([lindex $::editor_cmds_from 1],TTN)
		set idx 8
		while {$idx < [string length $::raw_sectors(0,VMG_TT_SRPT)]} {
			if {[get_int ::raw_sectors(0,VMG_TT_SRPT) [expr $idx+6] 1] == $vts && \
				[get_int ::raw_sectors(0,VMG_TT_SRPT) [expr $idx+7] 1] == $::pgcs([lindex $::editor_cmds_from 1],TTN)} {
				set puos [expr $puos | ([get_int ::raw_sectors(0,VMG_TT_SRPT) $idx 1] & 3)]
				break
			}
			incr idx 12
		}
	}
	trace_main_window_state normal
	if {$puos != 0} {
		set disabledcolor [$w.fl.fp.f2.b1 cget -disabledforeground]
		for {set p 0} {$p < 25} {incr p} {
			if {[expr {round(pow(2,$p)) & $puos}] != 0} {
				set ::tr::puo($p) 1
			} else {
				set ::tr::puo($p) 0
			}
			switch [expr {round(pow(2,$p)) & $puos}] {
				1 {;#puts "  time play or search" ;# (TODO:)}
				2 {$w.fl.fp.f3.mb1 configure -foreground $disabledcolor ;#puts "  PTT play or search"}
				4 {$w.fl.fp.f3.mb0 configure -foreground $disabledcolor ;#puts "  Title Play"}
				8 {;# $w.fl.fp.f1.b4 configure -state disabled ;#puts "  Stop"}
				16 {$w.fl.fp.f2.m7 configure -state disabled ;#puts "  GoUp"}
				32 {$w.fl.fp.f3.mb1 configure -foreground $disabledcolor ;#puts "  time or PTT search"}
				64 {$w.fl.fp.f2.b1 configure -state disabled ;#puts "  TopPG or PrevPG search"}
				128 {$w.fl.fp.f2.b8 configure -state disabled ;#puts "  NextPG search"}
				256 {;# $w.fl.fp.f1.b7 configure -state disabled ;#puts "  ff"}
				512 {;# $w.fl.fp.f1.b2 configure -state disabled ;#puts "  rew"}
				1024 {$w.fl.fp.f2.m1 configure -state disabled ;#puts "  menu Title"}
				2048 {$w.fl.fp.f2.m2 configure -state disabled ;#puts "  menu Root"}
				4096 {$w.fl.fp.f2.m5 configure -state disabled ;#puts "  menu SubPic"}
				8192 {$w.fl.fp.f2.m4 configure -state disabled ;#puts "  menu Audio"}
				16384 {$w.fl.fp.f2.m6 configure -state disabled ;#puts "  manu Angle"}
				32768 {$w.fl.fp.f2.m3 configure -state disabled ;#puts "  menu PTT"}
				65536 {;#puts "  resume" ;# (TODO:!)}
				131072 {;#puts "  button select or activate" ;# (TODO:)}
				262144 {;# $w.fl.fp.f1.b5 configure -state disabled ;#puts "  still off" ;# (TODO:?)}
				524288 {;# $w.fl.fp.f1.b6 configure -state disabled ;#puts "  pause on"}
				1048576 {$w.fl.fp.f3.mb2 configure -foreground $disabledcolor ;#puts "  audio stream change"}
				2097152 {$w.fl.fp.f3.mb3 configure -foreground $disabledcolor ;#puts "  subpic stream change"}
				4194304 {$w.fl.fp.f3.mb4 configure -foreground $disabledcolor ;#puts "  angle change"}
				8388608 {;#puts "  karaoke audio change"}
				16777216 {;#puts "  video presentation mode change"}
			}
		}
	} else {
		for {set p 0} {$p < 25} {incr p} {
			set ::tr::puo($p) 0
		}
	}
}

proc refresh_puos {puos} {
	if {$puos != 0} {
		for {set p 0} {$p < 25} {incr p} {
			if {[expr {round(pow(2,$p)) & $puos}] != 0} {
				set ::tr::puo($p) 1
			} else {
				set ::tr::puo($p) 0
			}
		}
	} else {
		for {set p 0} {$p < 25} {incr p} {
			set ::tr::puo($p) 0
		}
	}
	if {$puos > 0} {
		logprint puo "        PUOs: $puos (0x[format "%08X" $puos])"
	}
	set i 0
	foreach p $::pgced::puos_labels {
		if {[expr {round(pow(2,$i)) & $puos}] > 0} {
			logprint puo_details "              PUO: $p"
		}
		incr i
	}
}

proc build_info_window {w} {
	if {![winfo exists $w]} {
		toplevel $w
		wm withdraw $w
		wm title $w "Watch"
		if {$::config(tk) != "aqua"} {wm transient $w $::tr::watch_always_on_top}
		set closecmd {
			set ::tr::view_gprms 0
			set ::tr::view_sprms 0
			set ::tr::view_puos  0
			set ::tr::view_input 0
			::tr::configure_info_window
		}
		wm protocol $w WM_DELETE_WINDOW $closecmd
		bind $w <Escape> $closecmd

		pack [frame $w.f1] -side left -fill y

		pack [frame $w.f1.top] -fill y -anchor n
		info_window_gprms $w

		pack [frame $w.f1.bot] -side bottom -anchor w
		info_window_input $w

		pack [frame $w.f2] -side left -fill y
		info_window_sprms $w

		pack [frame $w.f3] -side left -fill y
		info_window_puos $w

		set geom $::config(trace_info_win_geometry)
		set idx [string first "+" $geom]
		if {$idx >= 0} {set geom [string range $geom $idx end]}
		if {[catch {wm geometry $w $geom}]} {
			wm geometry $w ""
		}
	}
}
proc configure_info_window {} {
	set w .trinfo
	if {[expr {$::tr::view_gprms + $::tr::view_sprms + $::tr::view_puos + $::tr::view_input}] == 0} {
		wm withdraw $w
	} else {
		info_window_input $w
		info_window_gprms $w
		info_window_sprms $w
		info_window_puos  $w
		if {![winfo ismapped $w]} {
			wm deiconify $w
			focus $::tr::wm_main_window
		}
	}
}
proc info_window_gprms {w} {
	if {$::tr::view_gprms} {
		if {![winfo exists $w.f1.top.top]} {
			pack [frame $w.f1.top.top -bd 2 -relief groove] -padx 2 -pady 2

			pack [frame $w.f1.top.top.fl -bd 1 -relief raised] -fill x -expand 1 -padx 1 -pady 1
			pack [button $w.f1.top.top.fl.lab -text "GPRMs" -bd 0 -padx 0 -pady 0 \
					-command "::tr::toggle ::tr::view_glabs; ::tr::info_window_gprms $w"] \
					-fill x -expand 1 -side left
			::balloon::set_balloon $w.f1.top.top.fl.lab "Show/Hide GPRM Labels"
			pack [button $w.f1.top.top.fl.x -image ::tr::bitmapx -width 8 -height 8 -bd 1 \
					-command "::tr::toggle ::tr::view_gprms; ::tr::configure_info_window"] \
					-side right -anchor ne -padx 1 -pady 1

			pack [frame $w.f1.top.top.f]
			for {set g 0} {$g < 16} {incr g} {
				grid [label $w.f1.top.top.f.l$g -text $g] -column 0 -row $g
				bind $w.f1.top.top.f.l$g <1> "::tr::toggle ::tr::gprmbreak($g)"
				if {$::tr::gprmbreak($g)} {
					$w.f1.top.top.f.l$g configure -foreground red
				}
				grid [button $w.f1.top.top.f.e$g -textvariable ::tr::gprm($g) \
						-width 5 -pady 0 -bd 1 \
						-command "::tr::highlight_gprms_changes gprm $g edit" \
						-foreground $::config(disabledforeground)] -column 1 -row $g
				if {$::tr::gprm_changed($g)>0} {
					$w.f1.top.top.f.e$g configure -foreground black
				}
			}
			grid [frame $w.f1.top.top.f.pad -width 4 -height 2] -column 2 -row 16
		}
		set view_glabs false
		for {set g 0} {$g < 16} {incr g} {
			if {$::gprm_labels($g) != ""} {
				set view_glabs $::tr::view_glabs
				break
			}
		}
		for {set g 0} {$g < 16} {incr g} {
			destroy $w.f1.top.top.f.c$g
			if {$view_glabs} {
				grid [label $w.f1.top.top.f.c$g -text $::gprm_labels($g) -anchor w] \
						-column 2 -row $g -sticky ew
				bind $w.f1.top.top.f.c$g <1> "::utils::set_gprm_labels $g $w"
			}
		}
	} else {
		destroy $w.f1.top.top
		$w.f1.top configure -width 1 -height 1
	}
}
proc info_window_sprms {w} {
	if {$::tr::view_sprms} {
		if {($::tr::view_slabs && ![winfo exists $w.f2.f2.f.d0]) ||
			(!$::tr::view_slabs && [winfo exists $w.f2.f2.f.d0])} {
			destroy $w.f2.f2
		}
		if {![winfo exists $w.f2.f2]} {
			pack [frame $w.f2.f2 -bd 2 -relief groove] -pady 2 -fill y -side left
			pack [frame $w.f2.f2.fl -bd 1 -relief raised] -fill x -expand 1 -padx 1 -pady 1
			pack [button $w.f2.f2.fl.lab -text "SPRMs" -bd 0 -padx 0 -pady 0 \
					-command "::tr::toggle ::tr::view_slabs; ::tr::info_window_sprms $w"] \
					-fill x -expand 1 -side left
			::balloon::set_balloon $w.f2.f2.fl.lab "Show/Hide SPRM Labels"
			pack [button $w.f2.f2.fl.x -image ::tr::bitmapx -width 8 -height 8 -bd 1 \
					-command "::tr::toggle ::tr::view_sprms; ::tr::configure_info_window"] \
					-side right -anchor ne -padx 1 -pady 1

			pack [frame $w.f2.f2.f]
			for {set s 0} {$s < 24} {incr s} {
				grid [label $w.f2.f2.f.l$s -text $s] -column 0 -row $s
				grid [button $w.f2.f2.f.e$s -textvariable ::tr::sprm($s) \
						-width 5 -pady 0 -borderwidth 1 \
						-command "::tr::highlight_sprms_changes sprm $s edit" \
						-foreground $::config(disabledforeground)] -column 1 -row $s
				if {$::tr::sprmbreak($s)} {
					$w.f2.f2.f.l$s configure -foreground red
				}
				if {$::tr::sprm_changed($s)>0} {
					$w.f2.f2.f.e$s configure -foreground black
				}
				bind $w.f2.f2.f.l$s <1> "::tr::toggle ::tr::sprmbreak($s)"
				if {$::tr::view_slabs} {
					grid [label $w.f2.f2.f.d$s -text [lindex $::disasm::SPRMlist $s] -anchor w] -column 2 -row $s -sticky ew
				}
			}
			grid [frame $w.f2.f2.f.pad -width 4 -height 2] -column 2 -row 24
		}
	} else {
		destroy $w.f2.f2
		$w.f2 configure -width 1 -height 1
	}
}
proc toggle {var} {
	if {[info exists $var]} {
		set val [subst \$$var]
		set $var [expr {! $val}]
	} else {
		set $var 1
	}
}
proc info_window_puos {w} {
	if {$::tr::view_puos} {
		if {![winfo exists $w.f3.f3]} {
			pack [frame $w.f3.f3 -bd 2 -relief groove] -padx 2 -pady 2 -side left -fill y
			pack [frame $w.f3.f3.lf] -fill x -expand 1
			pack [label $w.f3.f3.lf.lab -text "PUOs  (from IFOs only)"] -side left -fill x -expand 1
			pack [button $w.f3.f3.lf.x -image ::tr::bitmapx -width 8 -height 8 -bd 1 \
					-command "::tr::toggle ::tr::view_puos; ::tr::configure_info_window"] \
					-side right -anchor ne -padx 1
			pack [frame $w.f3.f3.pad1] -fill y -expand 1
			set puon 0
			foreach p $::pgced::puos_labels {
				checkbutton $w.f3.f3.puo$puon -text "$p" -anchor w -pady 0 -borderwidth 1 \
						-variable ::tr::puo($puon) -indicatoron false -selectcolor pink \
						-state disabled -disabledforeground black -offrelief sunken
				pack $w.f3.f3.puo$puon -fill x -padx 1
				incr puon
			}
			pack [frame $w.f3.f3.pad2] -pady 1
		}
	} else {
		destroy $w.f3.f3
		$w.f3 configure -width 1 -height 1
	}
}
proc info_window_input {w} {
	if {$::tr::view_input} {
		if {![winfo exists $w.f1.bot.bot]} {
			pack [frame $w.f1.bot.bot -bd 2 -relief groove] -padx 2 -pady 2 -fill both
			pack [frame $w.f1.bot.bot.fl] -fill x -expand 1
			pack [label $w.f1.bot.bot.fl.lab -text "Input"] -fill x -expand 1 -side left
			pack [button $w.f1.bot.bot.fl.x -image ::tr::bitmapx -width 8 -height 8 -bd 1 \
					-command "::tr::toggle ::tr::view_input; ::tr::configure_info_window"] \
					-side right -anchor ne -padx 1 -pady 1

			pack [label $w.f1.bot.bot.lab -textvariable ::tr::info_convert_lab -relief groove -width 8 -anchor w] \
					-fill x -expand 1 -padx 1 -pady 1
			bind $w.f1.bot.bot.lab <1> ::tr::dump_info_window_input_to_log
			::balloon::set_balloon $w.f1.bot.bot.lab "Last modified register.\nClick to dump to log."

			pack [entry $w.f1.bot.bot.dec -textvariable ::tr::info_convert_dec -relief groove -width 5 \
					-validate key -validatecommand "::tr::info_validate_entry dec %P %d"] \
					-fill x -expand 1 -padx 1 -pady 1
			bind $w.f1.bot.bot.dec <Return> ::tr::dump_info_window_input_to_log

			pack [frame $w.f1.bot.bot.hex] -fill x -expand 1 -padx 1 -pady 1
			pack [label $w.f1.bot.bot.hex.lab -text "hex" -anchor w] -side left -fill x
			pack [entry $w.f1.bot.bot.hex.hex -textvariable ::tr::info_convert_hex -relief groove -width 4 \
					-validate key -validatecommand "::tr::info_validate_entry hex %P %d"] -ipadx 2 -side right
			bind $w.f1.bot.bot.hex.hex <Return> ::tr::dump_info_window_input_to_log

			pack [frame $w.f1.bot.bot.asc] -fill x -expand 1 -padx 1 -pady 1
			pack [label $w.f1.bot.bot.asc.lab -text "lang." -anchor w] -side left -fill x
			pack [entry $w.f1.bot.bot.asc.asc -textvariable ::tr::info_convert_asc -relief groove -width 2 \
					-validate key -validatecommand "::tr::info_validate_entry asc %P %d"] -ipadx 2 -side right
			bind $w.f1.bot.bot.asc.asc <Return> ::tr::dump_info_window_input_to_log

			pack [frame $w.f1.bot.bot.but] -fill x -expand 1 -padx 1 -pady 1
			pack [label $w.f1.bot.bot.but.lab -text "button" -anchor w] -side left -fill x
			pack [entry $w.f1.bot.bot.but.but -textvariable ::tr::info_convert_but -relief groove -width 2 \
					-validate key -validatecommand "::tr::info_validate_entry but %P %d"] -ipadx 2 -side right
			bind $w.f1.bot.bot.but.but <Return> ::tr::dump_info_window_input_to_log

			pack [label $w.f1.bot.bot.use -textvariable ::tr::info_usage_count -anchor w] -side left -fill x

			#pack [button $w.f1.bot.bot.close -text "Reset" -bd 1 \
			#		-command "if {\$::tr::info_convert_type != {}} {set ::tr::info_convert_dec 0; ::tr::info_validate_entry dec 0 1; ::tr::dump_info_window_input_to_log}"] \
			#		-fill x -padx 2 -pady 2

			if {[info exists ::tr::info_convert_type]} {
				if {$::tr::info_convert_type == "gprm"} {
					$w.f1.bot.bot.lab configure -background yellow
				}
				if {$::tr::info_convert_type == "sprm"} {
					$w.f1.bot.bot.lab configure -background lightgreen
				}
			}
		}
	} else {
		destroy $w.f1.bot.bot
		$w.f1.bot configure -width 1 -height 1
	}
}
proc dump_info_window_input_to_log {} {
	if {$::tr::info_convert_type != ""} {
		reset_log_highlight
		set tmp1 $::tr::log_filter(gprm)
		set tmp2 $::tr::log_filter(sprm)
		set ::tr::log_filter(gprm) 1
		set ::tr::log_filter(sprm) 1
		reset_log_highlight
		logprint always "+++++++ Log: Dump forced by user."
		set [subst ::tr::$::tr::info_convert_type]($::tr::info_convert_reg) $::tr::info_convert_dec
		set ::tr::log_filter(gprm) $tmp1
		set ::tr::log_filter(sprm) $tmp2
		reset_log_highlight
	}
}
proc highlight_gprms_changes {name index op} {
	if {$::tr::view_gprms} {
		for {set g 0} {$g < 16} {incr g} {
			.trinfo.f1.top.top.f.e$g configure -background $::config(systembuttonface)
		}
		.trinfo.f1.top.top.f.e$index configure -background yellow
		if {$op == "write"} {
			.trinfo.f1.top.top.f.e$index configure -foreground black
		}
	}
	if {$op != "edit"} {
		incr ::tr::gprm_changed($index)
	}
	set val $::tr::gprm($index)
	if {$val == ""} {set val 0}
	if {$::tr::view_input} {
		.trinfo.f1.bot.bot.lab configure -background yellow
	}
	set ::tr::info_convert_type "gprm"
	set ::tr::info_convert_reg $index
	set ::tr::info_convert_lab "gprm($index)"
	if {[expr {$val%1024}]==0 && $val!=0} {
		set ::tr::info_convert_but [expr {$val/1024}]
	} else {
		set ::tr::info_convert_but ""
	}
	set langcode [d2ascii $val]
	if {[lsearch $::streams::valid_lang_codes $langcode] >= 0} {
		set ::tr::info_convert_asc [d2ascii $val]
	} else {
		set ::tr::info_convert_asc ""
	}
	set ::tr::info_convert_hex [format {%04X} $val]
	set ::tr::info_convert_dec $val
	set ::tr::info_usage_count "set $::tr::gprm_changed($index) x"
	if {$::tr::gprm_changed($index) >= 100000} {
		set ::tr::info_usage_count $::tr::gprm_changed($index)
	}

	if {$op == "edit"} {
		if {!$::tr::view_input} {
			set ::tr::view_input 1
			info_window_input .trinfo
		}
		focus .trinfo.f1.bot.bot.dec
		.trinfo.f1.bot.bot.dec icursor end
		.trinfo.f1.bot.bot.dec selection range 0 end
	} else {
		set txt "\t-> gprm($index) = $val (0x[format {%04X} $val])"
		if {[::streams::valid_langcode $::tr::info_convert_asc]} {
			set txt "${txt} = \"$::tr::info_convert_asc\""
		}
		set but [expr {$val/1024}]
		if {$but > 0 && $but <= 36 && [expr {$val%1024}] == 0} {
			set txt "${txt} = button [expr {$val/1024}]"
		}
		logprint gprm $txt

		if {[info exists ::tr::gprmbreak($index)] && $::tr::gprmbreak($index)} {
			set ::tr::wait_step 0
			wait_for_step
		}
	}
	if {$::tr::view_gprms} {
		set pause [expr {$::tr::delay / 10}]
		if {$pause > 0.15} {set pause 0.15}
		if {$pause < 0.05} {set pause 0.05}
		pause2 $pause
	}
}
proc highlight_sprms_changes {name index op} {
	if {$::tr::view_sprms} {
		for {set s 0} {$s < 24} {incr s} {
			.trinfo.f2.f2.f.e$s configure -background $::config(systembuttonface)
		}
		.trinfo.f2.f2.f.e$index configure -background lightgreen
		if {$op == "write"} {
			.trinfo.f2.f2.f.e$index configure -foreground black
		}
	}
	if {$op != "edit"} {
		incr ::tr::sprm_changed($index)
	}
	set val $::tr::sprm($index)
	if {$val == ""} {set val 0}
	if {$::tr::view_input} {
		.trinfo.f1.bot.bot.lab configure -background lightgreen
	}
	set ::tr::info_convert_type "sprm"
	set ::tr::info_convert_reg $index
	set ::tr::info_convert_lab "sprm($index)"
	if {[expr {$val%1024}]==0 && $val!=0} {
		set ::tr::info_convert_but [expr {$val/1024}]
	} else {
		set ::tr::info_convert_but ""
	}
	set langcode [d2ascii $val]
	if {[lsearch $::streams::valid_lang_codes $langcode] >= 0} {
		set ::tr::info_convert_asc [d2ascii $val]
	} else {
		set ::tr::info_convert_asc ""
	}
	set ::tr::info_convert_hex [format {%04X} $val]
	set ::tr::info_convert_dec $val
	set ::tr::info_usage_count "set $::tr::sprm_changed($index) x"
	if {$::tr::sprm_changed($index) >= 100000} {
		set ::tr::info_usage_count $::tr::sprm_changed($index)
	}

	if {$op == "edit"} {
		if {!$::tr::view_input} {
			set ::tr::view_input 1
			info_window_input .trinfo
		}
		focus .trinfo.f1.bot.bot.dec
		.trinfo.f1.bot.bot.dec icursor end
		.trinfo.f1.bot.bot.dec selection range 0 end
	} else {
		set txt "\t-> sprm($index:[lindex $::disasm::SPRMlist $index]) = $val (0x[format {%04X} $val])"
		if {[::streams::valid_langcode $::tr::info_convert_asc]} {
			set txt "${txt} = \"$::tr::info_convert_asc\""
		}
		set but [expr {$val/1024}]
		if {$but > 0 && $but <= 36 && [expr {$val%1024}] == 0} {
			set txt "${txt} = button [expr {$val/1024}]"
		}
		logprint sprm $txt

		if {[info exists ::tr::sprmbreak($index)] && $::tr::sprmbreak($index)} {
			set ::tr::wait_step 0
			wait_for_step
		}
	}
	if {$::tr::view_sprms} {
		set pause [expr {$::tr::delay / 10}]
		if {$pause > 0.25} {set pause 0.25}
		if {$pause < 0.05} {set pause 0.05}
		pause2 $pause
	}
}
proc d2ascii {num} {
	set ascii ""
	while {$num > 0} {
		set c [expr {$num % 256}]
		set num [expr {$num / 256}]
		if {$c <= 0X1F || ($c >= 0X80 && $c <= 0X9F)} {
			set ascii ".$ascii"
		} else {
			set ascii "[format {%c} $c]$ascii"
		}
	}
	while {[string length $ascii] < 2} {
		set ascii ".$ascii"
	}
	return $ascii
}
proc ascii2d {ascii} {
	set numchars [binary scan $ascii cc a(1) a(2)]
	set r 0
	for {set d 1} {$d <= $numchars} {incr d} {
		set r [expr {($r * 256) + $a($d)}]
	}
	return $r
}
proc info_validate_entry {entrytype value action} {

	if {$action == -1} {return true}
	if {![winfo exists .trinfo.f1.bot.bot.but.but]} {return true}
	if {$value == ""} {return true}
	if {$::tr::info_convert_type == ""} {bell; return false}
	switch $entrytype {
		"dec" {
			if {[catch {set val [expr {$value}]}]} {bell; return false}
			if {$val > 0xFFFF} {bell; return false}
			set ::tr::info_convert_asc [d2ascii $val]
			set ::tr::info_convert_but [expr {$val/1024}]
			set ::tr::info_convert_hex [format {%04X} $val]
		}
		"hex" {
			if {[catch {set val [x2d $value]}]} {bell; return false}
			if {$val > 0xFFFF} {bell; return false}
			set ::tr::info_convert_asc [d2ascii $val]
			set ::tr::info_convert_but [expr {$val/1024}]
			set ::tr::info_convert_dec $val
		}
		"asc" {
			set numchars [binary scan $value ccc a(1) a(2) a(3)]
			if {$numchars > 2} {bell; return false}
			set val 0
			for {set d 1} {$d <= $numchars} {incr d} {
				set val [expr {($val * 256) + $a($d)}]
			}
			set ::tr::info_convert_but [expr {$val/1024}]
			set ::tr::info_convert_hex [format {%04X} $val]
			set ::tr::info_convert_dec $val
		}
		"but" {
			if {[catch {set val [expr {$value}]}]} {bell; return false}
			if {$val > 36} {bell; return false}
			set val [expr {$value * 1024}]
			set ::tr::info_convert_asc [d2ascii $val]
			set ::tr::info_convert_hex [format {%04X} $val]
			set ::tr::info_convert_dec $val
		}
	}

	if {[lsearch $::streams::valid_lang_codes $::tr::info_convert_asc] == -1} {
		set ::tr::info_convert_asc ""
	}
	if {[expr {$val%1024}]!=0 || $val==0} {
		set ::tr::info_convert_but ""
	}

	trace remove variable ::tr::gprm write ::tr::highlight_gprms_changes
	trace remove variable ::tr::sprm write ::tr::highlight_sprms_changes
	set [subst ::tr::$::tr::info_convert_type]($::tr::info_convert_reg) $val
	trace add variable ::tr::gprm write ::tr::highlight_gprms_changes
	trace add variable ::tr::sprm write ::tr::highlight_sprms_changes

	return true
}

proc player_setup {} {
	set utiltitle "Virtual Player Setup"
	set w .playersetup
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::tr::playersetup_done 0}
	bind $w <Escape> "set ::tr::playersetup_done 0"
	bind $w <Return> "set ::tr::playersetup_done 1"

	### Language Setup
	pack [frame $w.f1 -bd 2 -relief groove -padx 4] -padx 4 -pady 2 -fill x -anchor w
	pack [frame $w.f1.f] -side left
	grid [label $w.f1.f.lab -text "Language Setup:" -pady 4] -sticky w -row 0 -column 0

	grid [button $w.f1.f.iso -text " ISO639 language codes " -command "::main::showtextfile .langcodes \
			{ISO639 language codes} {$::config(apppath)/lib/data/lang_codes.txt} grab" -pady 0] \
			-sticky ew -row 0 -column 3 -padx 2

	grid [label $w.f1.f.l0 -text "Preferred Menu Language code (sprm 0)"] -sticky w -row 1 -column 0
	grid [entry $w.f1.f.e0 -textvariable ::tr::playerconfig(0) -width 2 -validate key \
			-validatecommand {::streams::validate_langcode %P} -font $::config(fixed_width_font)] \
			-sticky w -row 1 -column 1

	grid [label $w.f1.f.l16 -text "Preferred Audio Language code (sprm 16)"] -sticky w -row 2 -column 0
	grid [entry $w.f1.f.e16 -textvariable ::tr::playerconfig(16) -width 2 -validate key \
			-validatecommand {::streams::validate_langcode %P} -font $::config(fixed_width_font)] \
			-sticky w -row 2 -column 1
	grid [label $w.f1.f.l17 -text "and Extension (sprm 17)"] -sticky w -row 2 -column 2
	grid [menubutton $w.f1.f.e17 -textvariable ::tr::playerconfig(17,text) -width 20 \
			-menu $w.f1.f.e17.m -pady 3] -sticky w -row 2 -column 3
	menu $w.f1.f.e17.m -tearoff 0
	set idx 0
	foreach i $::streams::audio_code_ext_list {
		$w.f1.f.e17.m add command -label $i \
				-command "set ::tr::playerconfig(17) $idx"
		incr idx
	}
	trace add variable ::tr::playerconfig(17) write ::tr::player_setup_code17

	grid [label $w.f1.f.l18 -text "Preferred Sub-Picture Language code (sprm 18)"] -sticky w -row 3 -column 0
	grid [entry $w.f1.f.e18 -textvariable ::tr::playerconfig(18) -width 2 -validate key \
			-validatecommand {::streams::validate_langcode %P} -font $::config(fixed_width_font)] \
			-sticky w -row 3 -column 1
	grid [label $w.f1.f.l19 -text "and Extension (sprm 19)"] -sticky w -row 3 -column 2
	grid [menubutton $w.f1.f.e19 -textvariable ::tr::playerconfig(19,text) -width 20 \
			-menu $w.f1.f.e19.m -pady 3] -sticky w -row 3 -column 3
	menu $w.f1.f.e19.m -tearoff 0
	set idx 0
	foreach i $::streams::subpic_code_ext_list {
		if {$i != "-"} {
			$w.f1.f.e19.m add command -label $i \
					-command "set ::tr::playerconfig(19) $idx"
		}
		incr idx
	}
	trace add variable ::tr::playerconfig(19) write ::tr::player_setup_code19
	grid [label $w.f1.f.lnote -text "The language codes (ISO 639) must be TWO lower case letters (like 'en', 'fr'...), or leave the box empty for 'unset'." \
			-justify left] -sticky w -row 4 -column 0 -columnspan 4
	grid rowconfigure $w.f1.f 1 -uniform 1
	grid rowconfigure $w.f1.f 2 -uniform 1
	grid rowconfigure $w.f1.f 3 -uniform 1

	### Video Setup:
	# TODO: Verifier usage du SPRM 14!
	pack [frame $w.f2 -bd 2 -relief groove -padx 4] -padx 4 -pady 2 -fill x
	pack [frame $w.f2.f] -side left
	grid [label $w.f2.f.lab -text "Video Setup:" -pady 4] -sticky w -row 0 -column 0

	grid [label $w.f2.f.l14 -text "TV Display Aspect and Video Preference (sprm 14)"] -sticky w -row 1 -column 0
	grid [radiobutton $w.f2.f.r14_1 -text "4:3 TV - Display 16:9 as Letterboxed" \
			-variable ::tr::playerconfig(14) -value 512] -sticky w -row 1 -column 1
	grid [radiobutton $w.f2.f.r14_2 -text "4:3 TV - Display 16:9 as Pan & Scan" \
			-variable ::tr::playerconfig(14) -value 256] -sticky w -row 2 -column 1
	grid [radiobutton $w.f2.f.r14_3 -text "16:9 TV" \
			-variable ::tr::playerconfig(14) -value 3072] -sticky w -row 3 -column 1

	### Parental Setup:
	pack [frame $w.f3 -bd 2 -relief groove -padx 4] -padx 4 -pady 2 -fill x
	pack [frame $w.f3.f] -side left
	grid [label $w.f3.f.lab -text "Parental Setup:" -pady 4] -sticky w -row 0 -column 0

	grid [label $w.f3.f.l12 -text "Parental Managment Country code (sprm 12)"] -sticky w -row 1 -column 0
	grid [entry $w.f3.f.e12 -textvariable ::tr::playerconfig(12) -width 2 -validate key \
		-validatecommand {::tr::validate_country_code %P} -font $::config(fixed_width_font)] \
		-sticky w -row 1 -column 1

	grid [label $w.f3.f.l13 -text "and Level (sprm 13)"] -sticky w -row 1 -column 2
	grid [menubutton $w.f3.f.e13 -textvariable ::tr::playerconfig(13,text) -width 12 \
			-menu $w.f3.f.e13.m -pady 3] -sticky w -row 1 -column 3
	menu $w.f3.f.e13.m -tearoff 0
	$w.f3.f.e13.m add command -label "Not set (15)" \
			-command "set ::tr::playerconfig(13) 15; set ::tr::playerconfig(13,text) {Not set (15)}"
	for {set i 1} {$i <= 8} {incr i} {
		$w.f3.f.e13.m add command -label $i \
				-command "set ::tr::playerconfig(13) $i; set ::tr::playerconfig(13,text) $i"
	}
	grid [label $w.f3.f.lpasswd -text "The country code (ISO 3166) must be TWO UPPER CASE letters (like 'US', 'UK', 'FR', 'BE'...), or empty for 'unset'.\nIt is not currently used by PgcEdit's Trace function (but the Level is used).\n\nNote: A real player should ask for a password, too." \
			-justify left] -sticky w -row 2 -column 0 -columnspan 100

	### Hardware Capabilities:
	pack [frame $w.f4 -bd 2 -relief groove -padx 4] -padx 4 -pady 2 -fill x
	pack [frame $w.f4.f] -side left
	grid [label $w.f4.f.lab -text "Hardware Capabilities:" -pady 4] -sticky w -row 0 -column 0 -columnspan 5

	grid [label $w.f4.f.l15 -text "Audio Capabilities (sprm 15)"] -sticky w -row 1 -column 0 -columnspan 2
	grid [label $w.f4.f.r15_pad -text "Allow"] -sticky w -row 2 -column 0
	grid [checkbutton $w.f4.f.r15_1 -text "SDDS" \
			-variable ::tr::playerconfig(15,10) -onvalue 0 -offvalue [expr {(0x0400)}]] -sticky w -row 2 -column 1
	grid [checkbutton $w.f4.f.r15_2 -text "DTS" \
			-variable ::tr::playerconfig(15,11) -onvalue 0 -offvalue [expr {(0x0800)}]] -sticky w -row 3 -column 1
	grid [checkbutton $w.f4.f.r15_3 -text "MPEG" \
			-variable ::tr::playerconfig(15,12) -onvalue 0 -offvalue [expr {(0x1000)}]] -sticky w -row 4 -column 1
	grid [checkbutton $w.f4.f.r15_4 -text "Dolby" \
			-variable ::tr::playerconfig(15,14) -onvalue 0 -offvalue [expr {(0x4000)}]] -sticky w -row 5 -column 1

	grid [label $w.f4.f.pad -text "" -width 15] -sticky w -row 2 -column 2

	grid [label $w.f4.f.l20 -text "Player Region Code (sprm 20)"] -sticky w -row 1 -column 3 -columnspan 3
	grid [label $w.f4.f.r20_pad -text "Allow zones"] -sticky w -row 2 -column 3
	for {set i 1} {$i <= 4} {incr i} {
		grid [checkbutton $w.f4.f.r20_$i -text "$i  " -onvalue 0 -offvalue [expr {round(pow(2,($i-1)))}] \
				-variable ::tr::playerconfig(20,$i)] -sticky w -row [expr {$i+1}] -column 4
	}
	for {set i 5} {$i <= 8} {incr i} {
		grid [checkbutton $w.f4.f.r20_$i -text "$i"   -onvalue 0 -offvalue [expr {round(pow(2,($i-1)))}] \
				-variable ::tr::playerconfig(20,$i)] -sticky w -row [expr {$i-3}] -column 5
	}
	grid columnconfigure $w.f4.f 1 -weight 1
	grid columnconfigure $w.f4.f 2 -weight 10
	grid columnconfigure $w.f4.f 4 -weight 0
	grid columnconfigure $w.f4.f 5 -weight 1

	### Compatibility:
	pack [frame $w.f5 -bd 2 -relief groove -padx 4] -padx 4 -pady 2 -fill x
	pack [frame $w.f5.f] -side left
	pack [label $w.f5.f.lab -text "Compatibility:" -pady 4] -anchor w
	pack [checkbutton $w.f5.f.sprm7 -justify left -text \
			"SPRM 7 holds the PGN (Program Number) instead of the PTTN (Chapter Number)" \
			-variable ::tr::playerconfig(sprm7holdsPG) -onvalue true -offvalue false] -anchor w
	pack [label $w.f5.f.sprm7l -justify left -text \
			"The SPRM 7 must normally hold the PTTN, but on many players, it holds the PGN!\nYou may download a little test-DVD on the PgcEdit's homepage to test your player." \
			] -anchor w -padx 20

	### buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	button $w.fb.load -text "Load defaults" -width 13 -command {::tr::init_trace_config; ::tr::configure_trace_config_window}
	pack   $w.fb.load -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::tr::playersetup_done 0}
	pack   $w.fb.cancel -side right -padx 2
	button $w.fb.use -text "Use" -width 10 -command {set ::tr::playersetup_done 1}
	pack   $w.fb.use -side right -padx 2
	button $w.fb.save -text "Save" -width 10 -command {set ::tr::playersetup_done 2}
	pack   $w.fb.save -side right -padx 2
	if {$::config(platform) == "mac"} {
		pack [frame $w.fb.macpad] -side left -padx 6
	}

	configure_trace_config_window

	update
	wm geometry $w +150+50
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::tr::playersetup_done
	grab release $w
	destroy $w

	if {$::tr::playersetup_done > 0} {
		set ::tr::playerconfig(15) [expr {$::tr::playerconfig(15,10) | $::tr::playerconfig(15,11) \
				| $::tr::playerconfig(15,12) | $::tr::playerconfig(15,14)}]

		set ::tr::playerconfig(20) 0
		for {set c 1} {$c <= 8} {incr c} {
			set ::tr::playerconfig(20) [expr {$::tr::playerconfig(20) + $::tr::playerconfig(20,$c)}]
		}

		reset_log_highlight
		logprint always "+++++++ Player config changed. Initializing the SPRMs to default/setup values."
		playerconfig2sprms
		reset_log_highlight

		# Save config
		if {$::tr::playersetup_done == 2} {
			set txt "# PgcEdit Virtual Player config file.  Saved automatically.  Do not edit!\n\n"
			set txt "${txt}set ::tr::playerconfig(0) \"$::tr::playerconfig(0)\"\n"
			for {set s 12} {$s <= 20} {incr s} {
				set txt "${txt}set ::tr::playerconfig($s) \"$::tr::playerconfig($s)\"\n"
			}
			set txt "${txt}set ::tr::playerconfig(sprm7holdsPG) \"$::tr::playerconfig(sprm7holdsPG)\"\n"
			saveFile [file join $::config(config_directory) "pgcedit.player.cfg"] auto $txt
		}
	}
}
proc init_trace_config {} {
	set ::tr::playerconfig(0)  ""
	set ::tr::playerconfig(12) ""
	set ::tr::playerconfig(13) 15
	set ::tr::playerconfig(14) 512
	set ::tr::playerconfig(15) 0
	set ::tr::playerconfig(16) ""
	set ::tr::playerconfig(17) 1
	set ::tr::playerconfig(18) ""
	set ::tr::playerconfig(19) 1
	set ::tr::playerconfig(20) 0
	set ::tr::playerconfig(sprm7holdsPG) false
	# Overwrite default values with user defined ones
	set configfile [file join $::config(config_directory) "pgcedit.player.cfg"]
	if {[file exists $configfile]} {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}
	set ::tr::num_active_wait_step 0
}
proc configure_trace_config_window {} {
	# init GUI control variables
	set ::tr::playerconfig(15,10) [expr {$::tr::playerconfig(15) & 0x0400}]
	set ::tr::playerconfig(15,11) [expr {$::tr::playerconfig(15) & 0x0800}]
	set ::tr::playerconfig(15,12) [expr {$::tr::playerconfig(15) & 0x1000}]
	set ::tr::playerconfig(15,14) [expr {$::tr::playerconfig(15) & 0x4000}]
	for {set i 1} {$i <= 8} {incr i} {
		set ::tr::playerconfig(20,$i) [expr {$::tr::playerconfig(20) & round(pow(2,$i-1))}]
	}
	if {$::tr::playerconfig(13) == 15} {
		set ::tr::playerconfig(13,text) {Not set}
	} else {
		set ::tr::playerconfig(13,text) $::tr::playerconfig(13)
	}
	# use the Tcl 'trace' on variables to initialize the menubuttons displays
	set ::tr::playerconfig(17) $::tr::playerconfig(17)
	set ::tr::playerconfig(19) $::tr::playerconfig(19)
}
proc playerconfig2sprms {} {
	trace remove variable ::tr::sprm write ::tr::highlight_sprms_changes
	if {[winfo exists .trinfo.f2.f2.f.e0] && $::tr::view_sprms} {
		for {set s 0} {$s < 24} {incr s} {
			.trinfo.f2.f2.f.e$s configure -background lightgreen \
					-foreground $::config(disabledforeground)
		}
	}

	if {$::tr::playerconfig(0) == ""} {
		set ::tr::sprm(0) [expr {(0xFFFF)}]
	} else {
		set ::tr::sprm(0)  [ascii2d $::tr::playerconfig(0)]		;# Preferred menu language
	}
	set ::tr::sprm(1)  15	;# Audio: none
	set ::tr::sprm(2)  62	;# Subpic: none
	set ::tr::sprm(3)  1	;# Angle
	set ::tr::sprm(4)  1	;# Title
	set ::tr::sprm(5)  1	;# TTN
	set ::tr::sprm(6)  0	;# PGC number
	set ::tr::sprm(7)  1	;# PTT number
	set ::tr::sprm(8)  1024	;# HL_BTN: assumes button 1 to begin
	set ::tr::sprm(9)  0	;# NAV Timer
	set ::tr::sprm(10) 0	;# Jump for NAV
	set ::tr::sprm(11) 0	;# Karaoke mix mode
	set ::tr::sprm(12) [ascii2d $::tr::playerconfig(12)]	;# Parental country code
	set ::tr::sprm(13) $::tr::playerconfig(13)				;# Parental level
	# 4:3 letterbox/pan&scan mode is set in real time
	set ::tr::sprm(14) [expr {$::tr::playerconfig(14) & 0x0C00}]	;# Video prefs & mode
	set ::tr::sprm(15) $::tr::playerconfig(15)				;# Player audio capabilities
	if {$::tr::playerconfig(16) == ""} {
		set ::tr::sprm(16) [expr {(0xFFFF)}]
	} else {
		set ::tr::sprm(16) [ascii2d $::tr::playerconfig(16)]	;# Pref audio
	}
	set ::tr::sprm(17) $::tr::playerconfig(17)				;# Pref audio ext
	if {$::tr::playerconfig(18) == ""} {
		set ::tr::sprm(18) [expr {(0xFFFF)}]
	} else {
		set ::tr::sprm(18) [ascii2d $::tr::playerconfig(18)]	;# Pref subpic
	}
	set ::tr::sprm(19) $::tr::playerconfig(19)				;# Pref subpic ext
	set ::tr::sprm(20) [expr 255 - $::tr::playerconfig(20)]	;# Region code
	set ::tr::sprm(21) 0	;# reserved
	set ::tr::sprm(22) 0	;# reserved
	set ::tr::sprm(23) 0	;# reserved

	if {[winfo exists .trinfo.f2.f2.f.e0] && $::tr::view_sprms} {
		pause2 0.2
		for {set s 0} {$s < 24} {incr s} {
			.trinfo.f2.f2.f.e$s configure -background $::config(systembuttonface)
		}
	}
	trace add variable ::tr::sprm write ::tr::highlight_sprms_changes
}
proc player_setup_code17 {name index op} {
	set ::tr::playerconfig(17,text) [lindex $::streams::audio_code_ext_list $::tr::playerconfig(17)]
}
proc player_setup_code19 {name index op} {
	set ::tr::playerconfig(19,text) [lindex $::streams::subpic_code_ext_list $::tr::playerconfig(19)]
}
proc validate_country_code {value} {
	if {$value == ""} {return true}
	if {! [string is alpha $value]} {bell; return false}
	if {! [string is upper $value]} {bell; return false}
	if {[string length $value] > 2} {bell; return false}
	return true
}
proc logprint {category txt} {
	set logit 1
	if {$category != "always"} {
		set logit $::tr::log_filter($category)
		if {$category == "if" && ! $::tr::log_filter(cmd)} {
			set logit 0
		}
	}
	if {! $logit} return
	if {! [info exists ::tr::wm_main_window]} return
	set w $::tr::wm_main_window.fr.a.txt
	$w insert end $txt
	$w yview moveto 1
	update	;# Don't display the last blank line.
	$w insert end "\n"

	# highlight last step messages by selecting the lines
	if {$::tr::highlight_log_step} {
		$w tag remove highlight 0.0 end
		$w tag add highlight $::tr::trace_highlight_startindex end
	}

	if {$::tr::break_at_log_output && $::main::trace_searchstring != ""} {
		if {$::main::trace_searchregexp == "-exact"} {
			if {$::main::trace_searchcase} {
				set idx [string first $::main::trace_searchstring $txt]
			} else {
				set idx [string first [string toupper $::main::trace_searchstring] [string toupper $txt]]
			}
			if {$idx >= 0} {set found true} else {set found false}
		} else {
			if {$::main::trace_searchcase} {
				set rc [catch {set found [regexp -- $::main::trace_searchstring $txt]} errstr]
			} else {
				set rc [catch {set found [regexp -nocase -- $::main::trace_searchstring $txt]} errstr]
			}
			if {$rc != 0} {
				tk_messageBox -title "Trace: Break at Log Output" -icon error \
						-message "Break at Log Output: regular expression error.\n\n[string map {": " ":\n"} $errstr]" -parent .
				set found true
			}
		}
		if {$found} {
			set ::tr::wait_step 0
		}

	}
}
proc reset_log_highlight {} {
	if {[info exists ::tr::wm_main_window] && [winfo exists $::tr::wm_main_window.fr.a.txt]} {
		set ::tr::trace_highlight_startindex [$::tr::wm_main_window.fr.a.txt index "end -1 lines"]
	}
}
proc set_run_button {name index op} {
	if {$::tr::wait_step == 2} {
		catch {$::tr::wm_main_window.fl.ft.f1.brun configure -text {->|  Break} -foreground red}
	} else {
		set ::tr::wait_for_step_pause 0
		catch {$::tr::wm_main_window.fl.ft.f1.brun configure -text "->>> Run" -foreground black}
	}
	update idletasks
}
proc highlight_gprm_break_changes {name index op} {
	if {$::tr::view_gprms} {
		if {$::tr::gprmbreak($index)} {
			.trinfo.f1.top.top.f.l$index configure -foreground red
		} else {
			.trinfo.f1.top.top.f.l$index configure -foreground black
		}
	}
}
proc highlight_sprm_break_changes {name index op} {
	if {$::tr::view_sprms} {
		if {$::tr::sprmbreak($index)} {
			.trinfo.f2.f2.f.l$index configure -foreground red
		} else {
			.trinfo.f2.f2.f.l$index configure -foreground black
		}
	}
}

################################### Trace procs ###################################

##############################
# Reset GPRMs and SPRMs, and go to FP-PGC, select *** pre label
proc reset_trace {{exit2play false}} {
	update
	set ::tr::trace_highlight_startindex 0.0

	set ::tr::disasmbled ""
	# exit wait for step state
	set ::tr::wait_for_step_pause 0
	set ::tr::wait_for_step_pause2 0
	set ::tr::wait_step -2
	set ::tr::now_playing false
	set ::tr::last_domain "fp 0"

	set ::menubuttons::menubuttons_done -1
	set ::buttons::default_button 1
	catch {destroy .menubuttons}
	catch {image delete ::menubuttons::preview_ppm}
	for {set i 1} {$i <= 32} {incr i} {
		if {[catch {image delete ::menubuttons::preview_but($i)}]} break
	}

	update idletasks

	# clear log
	if {[winfo exists $::tr::wm_main_window.fr.a.txt]} {
		$::tr::wm_main_window.fr.a.txt delete 1.0 end
	}

	if {$exit2play} {
		if {[info exists ::pgcs(Title,1)]} {
			#blocking_msg "Restarting DVD..."
			set header $::pgcs([lindex $::pgcs(Title,1) 0],[lindex $::pgcs(Title,1) 1],header)
			trace_main_window_state_puos [get_int header 8 4]
			set loghead "Exit domain"
		} else {
			bell
			return
		}
	} else {
		#blocking_msg "Inserting DVD..."
		::main::display_pgc pgc 0 0
		trace_main_window_state_puos [get_int ::pgcs(0,0,header) 8 4]
		set loghead "DVD inserted"
	}

	#logprint "****************************** PgcEdit Trace Log *******************************"
	logprint always "$loghead: \"[file nativename $::dvddir]\""
	logprint always "$loghead: \"[::utils::get_General_Name]\"; Provider: [::utils::get_provider_ID]"

	clear_changed_flags
	::main::init_tree_menu_no_move

	# reset GPRMs
	logprint always "$loghead: Resetting all GPRMs to 0."
	trace remove variable ::tr::gprm write ::tr::highlight_gprms_changes
	if {$::tr::view_gprms} {
		pause2 0.2
		for {set g 0} {$g < 16} {incr g} {
			.trinfo.f1.top.top.f.e$g configure -background yellow \
					-foreground $::config(disabledforeground)
		}
	}
	for {set g 0} {$g < 16} {incr g} {
		set ::tr::gprm($g) 0
	}
	if {$::tr::view_gprms} {
		pause2 0.2
		for {set g 0} {$g < 16} {incr g} {
			.trinfo.f1.top.top.f.e$g configure -background $::config(systembuttonface)
		}
	}
	trace add variable ::tr::gprm write ::tr::highlight_gprms_changes

	# reset SPRMs
	if {! $exit2play} {
		logprint always "$loghead: Initializing the SPRMs to default/setup values."
		playerconfig2sprms
	}

	# cosmetics
	set ::tr::info_convert_type ""
	set ::tr::info_convert_reg ""
	set ::tr::info_convert_lab ""
	set ::tr::info_convert_dec ""
	set ::tr::info_convert_hex ""
	set ::tr::info_convert_asc ""
	set ::tr::info_convert_but ""
	set ::tr::info_usage_count ""
	if {$::tr::view_input} {
		.trinfo.f1.bot.bot.lab configure -background $::config(systembuttonface)
	}

	# reset internal variables
	set ::tr::current(Title)  -1
	set ::tr::current(TTN)    -1
	set ::tr::current(LU)     -1
	set ::tr::current(PGC)    -1
	set ::tr::current(PTT)    -1
	set ::tr::current(PG)     -1
	set ::tr::current(Cell)   -1
	set ::tr::last_visited_vts 1
	set ::tr::now_playing false

	unset -nocomplain ::tr::rsm_vts ::tr::rsm_pgc ::tr::rsm_cell

	set ::tr::display(Title)  ""
	set ::tr::display(PTT)    ""
	set ::tr::display(Audio)  ""
	set ::tr::display(SubPic) ""
	set ::tr::display(Angle)  ""

	# Jump to FP PGC
	trace_main_window_state disabled
	if {! $exit2play} {
		display_VTST_PGC 0 0 0
	} else {
		if {[info exists ::pgcs(Title,1)]} {
			set targetvts [lindex $::pgcs(Title,1) 0]
			set targetpgc [lindex $::pgcs(Title,1) 1]
			if {[info exists ::pgcs($targetvts,$targetpgc)]} {
				logprint always "$loghead: Beginning navigation at Title 1"
				display_VTST_PGC $targetvts $targetpgc 1
			} else {
				bell
				logprint always "####### $loghead: ERROR: Cannot find Title 1!"
				display_VTST_PGC 0 0 0
			}
		}
	}

	# check region code
	set code [get_int ::raw_sectors(0,VMGM_MAT) 0X23 1]
	set prohibed true
	for {set c 0} {$c < 8} {incr c} {
		if {[expr {$code & round(pow(2,$c))}] == 0} {
			if {[expr {$::tr::sprm(20) & round(pow(2,$c))}] > 0} {
				set prohibed false
			}
		}
	}
	if {$prohibed} {
		logprint always "####### WARNING: Region Code prohibition!"
		set ::tr::wait_step 0
		tk_messageBox -title "Trace" -icon warning -message "WARNING: Region Code prohibition!"
	}

	set ::tr::abort_current_step 1
	init_pgc_playback
	#close_blocking_msg
	if {$::tr::num_active_wait_step <= 0} {
		step
	}
	return
}

proc exit2play {} {
	set txt "This function restarts the trace from the Exit domain."
	set txt "$txt\n"
	set txt "$txt\nWhen an Exit command is encountered, or when the playback cannot"
	set txt "$txt\ncontinue, or if the user has pressed twice the STOP button on the"
	set txt "$txt\n remote, the player enters the Exit domain. The playback is stopped."

	set txt "$txt\n\nAt this time, if the user presses Play again, the GPRMs are celared,"
	set txt "$txt\nand the DVD is restarted, but instead of beginning the navigation"
	set txt "$txt\nat the First-Play PGC, Title 1 is played."
	set txt "$txt\nThis is what this function simulates."

	set txt "$txt\n\nNote: This function is also available by right clicking on the \"Insert DVD\""
	set txt "$txt\nicon in the bottom left corner of the trace control panel, but this message"
	set txt "$txt\nis not displayed."
	tk_messageBox -message $txt -title "Restart playback from the Exit domain" -parent . \
			-skip "Restart playback from the Exit domain: usage"

	reset_trace true
}

###############################
# Execute current selected command (or label), and
# display the next command (or label) to be executed
proc step {} {
	while {$::tr::wait_step != -1} {
		if {$::tr::wait_step > 0} {
			set idx [lindex [::main::get_selection_indexes] 0]
			if {$idx == "-"} {
				::main::set_listbox_selection end
				set idx [lindex [::main::get_selection_indexes] 0]
				bell
			} elseif {$::tr::now_playing} {
				simulate_play
			} else {
				display_current_line
				if {![exec_current_line]} {return false}
			}
		}
		wait_for_step
	}
	return true
}
proc wait_for_step {} {
	update idletasks
	#logprint always " "
	reset_log_highlight
	if {$::tr::wait_step <= 1} {
		incr ::tr::num_active_wait_step
		tkwait variable ::tr::wait_step
		incr ::tr::num_active_wait_step -1
	} else {
		if {$::tr::now_playing} {
			if {[info exists ::tr::no_pause]} {
				unset ::tr::no_pause
				pause 0.0
			} else {
				pause $::tr::delay
			}
			if {$::config(use_preview) && $::config(trace_auto_preview)} {
				::preview::close_preview
			}
		} else {
			pause 0.0
		}
	}
}
proc pause {secs} {
	update idletasks
	set ::tr::wait_for_step_pause 0
	after [expr {round($secs * 1000)}] {set ::tr::wait_for_step_pause 1}
	tkwait variable ::tr::wait_for_step_pause
}
proc pause2 {secs} {
	update idletasks
	set ::tr::wait_for_step_pause2 0
	after [expr {round($secs * 1000)}] {set ::tr::wait_for_step_pause2 1}
	tkwait variable ::tr::wait_for_step_pause2
}

###############################
# Display the current selected command (or label)
# If applicable, also display the current comparison result
proc display_current_line {} {
	set ::tr::wait_step $::tr::wait_step

	set ::tr::pgclabel [$::widget(main_cmd_lab) cget -text]
	#wm title $::tr::wm_main_window "Trace - $::tr::pgclabel"

	set idx [lindex [::main::get_selection_indexes] 0]
	if {$idx == "-"} {
		reset_log_highlight
		logprint always "####### Unexpected error!  Trace end!"
		reset_log_highlight
		bell
		set ::tr::wait_step 0
		return false
	}
	set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set cmdtype [get_cmd_type cmds $idx]
	set vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vts > 0} {
		set ::tr::last_visited_vts $vts
	}


	switch $cmdtype {
		"* pre" {
			if {[info exists ::tr::breakpoints([lindex $::editor_cmds_from 0],[lindex $::editor_cmds_from 1],prepost)]} {
				set ::tr::wait_step 0
			}
			if {$::tr::break_at_all_titles && [lindex $::editor_cmds_from 0]=="pgc"} {
				set ::tr::wait_step 0
			}
			if {$::tr::break_at_all_menus && [lindex $::editor_cmds_from 0]=="menu"} {
				set ::tr::wait_step 0
			}
			logprint always "------> Entering $::tr::pgclabel"

			set titlepuos 0
			if {[lindex $::editor_cmds_from 0] == "pgc" && $vts != 0} {
				set ttn $::pgcs([lindex $::editor_cmds_from 1],TTN)
				set idx 8
				while {$idx < [string length $::raw_sectors(0,VMG_TT_SRPT)]} {
					if {[get_int ::raw_sectors(0,VMG_TT_SRPT) [expr $idx+6] 1] == $vts && \
						[get_int ::raw_sectors(0,VMG_TT_SRPT) [expr $idx+7] 1] == $::pgcs([lindex $::editor_cmds_from 1],TTN)} {
						set titlepuos [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) $idx 1] & 3]
						break
					}
					incr idx 12
				}
			}
			set header [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1],header)]
			refresh_puos [expr [get_int header 8 4] | $titlepuos]

			trace_main_window_state disabled
		}
		"* post" {
			if {[subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1],numprograms)] == 0} {
				set ::tr::now_playing true
				set ::tr::no_pause true
			} else {
				logprint always "******* Start playback $::tr::pgclabel..."
				set header [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1],header)]
				set tmp [get_int header 0xA2 1]
				if {$tmp > 0} {
					set n [expr {($tmp & 0x7F)}]
					if {[expr {$tmp & 0x80}] > 0} {
						logprint always "        Shuffle mode: playing $n programs."
					} else {
						logprint always "        Random mode: playing $n programs."
					}
				}
				set ::tr::now_playing true
				set ::tr::no_pause true
				# when entering playback mode, reset the collection of PGs already played in random/shuffle mode
				set ::tr::cur_pg_pb_rand_pgs {}
				if {$::tr::break_at_next_playback} {
					set ::tr::break_at_next_playback 0
					set ::tr::wait_step 0
				}
			}
			if {[info exists ::tr::breakpoints([lindex $::editor_cmds_from 0],[lindex $::editor_cmds_from 1],playback)]} {
				set ::tr::wait_step 0
			}
		}
		"* cell" {
			set type [lindex $::editor_cmds_from 0]
			if {$type == "pgc"} {
				set nextpgcn [get_int ::pgcs([lindex $::editor_cmds_from 1],header) 0x9C 2]
				if {$nextpgcn != 0} {
					logprint always "        End of post commands: Jumping to NextPGCN $nextpgcn"
					set err [display_VTST_PGC $vts $nextpgcn 0]
				} else {
					set err -1
				}
			} else {
				set nextpgcn [get_int ::menus([lindex $::editor_cmds_from 1],header) 0x9C 2]
				if {$nextpgcn != 0} {
					logprint always "        End of post commands: Jumping to NextPGCN $nextpgcn"
					set err [display_menu_PGC $vts $nextpgcn]
				} else {
					set err -1
				}
			}
			if {$err == 0} {
				set ::tr::abort_current_step 1
				trace_main_window_state disabled
				init_pgc_playback
				display_current_line
				return true
			} else {
				if {$err != -1} {
					logprint always "####### Target not found!"
				}
				logprint always "------> End of post commands: Exiting!  DVD Playback end!"
				trace_main_window_state disabled
				bell
				set ::tr::wait_step 0
				reset_log_highlight
				return false
			}
		}
		default {
			set ::tr::disasmbled "  -- $cmdtype [get_cmd_linenum cmds $idx] : [::disasm::disasm [lindex $cmds $idx]]"
			logprint cmd $::tr::disasmbled
			check_condition true true
		}
	}
	update idletasks
	return true
}

###############################
# Execute the current selected command (or label)
# and change the current selection accordingly
proc exec_current_line {} {
	set idx [lindex [::main::get_selection_indexes] 0]
	if {$idx == "-"} {
		logprint always "####### Unexpected error!  Trace end!"
		set ::tr::wait_step 0
		bell
		reset_log_highlight
		return false
	}
	if {[catch {set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]}]} {
		set ::tr::wait_step 0
		reset_log_highlight
		return false
	}

	set cmdtype [get_cmd_type cmds $idx]
	switch $cmdtype {
		"* pre" {
			incr idx
			::main::set_listbox_selection $idx
		}
		"* post" {
			::main::set_listbox_selection $idx
		}
		"* cell" {
			bell
			return false
		}
		default {
			if {[string first "ILLEGAL" $::tr::disasmbled]!= -1 ||
					[$::widget(main_listbox) itemcget $idx -background] == "pink"} {
				set txt "####### WARNING: Illegal command found!\nThis command will be skipped!"
				logprint always [string map {\n " "} $txt]
				set ::tr::wait_step 0
				tk_messageBox -title "Trace" -icon warning -message $txt
				if {$cmdtype != "cell"} {
					incr idx
					::main::set_listbox_selection $idx
				} else {
					# return to playback if it's a cell command
					set ::tr::now_playing true
					incr ::tr::current(Cell)
					::main::set_listbox_selection [expr {[llength [get_type_cmds pre cmds]]+1}]
				}
			} elseif {![check_condition false true]} {
				if {$cmdtype != "cell"} {
					incr idx
					::main::set_listbox_selection $idx
				} else {
					# return to playback if it's a cell command
					set ::tr::now_playing true
					incr ::tr::current(Cell)
					::main::set_listbox_selection [expr {[llength [get_type_cmds pre cmds]]+1}]
				}
			} else {
				trace_disasm
			}
		}
	}
	return true
}

proc menucall {menubutton} {
	if {! $::tr::now_playing} {return}

	reset_log_highlight
	logprint always "!!!!!!! USER: $menubutton Menu Button"

	set cur_type [lindex $::editor_cmds_from 0]
	set index    [lindex $::editor_cmds_from 1]
	set cur_vts  [lindex [split $index ","] 0]

	set menutype 0
	set jumpfound false

	if {$cur_type == "pgc"} {

		# remember resume point
		set ::tr::rsm_vts $cur_vts
		set ::tr::rsm_pgc [lindex [split $index ","] 1]
		set ::tr::rsm_cell [expr {$::tr::current(Cell) -1}]
		if {$::tr::rsm_cell < 1} {set ::tr::rsm_cell 1}

		switch $menubutton {
			"Title" {
				# simulate a CallSS-VMGM-menu
				if {[info exists ::menus(0,1,numpgcs)]} {
					for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
						if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
							logprint always "!!!!!!! USER: Jumping to VMGM Title menu."
							display_menu_PGC 0 $pgc
							trace_main_window_state disabled
							init_pgc_playback
							set ::tr::wait_step 1
							set ::tr::wait_for_step_pause 0
							set jumpfound true
							break
						}
					}
				}
				if {! $jumpfound} {
					logprint always "!!!!!!! USER: Target menu not found.  No action."
				}
			}
			"Root"		{set menutype 3}
			"SubPic"	{set menutype 4}
			"Audio"		{set menutype 5}
			"Angle"		{set menutype 6}
			"Chapter"	{set menutype 7}
		}
	}

	if {$cur_type == "menu"} {
		if {$menubutton == "Title"} {
			if {$cur_vts == 0} {
				# Simulate a RSM, if possible.
				logprint always "!!!!!!! USER: Resuming to Title domain."
				if {[info exists ::tr::rsm_vts]} {
					# close menu button window if needed
					set ::menubuttons::menubuttons_done -1

					display_VTST_PGC $::tr::rsm_vts $::tr::rsm_pgc 0
					init_pgc_playback
					set cmds $::pgcs($::tr::rsm_vts,$::tr::rsm_pgc)
					set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
					set nextCell $::tr::rsm_cell
					set ::tr::current(Cell) $nextCell
					::main::set_listbox_selection $nexteditorline
					set ::tr::now_playing true
					set jumpfound true
				} else {
					bell
					logprint always "####### WARNING: Resume point not yet defined!  User action ignored!"
				}
			} else {
				# Jump to Title menu
				if {[info exists ::menus(0,1,numpgcs)]} {
					for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
						if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
							logprint always "!!!!!!! USER: Jumping to VMGM Title menu."
							set ::menubuttons::menubuttons_done -1
							if {![display_menu_PGC 0 $pgc]} {
								trace_main_window_state disabled
								init_pgc_playback
								set ::tr::wait_step 1
								set ::tr::wait_for_step_pause 0
								set jumpfound true
							}
							break
						}
					}
				}
			}
		} else {
			switch $menubutton {
				"Root" {
					# Simulate a RSM, if possible.
					logprint always "!!!!!!! USER: Resuming to Title domain."
					if {[info exists ::tr::rsm_vts]} {
						# close menu button window if needed
						set ::menubuttons::menubuttons_done -1

						display_VTST_PGC $::tr::rsm_vts $::tr::rsm_pgc 0
						init_pgc_playback
						set cmds $::pgcs($::tr::rsm_vts,$::tr::rsm_pgc)
						set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
						set nextCell $::tr::rsm_cell
						set ::tr::current(Cell) $nextCell
						::main::set_listbox_selection $nexteditorline
						set ::tr::now_playing true
						set jumpfound true
					} else {
						bell
						logprint always "####### WARNING: Resume point not yet defined!  User action ignored!"
					}
				}
				"SubPic"	{set menutype 4}
				"Audio"		{set menutype 5}
				"Angle"		{set menutype 6}
				"Chapter"	{set menutype 7}
			}
		}
	}
	if {$menutype > 0} {
		# simulate a CallSS-VTSM
		if {[info exists ::menus($cur_vts,1,numpgcs)]} {
			for {set pgc 1} {$pgc <= $::menus($cur_vts,1,numpgcs)} {incr pgc} {
				if {[expr {$::menus($cur_vts,1,$pgc,mask) & 0x7F}] == $menutype} {
					logprint always "!!!!!!! USER: Jumping to VTSM $menubutton menu."
					set ::menubuttons::menubuttons_done -1
					if {![display_menu_PGC $cur_vts $pgc]} {
						trace_main_window_state disabled
						init_pgc_playback
						set ::tr::wait_step 1
						set ::tr::wait_for_step_pause 0
						set jumpfound true
					}
					break
				}
			}
		}
		if {! $jumpfound} {
			logprint always "!!!!!!! USER: Target menu not found.  No Action."
		}
	}
	if {$jumpfound} {
		set ::tr::abort_current_step 1
	}
	reset_log_highlight
	return
}

proc goup_button {} {
	if {! $::tr::now_playing} {return}

	set cur_type [lindex $::editor_cmds_from 0]
	set index    [lindex $::editor_cmds_from 1]
	set cur_vts  [lindex [split $index ","] 0]

	reset_log_highlight
	set jumpfound false

	if {$cur_type == "pgc"} {
		set menutype 0

		# remember resume point
		set ::tr::rsm_vts $cur_vts
		set ::tr::rsm_pgc [lindex [split $index ","] 1]
		set ::tr::rsm_cell [expr {$::tr::current(Cell) -1}]
		if {$::tr::rsm_cell < 1} {set ::tr::rsm_cell 1}

		set cur_header $::pgcs($index,header)
		set followingpgc [get_int cur_header 0xA0 2]
		if {$followingpgc > 0} {
			logprint always "!!!!!!! USER: GoUp (Return) Button: Jumping to GoUpPGCN $followingpgc"
			if {![display_VTST_PGC $cur_vts $followingpgc 1]} {
				trace_main_window_state disabled
				init_pgc_playback
				display_current_line
				exec_current_line
				set jumpfound true
			} else {
				logprint always "####### WARNING: Invalid GoUpPGCN $followingpgc!  User action ignored."
				bell
			}
		} else {
			logprint always "!!!!!!! USER: GoUp Button: GoUpPGCN = 0.  No action."
		}
	} else {
		# cur_type is "menu"
		set cur_header $::menus($index,header)
		set followingpgc [get_int cur_header 0xA0 2]
		if {$followingpgc == 65535} {
			logprint always "!!!!!!! USER: GoUp (Return) Button: GoUpPGCN = 0xFFFF: resuming to Title domain."
			if {[info exists ::tr::rsm_vts]} {
				display_VTST_PGC $::tr::rsm_vts $::tr::rsm_pgc 0
				init_pgc_playback
				set cmds $::pgcs($::tr::rsm_vts,$::tr::rsm_pgc)
				set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
				set nextCell $::tr::rsm_cell
				set ::tr::now_playing true
				set jumpfound true
			} else bell
		} elseif {$followingpgc > 0} {
			logprint always "!!!!!!! USER: GoUp (Return) Button: Jumping to GoUpPGCN $followingpgc"
			if {![display_menu_PGC $cur_vts $followingpgc]} {
				set ::menubuttons::menubuttons_done -1
				trace_main_window_state disabled
				init_pgc_playback
				display_current_line
				exec_current_line
				set jumpfound true
			} else {
				logprint always "####### WARNING: Invalid GoUpPGCN $followingpgc!  User action ignored."
				bell
			}
		} else {
			logprint always "!!!!!!! USER: GoUp Button: GoUpPGCN = 0.  No action."
		}
	}
	if {$jumpfound} {
		set ::tr::abort_current_step 1
	}
	return
}

proc chapter_button {direction} {
	if {! $::tr::now_playing} {return}
	set cur_type [lindex $::editor_cmds_from 0]
	set index    [lindex $::editor_cmds_from 1]
	set cur_vts  [lindex [split $index ","] 0]
	set cur_pgc  [lindex [split $index ","] 1]

	reset_log_highlight
	if {$direction == 1} {
		if {$::tr::current(PG) > 0} {
			set pg [expr {$::tr::current(PG) + 1}]
			set cell [program2cell $cur_vts $cur_pgc $pg]
			if {$cell != -1} {
				set ::tr::current(Cell) $cell
				logprint always "!!!!!!! USER: Next Chapter Button: jumping to program $pg"
				set ::menubuttons::menubuttons_done -1
				set ::tr::abort_current_step 1
				set ::tr::wait_step 1
				set ::tr::wait_for_step_pause 0
				return
			} else {
				if {$cur_type == "menu"} {
					set nextpgcn [get_int ::menus($index,header) 0x9C 2]
				} else {
					set nextpgcn [get_int ::pgcs($index,header) 0x9C 2]
					if {$nextpgcn > 0 && $nextpgcn <= $::pgcs($cur_vts,numpgcs) && \
							$::pgcs($cur_vts,$cur_pgc,TTN) != $::pgcs($cur_vts,$nextpgcn,TTN)} {
						logprint always "####### WARNING: Invalid NextPGCN $nextpgcn to another TTN!  User action ignored."
						return
					}
				}
				if {$nextpgcn == 0} {
					logprint always "!!!!!!! USER: Next Chapter Button: NextPGCN = 0.  No action."
					return
				}
				logprint always "!!!!!!! USER: Next Chapter Button: Jumping to post commands."
				# select next line
				set idx [expr {[lindex [::main::get_selection_indexes] 0] +1}]
				::main::set_listbox_selection $idx
				trace_main_window_state disabled
				set ::menubuttons::menubuttons_done -1
				set ::tr::abort_current_step 1
				set ::tr::now_playing false
				reset_log_highlight
#				set ::tr::wait_step 1
#				set ::tr::wait_for_step_pause 0
				return
			}
		}
	} else {
		if {$::tr::current(PG) > 1} {
			set pg [expr {$::tr::current(PG) - 1}]
			set cell [program2cell $cur_vts $cur_pgc $pg]
			if {$cell != -1} {
				set ::tr::current(Cell) $cell
				logprint always "!!!!!!! USER: Prev Chapter Button: jumping to program $pg"
				set ::menubuttons::menubuttons_done -1
				set ::tr::abort_current_step 1
				set ::tr::wait_step 1
				set ::tr::wait_for_step_pause 0
				return
			}
		} else {
			if {$cur_type == "menu"} {
				set prevpgcn [get_int ::menus($index,header) 0x9E 2]
			} else {
				set prevpgcn [get_int ::pgcs($index,header) 0x9E 2]
				if {$prevpgcn > 0 && $prevpgcn <= $::pgcs($cur_vts,numpgcs) && \
						$::pgcs($cur_vts,$cur_pgc,TTN) != $::pgcs($cur_vts,$prevpgcn,TTN)} {
					logprint always "####### WARNING: Invalid PrevPGCN $prevpgcn to another TTN!  User action ignored."
					return
				}
			}
			if {$prevpgcn == 0} {
				logprint always "!!!!!!! USER: Prev Chapter Button: PrevPGCN = 0.  No action."
				return
			}
			logprint always "!!!!!!! USER: Prev Chapter Button: Jumping to PrevPGCN $prevpgcn"

			if {$cur_type == "menu"} {
				set rc [display_menu_PGC $cur_vts $prevpgcn]
			} else {
				set rc [display_VTST_PGC $cur_vts $prevpgcn 1]
			}
			if {!$rc} {
				set ::menubuttons::menubuttons_done -1
				set ::tr::abort_current_step 1
				trace_main_window_state disabled
				init_pgc_playback
				display_current_line
				exec_current_line
			} else {
				bell
				logprint always "####### WARNING: Invalid PrevPGCN $prevpgcn!  User action ignored."
			}
			return
		}
	}
}

#####################################
# Check if the command has an IF condition,
# and return FALSE if the condition doesn't match.
# Optionally, display the comparison result in the status line.
proc check_condition {verbose specialtrue} {
	set idx [string first "if ( " $::tr::disasmbled]
	if {$idx < 0} {return true}
	if {[string first "ILLEGAL" $::tr::disasmbled]!= -1} {return true}

	set ifcmd [string range $::tr::disasmbled [expr {$idx+5}] end]

	set endidx [string first ")" $ifcmd]
	set op1 [string range $ifcmd 0 $endidx]
	set ifcmd [string range $ifcmd [expr {$endidx+1}] end]

	scan $ifcmd "%s %s" operator op2

	if {[string first "gprm(" $op1] >= 0} {
		set cmtidx [string first ":" $op1]
		if {$cmtidx >= 0} {
			set val1 $::tr::gprm([string range $op1 5 [expr {$cmtidx-1}]])
		} else {
			set val1 $::tr::gprm([string range $op1 5 end-1])
		}
	} else {
		set val1 "<ILLEGAL SPRM OPERAND $::tr::sprm([string range $op1 5 [expr {[string first ":" $op1]-1}]])>"
	}

	if {[string first "gprm(" $op2] >= 0} {
		set cmtidx [string first ":" $op2]
		if {$cmtidx >= 0} {
			set val2 $::tr::gprm([string range $op2 5 [expr {$cmtidx-1}]])
		} else {
			set val2 $::tr::gprm([string range $op2 5 end-1])
		}
	} elseif {[string first "sprm(" $op2] >= 0} {
		set val2 $::tr::sprm([string range $op2 5 [expr {[string first ":" $op2]-1}]])
	} else {
		set val2 $op2
	}

	set r false
	switch $operator {
		"==" {if {$val1 == $val2} {set r true}}
		"!=" {if {$val1 != $val2} {set r true}}
		"<"  {if {$val1 <  $val2} {set r true}}
		"<=" {if {$val1 <= $val2} {set r true}}
		">"  {if {$val1 >  $val2} {set r true}}
		">=" {if {$val1 >= $val2} {set r true}}
		"&"  {if {[expr {$val1 & $val2}] != 0} {set r true}}
	}
	# CSetLnk and SetCLnk must be executed even if the confition is false!
	if {$specialtrue && [string first "(SetCLnk)" $::tr::disasmbled] >= 0} {
		set verbose false
	}
	# display the condition
	if {$verbose} {
		logprint if "\tif ( $val1 $operator $val2 ) : $r."
	}
	# CSetLnk and SetCLnk must be executed even if the confition is false!
	if {$specialtrue && ([string first "(CSetLnk)" $::tr::disasmbled] >= 0 || \
			[string first "(SetCLnk)" $::tr::disasmbled] >= 0)} {
		set r true
	}
	return $r
}
#########################################
# Disassemble the current command,
# set the GPRMs and SPRMs accordingly,
# and select the destination PGC and line.
proc trace_disasm {} {
	set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set editorlinenum  [lindex [::main::get_selection_indexes] 0]
	set hex [string range [lindex $::editor_cmds_list $editorlinenum] 1 23]

	set nexteditorline [expr {$editorlinenum+1}]	;# by default, assumes the next command is the next in the section
	set cursection [get_cmd_type cmds $editorlinenum]
	switch $cursection {
		"pre"  {set lineoffset 0}
		"post" {set lineoffset [expr {[num_type_cmds pre cmds]+1}]}
		"cell" {
			set lineoffset [expr {[num_type_cmds pre cmds]+[num_type_cmds post cmds]+2}]
			set nexteditorline [expr {[num_type_cmds pre cmds]+1}]	;# return to playback after cell command
			incr ::tr::current(Cell)
			#set ::tr::now_playing true	;# removed in 0.5.0beta5
		}
	}


	set cmdtype [::disasm::get_nibble $hex 0 0]

	switch $cmdtype {
		0 {
			# flow control commands
			switch [::disasm::get_nibble $hex 1 1] {
				0 {
					# NOP
				}
				1 {
					# Goto
					set nexteditorline [expr {[::disasm::get_byte_val $hex 7 1 128]+$lineoffset}]
				}
				2 {
					# Break
					if {$cursection=="pre"} {
						set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
					} else {
						set cur_type [lindex $::editor_cmds_from 0]
						set cur_header [subst \$::${cur_type}s([lindex $::editor_cmds_from 1],header)]
						set nextpgc [get_int cur_header 0x9C 2]
						set cur_vts  [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
						if {$cur_type == "menu"} {
							set numpgcs $::menus($cur_vts,[lindex [split [lindex $::editor_cmds_from 1] ","] 1],numpgcs)
						} else {
							set numpgcs $::pgcs($cur_vts,numpgcs)
							set cur_ttn $::pgcs([lindex $::editor_cmds_from 1],TTN)
						}
						if {$nextpgc > 0 && $nextpgc <= $numpgcs} {
							if {$cur_type == "menu"} {
								if {![display_menu_PGC $cur_vts $nextpgc]} {
									set nexteditorline 0
								}
							} else {
								if {$::pgcs($cur_vts,$nextpgc,TTN) == $cur_ttn} {
									if {![display_VTST_PGC $cur_vts $nextpgc 1]} {
										set nexteditorline 0
									}
								} else {
									set nexteditorline $editorlinenum
									logprint always "------> Break in post-commands:  NextPGCN link $nextpgc points to another TTN!  Exiting."
									reset_log_highlight
									set ::tr::wait_step 0
								}
							}
						} else {
							set nexteditorline $editorlinenum
							logprint always "------> Break in post-commands:  Invalid NextPGCN link $nextpgc!  Exiting."
							reset_log_highlight
							set ::tr::wait_step 0
						}
					}
				}
				3 {
					# SetTmpPML
					if {[::disasm::get_nibble $hex 6 1] == 15} {
						set linenum [::disasm::get_byte $hex 7]
						if {$linenum > 0} {
							set nexteditorline [expr {$linenum+$lineoffset}]
						}
					} else {
						set level [::disasm::get_nibble_val $hex 6 1 1 8]
						if {$level > $::tr::sprm(13)} {
							set txt "You have configured the parental level to $::tr::sprm(13)."
							set txt "$txt\n\nThe temporary parental level is now set to $level."
							set txt "$txt\n\nDo you accept the new level?"
							set txt "$txt\n(A real player should prompt for a password.)"
							set r [tk_messageBox -title "Trace" -icon question -message $txt \
									-type yesno -default no]
						} else {
							set r "yes"
						}
						if {$r == "yes"} {
							set nexteditorline [expr {[::disasm::get_byte_val $hex 7 1 128]+$lineoffset}]
						}
					}
				}
			}
		}
		2 {
			# link commands
			set k [::disasm::get_nibble $hex 1 1]
			if {$k == 1} {
				set ll [::disasm::get_byte $hex 7]
			} else {
				set ll 0
			}
			if {[::disasm::get_nibble $hex 1 1] != 4} {
				set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
				if {$hl_button > 0} {
					set ::tr::sprm(8) $hl_button
				}
			}
			set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline $editorlinenum]
		}
		3 {
			# jump/call commands
			set jumpfound false
			switch [::disasm::get_nibble $hex 1 1] {
				1 {
					# Exit
					set nexteditorline $editorlinenum
					logprint always "------> Exiting!  DVD Playback end."
					reset_log_highlight
					set ::tr::wait_step 0
					set jumpfound true
				}
				2 {
					# JumpTT
					set title [::disasm::get_byte_val $hex 5 1 99]
					set jumpfound false
					if {$title>0 && $title<100} {
#						for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
#							for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
#								if {[info exists ::pgcs($vts,$pgc,Title)] && \
#										$::pgcs($vts,$pgc,Title) == $title} {
#									set targetvts $vts
#									set targetpgc $pgc
#									set jumpfound true
#									break
#								}
#							}
#							if {$jumpfound} break
#						}
						if {[info exists ::pgcs(Title,$title)]} {
							set vts [lindex $::pgcs(Title,$title) 0]
							set pgc [lindex $::pgcs(Title,$title) 1]
							if {[info exists ::pgcs($vts,$pgc)]} {
								set targetvts $vts
								set targetpgc $pgc
								set jumpfound true
							}
						}
					}
					if {$jumpfound} {
						display_VTST_PGC $targetvts $targetpgc 1
						init_pgc_playback
						set nexteditorline 0
					}
				}
				3 {
					# JumpVTS_TT
					set ttn [::disasm::get_byte_val $hex 5 1 99]
					set vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
					if {[info exists ::ttns($vts,$ttn)]} {
						set pgc $::ttns($vts,$ttn)
						if {![display_VTST_PGC $vts $pgc 1]} {
							init_pgc_playback
							set nexteditorline 0
							set jumpfound true
						}
					}
				}
				5 {
					# JumpVTS_PTT
					set pgclist [split [lindex $::editor_cmds_from 1] ","]
					set vts [lindex $pgclist 0]
					set ttn [::disasm::get_byte_val $hex 5 1 99]
					set chapter [expr {[::disasm::get_word $hex 2 false] & 0X03FF}]
					set jumpfound false
					for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
						if {[info exists ::pgcs($vts,$p,TTN)] && $::pgcs($vts,$p,TTN) == $ttn} {
							foreach j $::pgcs($vts,$p,pttlist) {
								if {[lindex $j 1] == $chapter} {
									set jumpfound true
									set pgc $p
									break
								}
							}
						}
						if {$jumpfound} break
					}
					if {$jumpfound} {
						if {![display_VTST_PGC $vts $pgc $chapter]} {
							set cell [chapter2cell $vts $pgc $chapter]
							if {$cell != -1} {
								set cmds $::pgcs($vts,$pgc)
								set nexteditorline 0
								init_pgc_playback
								set ::tr::current(Cell) $cell
								set ::tr::current(PTT) $chapter
								set ::tr::display(PTT) $chapter
								set ::tr::now_playing false
							} else {
								set jumpfound false
							}
						}
					}
				}
				6 {
					# JumpSS *
					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# JumpSS-FP
							display_VTST_PGC 0 0 0
							set nexteditorline 0
							init_pgc_playback
							set jumpfound true
						}
						4 {
							# JumpSS-VMGM-menu
							if {[info exists ::menus(0,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
										if {![display_menu_PGC 0 $pgc]} {
											set nexteditorline 0
											init_pgc_playback
											set jumpfound true
										}
										break
									}
								}
							}
						}
						8 {
							# JumpSS-VTSM
							set ttn [::disasm::get_byte_val $hex 3 0 99]
							if {$ttn > 0} {
								set ::tr::sprm(5) $ttn
							}
							set vts [::disasm::get_byte_val $hex 4 0 99]
							if {$vts == 0} {	;# VTS=0 means Last visited VTS, or current VTS
								set vts $::tr::last_visited_vts
							}
							set menutype [::disasm::get_nibble $hex 5 1]
							if {[info exists ::menus($vts,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus($vts,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus($vts,1,$pgc,mask) & 0x7F}] == $menutype} {
										if {![display_menu_PGC $vts $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						12 {
							# JumpSS-VMGM-pgc
							if {![display_menu_PGC 0 [::disasm::get_word_val $hex 2 1 999]]} {
								init_pgc_playback
								set nexteditorline 0
								set jumpfound true
							}
						}
					}
				}
				8 {
					# CallSS *
					set ::tr::rsm_vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
					set ::tr::rsm_pgc [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
					set ::tr::rsm_cell [::disasm::get_byte_val $hex 4 0 255]
					if {$::tr::rsm_cell == 0} {
						set ::tr::rsm_cell [expr {$::tr::current(Cell) -1}]
					}

					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# CallSS-FP
							display_VTST_PGC 0 0 0
							#init_pgc_playback
							set nexteditorline 0
							set jumpfound true
						}
						4 {
							# CallSS-VMGM-menu
							if {[info exists ::menus(0,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
										display_menu_PGC 0 $pgc
										init_pgc_playback
										set nexteditorline 0
										set jumpfound true
										break
									}
								}
							}
						}
						8 {
							# CallSS-VTSM
							set vts $::tr::last_visited_vts
							set menutype [::disasm::get_nibble $hex 5 1]
							if {[info exists ::menus($vts,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus($vts,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus($vts,1,$pgc,mask) & 0x7F}] == $menutype} {
										if {![display_menu_PGC $vts $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						12 {
							# CallSS-VMGM-pgc
							if {![display_menu_PGC 0 [::disasm::get_word_val $hex 2 1 999]]} {
								init_pgc_playback
								set nexteditorline 0
								set jumpfound true
							}
						}
					}
				}
			}
			if {! $jumpfound} {
				bell
				logprint always "####### Target not found!"
				reset_log_highlight
				set nexteditorline $editorlinenum
				set ::tr::wait_step 0
			}
		}
		4 -
		5 {
			# Set SPRM commands

			switch [::disasm::get_nibble $hex 0 1] {
				1 {
					# SetSTN
					set audio [::disasm::get_byte $hex 3]
					if {[expr {$audio & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(1) [expr {$audio & 0x0F}]
						} else {
							set ::tr::sprm(1) $::tr::gprm([expr {$audio & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(Audio)  \
									"[::streams::get_pgc_audio_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(1)]"
						}
					}
					set subpic [::disasm::get_byte $hex 4]
					if {[expr {$subpic & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(2) [expr {$subpic & 0x7F}]
						} else {
							set ::tr::sprm(2) $::tr::gprm([expr {$subpic & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(SubPic) \
									"[::streams::get_pgc_subpic_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(2)]"
						}
					}
					set angle [::disasm::get_byte $hex 5]
					if {[expr {$angle & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(3) [expr {$angle & 0x0F}]
						} else {
							set ::tr::sprm(3) $::tr::gprm([expr {$angle & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(Angle)  $::tr::sprm(3)
						}
					}
				}
				2 {
					# SetNVTMR
					if {$cmdtype == 5} {
						set ::tr::sprm(9) [::disasm::get_word $hex 2 false]
					} else {
						set reg [::disasm::get_byte $hex 3]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(9) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(9) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					set ::tr::sprm(10) [::disasm::get_word_val $hex 4 1 999]
					if {$::tr::sprm(9) > 0} {
						set txt "The Navigation timer has been set to $::tr::sprm(9) seconds,"
						set txt "$txt\nand the Jump for Nav Timer to PGC number $::tr::sprm(10)."
						set txt "$txt\n\nThis jump cannot be simulated with Trace..."	;# TODO:
						tk_messageBox -title "Trace" -icon warning -message $txt -parent .
					}
				}
				3 {
					# SetGPRMMD
					set gprm [::disasm::get_nibble $hex 5 1]
					if {$cmdtype == 5} {
						set ::tr::gprm($gprm) [::disasm::get_word $hex 2 false]
					} else {
						set reg [::disasm::get_byte $hex 3]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::gprm($gprm) $::tr::gprm($reg)
						} else {
							set ::tr::gprm($gprm) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					set mode [expr {[::disasm::get_nibble $hex 5 0] & 8}]
					if {$mode == 8} {
						set txt "The GPRM($gprm) has been set to $::tr::gprm($gprm) in counter mode."
						set txt "$txt\n\nThe counter mode cannot be simulated with Trace..."	;# TODO:
						tk_messageBox -title "Trace" -icon warning -message $txt -parent .
					}
				}
				4 {
					# SetAMXMD
					if {$cmdtype == 5} {
						set ::tr::sprm(11) [::disasm::get_word_val $hex 4 0 7196]
					} else {
						set reg [::disasm::get_byte $hex 5]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(11) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(11) $::tr::sprm([expr {$reg-0x80}])
						}
					}
				}
				6 {
					# SetHL_BTN
					if {$cmdtype == 5} {
						set but [expr {([::disasm::get_byte $hex 4] / 4) * 1024}]
						set ::tr::sprm(8) $but
					} else {
						set reg [::disasm::get_byte $hex 5]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(8) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(8) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					# TODO: no button here because it is already set!
				}
			}

			# Set + Link
			# (ces commandes ne peuvent pas avoir en meme temps un IF et un Link)
			set k [::disasm::get_nibble $hex 1 1]
			if {$k > 0} {
				if {$k == 1} {
					set ll [::disasm::get_byte $hex 7]
				} else {
					set ll 0
				}
				if {$k != 4} {
					set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
					if {$hl_button > 0} {
						set ::tr::sprm(8) $hl_button
					}
				}
				set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline $editorlinenum]
				#init_pgc_playback
			}
		}
		6 -
		7 {
			# Set command
			if {[expr {$cmdtype & 1}]} {
				set val [::disasm::get_word $hex 4 false]
			} else {
				set reg [::disasm::get_byte $hex 5]
				if {[expr {$reg & 0x80}] == 0} {
					set val $::tr::gprm($reg)
				} else {
					set val $::tr::sprm([expr {$reg & 0x1F}])
				}
			}
			set gprm1 [::disasm::get_byte $hex 3]
			set op1 $::tr::gprm($gprm1)
			set operand [::disasm::get_nibble $hex 0 1]
			if {$operand == 2} {
				# SWAP operation
				set ::tr::gprm($gprm1) $val
				if {[expr {$reg & 0x80}] == 0} {
					set ::tr::gprm($reg) $op1
				} else {
					set ::tr::sprm([expr {$reg & 0x1F}]) $op1
				}
			} else {
				set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
			}

			# Link part
			# Set ne peut pas avoir en meme temps un IF et un Link
			set k [::disasm::get_nibble $hex 1 1]
			if {$k > 0} {
				if {$k == 1} {
					set ll [::disasm::get_byte $hex 7]
				} else {
					set ll 0
				}
				if {$k != 4} {
					set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
					if {$hl_button > 0} {
						set ::tr::sprm(8) $hl_button
					}
				}
				set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline $editorlinenum]
				#init_pgc_playback
			}
		}
		8 -
		9 {
			# SetCLnk
			# Set part
			if {[expr {$cmdtype & 1}]} {
				set val [::disasm::get_word $hex 2 false]
			} else {
				set reg [::disasm::get_byte $hex 3]
				if {[expr {$reg & 0x80}] == 0} {
					set val $::tr::gprm($reg)
				} else {
					set val $::tr::sprm([expr {$reg & 0x1F}])
				}
			}
			set gprm1 [::disasm::get_nibble $hex 1 1]
			set op1 $::tr::gprm($gprm1)
			set operand [::disasm::get_nibble $hex 0 1]
			if {$operand == 2} {
				# SWAP operation
				set ::tr::gprm($gprm1) $val
				if {[expr {$reg & 0x80}] == 0} {
					set ::tr::gprm($reg) $op1
				} else {
					set ::tr::sprm([expr {$reg & 0x1F}]) $op1
				}
			} else {
				set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
			}

			# if then link part
			if {[check_condition true false]} {
				set ll [::disasm::get_byte $hex 7]
				set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
				if {$hl_button > 0} {
					set ::tr::sprm(8) $hl_button
				}
				set nexteditorline [trace_disasm_link $hex 1 $ll $nexteditorline $editorlinenum]
				#init_pgc_playback
			}
		}
		10 -
		11 -
		12 -
		13 {
			# CSetCLnk and CSetLnk
			# CSetCLnk don't even enter here if the condition is not true

			# set part: CSetLnk performs the set conditionally
			if {$cmdtype < 12 || [check_condition false false]} {
				if {[expr {$cmdtype & 1}]} {
					set val [::disasm::get_word $hex 2 false]
				} else {
					set reg [::disasm::get_byte $hex 2]
					if {[expr {$reg & 0x80}] == 0} {
						set val $::tr::gprm($reg)
					} else {
						set val $::tr::sprm([expr {$reg & 0x1F}])
					}
				}
				set gprm1 [::disasm::get_nibble $hex 1 1]
				set op1 $::tr::gprm($gprm1)
				set operand [::disasm::get_nibble $hex 0 1]
				if {$operand == 2} {
					# SWAP operation
					set ::tr::gprm($gprm1) $val
					if {[expr {$reg & 0x80}] == 0} {
						set ::tr::gprm($reg) $op1
					} else {
						set ::tr::sprm([expr {$reg & 0x1F}]) $op1
					}
				} else {
					set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
				}
			}

			# Link part: CSetLnk performs the link unconditionally
			set ll [::disasm::get_byte $hex 7]
			set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
			if {$hl_button > 0} {
				set ::tr::sprm(8) $hl_button
			}
			set nexteditorline [trace_disasm_link $hex 1 $ll $nexteditorline $editorlinenum]
			#init_pgc_playback
		}
	}
	# select the line to be executed next
	if {$nexteditorline != -1} {
		::main::set_listbox_selection $nexteditorline
	}
}


proc trace_disasm_link {hex k ll nexteditorline editorlinenum} {
	# link commands
	set nextCell -1
	set nextPG   -1
	set jumpfound false

	set cur_type [lindex $::editor_cmds_from 0]
	set index    [split [lindex $::editor_cmds_from 1] ","]
	set cur_vts  [lindex $index 0]
	if {$cur_type == "menu"} {
		set cur_lu  [lindex $index 1]
		set cur_pgc [lindex $index 2]
		set cur_numpgcs $::menus($cur_vts,$cur_lu,numpgcs)
		set cur_numpgs  $::menus($cur_vts,$cur_lu,$cur_pgc,numprograms)
		set cmds $::menus($cur_vts,$cur_lu,$cur_pgc)
	} else {
		set cur_lu  -1
		set cur_pgc [lindex $index 1]
		set cur_numpgcs $::pgcs($cur_vts,numpgcs)
		set cur_numpgs  $::pgcs($cur_vts,$cur_pgc,numprograms)
		set cur_ttn $::pgcs($cur_vts,$cur_pgc,TTN)
		set cmds $::pgcs($cur_vts,$cur_pgc)
	}
	set cur_numcells   [subst \$::${cur_type}s([lindex $::editor_cmds_from 1],numcells)]
	set cur_header     [subst \$::${cur_type}s([lindex $::editor_cmds_from 1],header)]

	switch $k {
		1 {
			# Link* (k=1)
			switch $ll {
				0 {
					# LinkNoLink
					set jumpfound true
				}
				1 {
					# LinkTopCell
					set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
					if {$editorlinenum == -200} {
						set nextCell $::tr::current(Cell)				;# menu button bug fixed in 0.5.7
					} else {
						set nextCell [expr {$::tr::current(Cell) -1}]
					}
					set jumpfound true
				}
				2 {
					# LinkNextCell
					if {$editorlinenum == -200} {
						set newcell [expr {$::tr::current(Cell) + 1}]	;# menu button bug fixed in 0.5.2
					} else {
						set newcell $::tr::current(Cell)
					}
					if {$newcell <= $cur_numcells} {
						set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
						set nextCell $newcell
						set jumpfound true
					}
				}
				3 {
					# LinkPrevCell
					if {$editorlinenum == -200} {
						set newcell [expr {$::tr::current(Cell) - 1}]	;# menu button bug fixed in 0.5.2
					} else {
						set newcell [expr {$::tr::current(Cell) - 2}]
					}
					if {$newcell >= 1} {
						set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
						set nextCell $newcell
						set jumpfound true
					}
				}
				5 {
					# LinkTopPG
					set nextPG $::tr::current(PG)
					#set nexteditorline -1
					#set jumpfound true
				}
				6 {
					# LinkNextPG
					set pg [expr {$::tr::current(PG) + 1}]
					if {$pg <= $cur_numpgs} {
						set nextPG $pg
						#set nexteditorline -1
						#set jumpfound true
					}
				}
				7 {
					# LinkPrevPG
					set pg [expr {$::tr::current(PG) - 1}]
					if {$pg > 0} {
						set nextPG $pg
						#set nexteditorline -1
						#set jumpfound true
					}
				}
				9 {
					# LinkTopPGC
					set nexteditorline 0
					set jumpfound true
					init_pgc_playback
				}
				10 {
					# LinkNextPGC
					set nextpgc [get_int cur_header 0x9C 2]
					if {$nextpgc > 0 && $nextpgc <= $cur_numpgcs} {
						if {$cur_type == "menu"} {
							if {![display_menu_PGC $cur_vts $nextpgc]} {
								set nexteditorline 0
								set jumpfound true
							}
						} else {
							if {$cur_ttn == $::pgcs($cur_vts,$nextpgc,TTN)} {
								if {![display_VTST_PGC $cur_vts $nextpgc 1]} {
									set nexteditorline 0
									set jumpfound true
								}
							}
						}
					}
				}
				11 -
				12 {
					# LinkPrevPGC & LinkGoUpPGC
					if {$ll == 11} {
						set followingpgc [get_int cur_header 0x9E 2]
					} else {
						set followingpgc [get_int cur_header 0xA0 2]
					}
					if {$followingpgc > 0 && $followingpgc <= $cur_numpgcs} {
						if {$cur_type == "menu"} {
							if {![display_menu_PGC $cur_vts $followingpgc]} {
								set nexteditorline 0
								init_pgc_playback
								set jumpfound true
							}
						} else {
							if {$cur_ttn == $::pgcs($cur_vts,$followingpgc,TTN)} {
								if {![display_VTST_PGC $cur_vts $followingpgc 1]} {
									set nexteditorline 0
									init_pgc_playback
									set jumpfound true
								}
							}
						}
					}
				}
				13 {
					# LinkTailPGC
					set nexteditorline [expr {[num_type_cmds pre cmds]+2}]
					set jumpfound true
				}
				16 {
					# RSM
					if {[info exists ::tr::rsm_vts]} {
						display_VTST_PGC $::tr::rsm_vts $::tr::rsm_pgc 0
						init_pgc_playback
						set cmds $::pgcs($::tr::rsm_vts,$::tr::rsm_pgc)
						set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
						set nextCell $::tr::rsm_cell
						set ::tr::now_playing true
						set jumpfound true
					} else bell
				}
			}
		}
		4 {
			# LinkPGCN
			set nextpgc [::disasm::get_word_val $hex 6 1 32767]
			set pgclist [split [lindex $::editor_cmds_from 1] ","]
			if {$cur_type == "menu"} {
				if {![display_menu_PGC $cur_vts $nextpgc]} {
					set nexteditorline 0
					init_pgc_playback
					set jumpfound true
				}
			} else {
				if {$cur_type != "menu"} {
				    if {$cur_ttn != $::pgcs($cur_vts,$nextpgc,TTN)} {
						bell
						logprint always "####### Prohibited link from TTN $cur_ttn to TTN $::pgcs($cur_vts,$nextpgc,TTN) !"
						reset_log_highlight
						set ::tr::wait_step 0
						return $editorlinenum
				    }
				}
				if {![display_VTST_PGC $cur_vts $nextpgc 1]} {
					set nexteditorline 0
					init_pgc_playback
					set jumpfound true
				}
			}
		}
		5 {
			# LinkPTT
			set chapter [expr {[::disasm::get_word $hex 6 false] & 0X03FF}]
			set pgclist [split [lindex $::editor_cmds_from 1] ","]
			set vts [lindex $pgclist 0]
			set origpgc [lindex $pgclist 1]
			set jumpfound false
			if {[info exists ::pgcs($vts,$origpgc,TTN)]} {
				set ttn $::pgcs($vts,$origpgc,TTN)
				for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
					if {$::pgcs($vts,$p,TTN) == $ttn} {
						foreach j $::pgcs($vts,$p,pttlist) {
							if {[lindex $j 1] == $chapter} {
								set jumpfound true
								set pgc $p
								set nextPG [lindex $j 0]
								break
							}
						}
					}
					if {$jumpfound} break
				}
			}
			if {$jumpfound} {
				if {![display_VTST_PGC $vts $pgc $chapter]} {
					set cell [chapter2cell $vts $pgc $chapter]
					if {$cell != -1} {
						set cmds $::pgcs($vts,$pgc)
						set nexteditorline 0
						init_pgc_playback
						set ::tr::current(Cell) $cell
						set ::tr::current(PTT) $chapter
						set ::tr::display(PTT) $chapter
						set ::tr::now_playing false
					} else {
						set jumpfound false
					}
				}
			}
		}
		6 {
			# LinkPGN
			set nextPG [::disasm::get_byte_val $hex 7 1 99]
		}
		7 {
			# LinkCN
			set cell [::disasm::get_byte_val $hex 7 1 255]
			if {$cell > 0 && $cell <= $cur_numcells} {
				set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
				set nextCell $cell
				set jumpfound true
			}
		}
	}
	# if link to a specific program, compute the cell number
	if {$nextPG != -1} {
		if {$nextPG > 0 && $nextPG <= $cur_numpgs} {
			set programmap [subst \$::${cur_type}s([lindex $::editor_cmds_from 1],programmap)]
			set nextCell [get_int programmap [expr {$nextPG-1}] 1]
		}
	}

	set new_numcells [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1],numcells)]
	if {$nextCell > 0 && $nextCell <= $new_numcells} {
		set ::tr::current(Cell) $nextCell
		set nexteditorline [expr {[num_type_cmds pre cmds]+1}]
		set jumpfound true
	}

	if {! $jumpfound} {
		bell
		logprint always "####### Target not found!"
		reset_log_highlight
		set ::tr::wait_step 0
		return $editorlinenum
	} else {
		return $nexteditorline
	}
}

## compute the expression, and return the value
## WARNING: this function will return the value of the SWAP operation for the target
## register.  The source register is not modified!
proc operation_result {op1 operand val} {
	switch $operand {
		1 { ;# = SET
			return $val
		}
		2 { ;# <-> SWAP !!! NOT SWAPPED: actually, this function allow only to return ONE value!
			return $val
		}
		3 { ;# += ADD
			set r [expr {$op1 + $val}]
			return [expr $r & 0xFFFF]
		}
		4 { ;# -= SUB
			set r [expr {$op1 - $val}]
			if {$r < 0} {return [expr $r + 0x10000]}
			return $r
		}
		5 { ;# *= MUL
			set r [expr {$op1 * $val}]
			return [expr $r & 0xFFFF]
		}
		6 { ;# /= DIV
			if {$val == 0} {return 65535}
			return [expr {$op1 / $val}]
		}
		7 { ;# %= MOD
			if {$val == 0} {return 65535}
			return [expr {$op1 % $val}]
		}
		8 { ;# ?= RND
			# TODO: What to do here if operator2 ($val) is 0 ?
			# Software players are returning 1 but Sony standalone return 0
			if {$val == 0} {return 0}
			return [expr {round( ( rand() * ($val-1) ) ) +1}]	;# return a value between 1 and $val
		}
		9 { ;# &= AND
			return [expr {$op1 & $val}]
		}
		10 { ;# |= OR
			return [expr {$op1 | $val}]
		}
		11 { ;# ^= XOR
			return [expr {$op1 ^ $val}]
		}
		default {
			bell
			return 0
		}
	}
}


# return the Cell number for a gived program number, or -1 if not found.
proc program2cell {vts pgc pg} {
	if {$pg > 0} {
		if {[string length $::pgcs($vts,$pgc,programmap)] < $pg} {
			return -1
		}
		if {[catch {set c [get_int ::pgcs($vts,$pgc,programmap) [expr {$pg-1}] 1]}]} {
			return -1
		}
		if {$c > 0} {
			return $c
		}
	}
	return -1
}

# return the Cell number for a gived chapter number, or -1 if not found.
proc chapter2cell {vts pgc pttn} {
	if {$pttn > 0} {
		set jumpfound false
		if {[info exists ::pgcs($vts,$pgc,TTN)]} {
			set ttn $::pgcs($vts,$pgc,TTN)
			for {set i 1} {$i <= $::pgcs($vts,numpgcs)} {incr i} {
				if {$::pgcs($vts,$i,TTN) == $ttn} {
					foreach j $::pgcs($vts,$i,pttlist) {
						if {[lindex $j 1] == $pttn} {
							set jumpfound true
							set pgc $i
							break
						}
					}
				}
				if {$jumpfound} break
			}
		}
		if {! $jumpfound} {
			return -1
		}
		set numcells $::pgcs($vts,$pgc,numcells)
		set pttlist  $::pgcs($vts,$pgc,pttlist)
		set numptts  [llength $pttlist]
		foreach pair $pttlist {
			if {[lindex $pair 1] == $pttn} {
				set pg [lindex $pair 0]
				if {$pg > 0} {
					if {[catch {set c [get_int ::pgcs($vts,$pgc,programmap) [expr {$pg-1}] 1]}]} {
						return -1
					}
					return $c
				}
			}
		}
	}
	return -1
}

proc init_pgc_playback {} {
	set ::tr::current(Cell)   1
	set ::tr::current(PG)     1
	set ::tr::current(PTT)    1

	# init PGC playback state
	set cur_type [lindex $::editor_cmds_from 0]
	set index    [split [lindex $::editor_cmds_from 1] ","]
	set cur_vts  [lindex $index 0]
	if {$cur_type == "menu"} {
		# reset NAV timer if we enter a new domain
		if {$::tr::sprm(9) != 0 || $::tr::sprm(10) != 0} {
			set ::tr::sprm(9)  0
			set ::tr::sprm(10) 0
		}
		set cur_lu  [lindex $index 1]
		set ::tr::current(PGC) [lindex $index 2]
		#set ::tr::current(TTN) -1
		#set ::tr::current(Title) -1
		if {[winfo exists .mf.f2.selector.a.lb]} {
			set lab [::main::pgc_label $::config(global_font) menu $cur_vts $cur_lu $::tr::current(PGC)]
			.mf.f2.selector.a.lb itemconfigure [lsearch $::main::labelslist $lab] -background #FFFFAF
		}
	} else {
		# reset NAV timer if we enter a new domain
		if {$cur_vts == 0 && ($::tr::sprm(9) != 0 || $::tr::sprm(10) != 0)} {
			set ::tr::sprm(9)  0
			set ::tr::sprm(10) 0
		}
		set cur_lu  -1
		set ::tr::current(PGC) [lindex $index 1]
		if {[info exists ::pgcs($cur_vts,$::tr::current(PGC),TTN)]} {
			set ::tr::current(TTN) $::pgcs($cur_vts,$::tr::current(PGC),TTN)
		}
		if {[info exists ::pgcs($cur_vts,$::tr::current(PGC),Title)]} {
			set ::tr::current(Title) $::pgcs($cur_vts,$::tr::current(PGC),Title)
		} elseif {$cur_vts != 0} {
			for {set i 1} {$i <= $::pgcs($cur_vts,numpgcs)} {incr i} {
				if {[info exists ::pgcs($cur_vts,$i,TTN)] && \
						$::pgcs($cur_vts,$i,TTN) == $::tr::current(TTN) && \
						[info exists ::pgcs($cur_vts,$i,Title)]} {
					set ::tr::current(Title) $::pgcs($cur_vts,$i,Title)
				}
			}
		}
		set ::tr::display(Title)  $::tr::current(Title)
		set ::tr::display(PTT)    $::tr::current(PTT)
		set ::tr::display(Audio)  \
				"[::streams::get_pgc_audio_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(1)]"
		set ::tr::display(SubPic) \
				"[::streams::get_pgc_subpic_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(2)]"
		set ::tr::display(Angle)  $::tr::sprm(3)
		if {[winfo exists .mf.f2.selector.a.lb]} {
			set lab [::main::pgc_label $::config(global_font) pgc $cur_vts $::tr::current(PGC)]
			.mf.f2.selector.a.lb itemconfigure [lsearch $::main::labelslist $lab] -background #FFFFAF
		}
	}
	snapshot_copy pgc last_pgc
	snapshot_state pgc
	if {$::tr::last_domain != [list $cur_type $cur_vts]} {
		snapshot_copy domain last_domain
		snapshot_state domain
	}
	set ::tr::last_domain [list $cur_type $cur_vts]
	update idletasks
}

###############################
# simulate the playing of a Title or Menu PGC
# TODO: update/use sprm(14:Video prefs & current mode)
proc simulate_play {} {
	set ::tr::abort_current_step 0	;# changed to 1 if a menu button was pressed during playback

	# check if we are on a "* post" label.
	# If not, then the user has selected another line,
	# and we must leave the now_playing mode
	set idx [lindex [::main::get_selection_indexes] 0]
	if {$idx == "-"} {bell; set ::tr::now_playing false; return}
	set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set cmdtype [get_cmd_type cmds $idx]
	if {$cmdtype != "* post"} {
		set ::tr::now_playing false
		trace_main_window_state disabled
		return
	}
	set ::tr::now_playing true

	# init current state
	set cur_type [lindex $::editor_cmds_from 0]
	set index    [split [lindex $::editor_cmds_from 1] ","]
	set cur_vts  [lindex $index 0]
	if {$cur_type == "menu"} {
		set cur_lu  [lindex $index 1]
		set cur_pgc [lindex $index 2]
		set cur_numcells  $::menus($cur_vts,$cur_lu,$cur_pgc,numcells)
		set cur_pttlist ""
		set cur_numptts 0

		set programmap $::menus($cur_vts,$cur_lu,$cur_pgc,programmap)
		for {set p 1} {$p <= $::menus($cur_vts,$cur_lu,$cur_pgc,numprograms)} {incr p} {
			set cell_program([get_int programmap [expr {$p-1}] 1]) $p
		}
		set old_pg 1
		for {set c 1} {$c <= $::menus($cur_vts,$cur_lu,$cur_pgc,numcells)} {incr c} {
			if {[info exists cell_program($c)]} {
				set old_pg $cell_program($c)
			} else {
				set cell_program($c) $old_pg
			}
		}
		set cellpbtable $::menus($cur_vts,$cur_lu,$cur_pgc,cellpbtable)
	} else {
		set cur_lu  -1
		set cur_pgc [lindex $index 1]
		set cur_numcells  $::pgcs($cur_vts,$cur_pgc,numcells)
		set cur_pttlist $::pgcs($cur_vts,$cur_pgc,pttlist)
		set cur_numptts [llength $cur_pttlist]
		set programmap $::pgcs($cur_vts,$cur_pgc,programmap)

		for {set p 1} {$p <= $::pgcs($cur_vts,$cur_pgc,numprograms)} {incr p} {
			set cell_program([get_int programmap [expr {$p-1}] 1]) $p
		}
		set old_pg 1

		for {set c 1} {$c <= $::pgcs($cur_vts,$cur_pgc,numcells)} {incr c} {
			if {[info exists cell_program($c)]} {
				set old_pg $cell_program($c)
			} else {
				set cell_program($c) $old_pg
			}
		}
		set cellpbtable $::pgcs($cur_vts,$cur_pgc,cellpbtable)
	}

	# set PUOs for current PGC
	set header [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1],header)]
	#puts "PUOs: [get_int header 8 4]"
	trace_main_window_state_puos [get_int header 8 4]

	# determine the current PG Playback mode
	set tmp [get_int header 0xA2 1]
	if {$tmp > 0} {
		set numtoplay [expr {($tmp & 0x7F)}]
		set numplayed [llength $::tr::cur_pg_pb_rand_pgs]
		set cur_numprograms $::pgcs($cur_vts,$cur_pgc,numprograms)
		if {[expr {$tmp & 0x80}] == 0} {
			if {[llength $::tr::cur_pg_pb_rand_pgs] == 0 || $::tr::current(Cell) > $cur_numcells || \
					$cell_program($::tr::current(Cell)) != [lindex $::tr::cur_pg_pb_rand_pgs end]} {
				# random mode: randomly pick a program
				if {$numplayed < $numtoplay} {
					set rnd [expr {round(rand()*($cur_numprograms-1))}]
					set ::tr::current(Cell) [get_int programmap $rnd 1]
					set pg $cell_program($::tr::current(Cell))
					lappend ::tr::cur_pg_pb_rand_pgs $pg
				} else {
					set ::tr::current(Cell) [expr {$cur_numcells + 1}]
					unset ::tr::cur_pg_pb_rand_pgs
				}
			}
		} else {
			# shuffle mode: randomly pick an unplayed program
			if {[llength $::tr::cur_pg_pb_rand_pgs] == 0 || $::tr::current(Cell) > $cur_numcells || \
					$cell_program($::tr::current(Cell)) != [lindex $::tr::cur_pg_pb_rand_pgs end]} {
				if {$numplayed < $numtoplay} {
					set lastplayed $::tr::cur_pg_pb_rand_pgs
					while {[llength $lastplayed] >= $cur_numprograms} {
						set lastplayed [lrange $lastplayed $cur_numprograms end]
					}

					set rnd [expr {round(rand()*($cur_numprograms-1))}]
					set ::tr::current(Cell) [get_int programmap $rnd 1]
					set pg $cell_program($::tr::current(Cell))
					while {[lsearch $lastplayed $pg] != -1} {
						set rnd [expr {round(rand()*($cur_numprograms-1))}]
						set ::tr::current(Cell) [get_int programmap $rnd 1]
						set pg $cell_program($::tr::current(Cell))
					}
					lappend ::tr::cur_pg_pb_rand_pgs $pg
				} else {
					set ::tr::current(Cell) [expr {$cur_numcells + 1}]
					unset ::tr::cur_pg_pb_rand_pgs
				}
			}
		}
	}
	if {$::tr::current(Cell) <= 0} {
		set ::tr::current(Cell) 1
	}

	# skip angle cell if it is not the current one
	while {$::tr::current(Cell) <= $cur_numcells} {
		set angle 0
		set cur_celltype [get_int cellpbtable [expr {($::tr::current(Cell) - 1) * 24}] 1]
		if {[expr {$cur_celltype & 0x10}] == 0x10} {
			# we are in an angle cell.
			if {! [info exists ::tr::last_angle]} {set ::tr::last_angle 1}	;# normally not needed
			if {[expr {$cur_celltype & 0xC0}] == 0x40} {
				# first angle of block
				set angle 1
				set ::tr::last_angle 1
			} else {
				# middle or last angle of block
				incr ::tr::last_angle
				set angle $::tr::last_angle
			}
		}
		if {$angle == 0 || $angle == $::tr::sprm(3)} break
		incr ::tr::current(Cell)
	}

	# simulate play of the current cell
	if {$::tr::current(Cell) <= $cur_numcells} {
		set ::tr::current(PG) $cell_program($::tr::current(Cell))

		set ::tr::current(PTT) ""
		foreach pair $cur_pttlist {
			if {[lindex $pair 0] == $cell_program($::tr::current(Cell))} {
				set ::tr::current(PTT) [lindex $pair 1]
				break;
			}
		}
		set bcdcelltime [string range $cellpbtable [expr {(($::tr::current(Cell) - 1) * 24) + 4}] [expr {(($::tr::current(Cell) - 1) * 24) + 7}]]
		set ncelltime [::pgced::short_BCDtime $bcdcelltime]
		set msg "******* Playing Program $::tr::current(PG), Cell $::tr::current(Cell) ($ncelltime"
		.mf.cf.trace.fl.ft.f1.preview configure -state normal -command \
			"::preview::preview2 [lindex $::editor_cmds_from 0] \
					[lindex $::editor_cmds_from 1] $::tr::current(Cell) false false 0"
		set pgcstilltime [get_int header 0xA3 1]
		if {$pgcstilltime > 0 && $::tr::current(Cell) == $cur_numcells} {
			set nstilltime $pgcstilltime
		} else {
			set nstilltime [get_int cellpbtable [expr {(($::tr::current(Cell) - 1) * 24) + 2}] 1]
		}
		if {$nstilltime > 0} {
			if {$nstilltime == 255} {
				set msg "$msg + infinite pause"
			} else {
				set msg "$msg + $nstilltime s. pause"
			}
		}
		set msg "$msg)"
		if {$::tr::current(PTT) != ""} {
			set msg "$msg (Chapter $::tr::current(PTT))"
			if {$angle != 0} {set msg "$msg, Angle $angle"}
			logprint always $msg
			if {$::tr::playerconfig(sprm7holdsPG)} {
				set ::tr::sprm(7) $::tr::current(PG)
			} else {
				set ::tr::sprm(7) $::tr::current(PTT)
			}
			set ::tr::display(PTT) $::tr::current(PTT)
		} else {
			if {$cur_type == "menu"} {
				if {[info exists ::buttons::butsinfo(menu,$cur_vts,errorcode)]} {
					set errorcode $::buttons::butsinfo(menu,$cur_vts,errorcode)
				} else {
					set errorcode 0
				}
				if {[info exists ::menus($cur_vts,$cur_lu,$cur_pgc,numbuttons)] && $errorcode==0} {
					set cellpostable $::menus($cur_vts,$cur_lu,$cur_pgc,cellpostable)
					set idx [expr {($::tr::current(Cell) - 1) * 4}]
					set vobid  [get_int cellpostable $idx 2]
					set cellid [get_int cellpostable [expr {$idx + 3}] 1]
					if {$::buttons::butsinfo(menu,$cur_vts,totalbuttons)==0} {
						set numbuts 0
					} else {
						set numbuts $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numbuts)
					}
					set msg "$msg ($numbuts buttons)"
				} else {
					set msg "$msg (WARNING: ? buttons)"
					set vobid  0
					set cellid 0
					set ::tr::wait_step 0
				}
			}
			if {$angle != 0} {set msg "$msg, Angle $angle"}
			logprint always $msg
		}
		snapshot_state pb
		if {$cur_type == "menu"} {
			#puts "Playing menu VTS $cur_vts, LU $cur_lu, PGC $cur_pgc (numcells = $::menus($cur_vts,$cur_lu,$cur_pgc,numcells))"
			if {$::menus($cur_vts,$cur_lu,$cur_pgc,numcells) > 0} {
				# display menu buttons
				if {$ncelltime == "0:00.12" && $nstilltime == 0} {
					logprint always "        No time left to select a button! (Blanked menu)"
					set but 0
				} else {
					# determine which button group to use
					set ::buttons::groupnum 1
					if {[info exists ::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numgroups)] && \
							$::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numgroups) > 1} {
						for {set g 1} {$g <= $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numgroups)} {incr g} {
							set type $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,$g,grouptype)
							if {$type == 0} {
								set ::buttons::groupnum $g
								break
							}
						}
						if {$cur_vts == 0} {
							set vattr [get_int ::raw_sectors(0,VMGM_MAT) 0X100 2]
						} else {
							set vattr [get_int ::raw_sectors($cur_vts,VTSI_MAT) 0X100 2]
						}
						if {[expr {$vattr & 0X0C00}] != 0} {	;# 16:9 video
							switch $::tr::playerconfig(14) {
								3072 {	;# 16:9 TV
									set mode 1
								}
								512 {	;# 4:3 TV, LetterBox
									set mode 2
								}
								256 {	;# 4:3 TV, Pan&Scan
									set mode 4
								}
							}
							for {set g 1} {$g <= $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numgroups)} {incr g} {
								set type $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,$g,grouptype)
								if {[expr {$type & $mode}] == $mode} {
									set ::buttons::groupnum $g
									break
								}
							}
						}
					}
					snapshot_state menu
					set but [::menubuttons::display_menu_buttons $cur_vts $cur_lu $cur_pgc $::tr::sprm(8) trace $::tr::current(Cell)]
					if {$::tr::wait_step < 0 || $::menubuttons::menubuttons_done == -1} {
						set ::tr::now_playing false
						return
					}
				}
				#puts "button = $but"
				if {$but > 0} {
					set cellpostable $::menus($cur_vts,$cur_lu,$cur_pgc,cellpostable)
					set idx [expr {($::tr::current(Cell)-1) * 4}]
					set cur_vobid [get_int cellpostable $idx 2]
					incr idx 3
					set cur_cellid [get_int cellpostable $idx 1]
					set cmd $::buttons::butsinfo(menu,$cur_vts,$cur_vobid,$cur_cellid,$::buttons::groupnum,$but,command)
					logprint always "        Selected menu button: $but"
					logprint cmd "  -- button : [::disasm::disasm $cmd]"
					set ::tr::disasmbled [::disasm::disasm $cmd]
					if {[check_condition true false]} {
						set nexteditorline [trace_disasm_button_cmd $cmd]
						::main::set_listbox_selection $nexteditorline

						set ::tr::now_playing false
						trace_main_window_state disabled
						pause2 0.3	;# delay needed to allow the windows to refresh
						return
					} else {
						pause2 0.3	;# delay needed to allow the windows to refresh
					}
			} else {
					if {$::menubuttons::menubuttons_done==0} {
						logprint always "        End of cell playback reached. No button selected."
					}
				}
			}
			set cmds $::menus($cur_vts,$cur_lu,$cur_pgc)
		} else {	;# cur_type is PGC: Check for BOVs
			#puts "Playing Title VTS $cur_vts, PGC $cur_pgc (numcells = $::pgcs($cur_vts,$cur_pgc,numcells))"
			if {[info exists ::pgcs($cur_vts,$cur_pgc,numbuttons)] && \
					$::pgcs($cur_vts,$cur_pgc,numbuttons)>0 && \
					$::pgcs($cur_vts,$cur_pgc,numcells)>0} {
				# display BOV buttons, if any
				# assumes subpic stream 0 is used, or the enabled subpic stream,
				# or stream 0 if forced subpic.
				set substream 0
				if {$::tr::sprm(2)>=64 && $::tr::sprm(2)<=95} {
					set substream [expr {$::tr::sprm(2) & 0x1F}]
				}
				if {$::tr::sprm(2)==127} {
					set substream 0
				}
				# get stream status for the current BOV stream
				set numsubpics [::streams::get_pgc_num_subpic pgc $cur_vts,$cur_pgc]
				if {$numsubpics <= $substream} {
					logprint always "WARNING: Current subpic stream $substream out of range!  Assuming subpoic 0."
					set substream 0
				}
				set streamstatus [get_int ::pgcs($cur_vts,$cur_pgc,header) [expr {(0x1C + ($substream*4))}] 4]
				# Process BOVs only if stream status of current subpic is > 0 (subpic exists)
				# and if a valid button is selected
				if {$streamstatus > 0 && $::tr::sprm(8) >= 1024 && $::tr::sprm(8) <= 36864 \
						&& [expr {$::tr::sprm(8) % 1024}] == 0} {
					set cellpostable $::pgcs($cur_vts,$cur_pgc,cellpostable)
					set idx [expr {($::tr::current(Cell) - 1) * 4}]
					set vobid  [get_int cellpostable $idx 2]
					set cellid [get_int cellpostable [expr {$idx + 3}] 1]

					# don't process BOVs if the current PGC has been blanked out
					if {$ncelltime == "0:00.12" && $nstilltime == 0} {
						logprint always "        No time left to select a button! (Blanked BOV)"
						set but 0
					} else {
						# determine which button group to use
						set ::buttons::groupnum 1
						if {[info exists ::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numgroups)] && \
								$::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numgroups) > 1} {
							for {set g 1} {$g <= $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numgroups)} {incr g} {
								set type $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,$g,grouptype)
								if {$type == 0} {
									set ::buttons::groupnum $g
									break
								}
							}
							set vattr [get_int ::raw_sectors($cur_vts,VTSI_MAT) 0X200 2]
							if {[expr {$vattr & 0X0C00}] != 0} {	;# 16:9 video
								switch $::tr::playerconfig(14) {
									3072 {	;# 16:9 TV
										set mode 1
									}
									512 {	;# 4:3 TV, LetterBox
										set mode 2
									}
									256 {	;# 4:3 TV, Pan&Scan
										set mode 4
									}
								}
								for {set g 1} {$g <= $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numgroups)} {incr g} {
									set type $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,$g,grouptype)
									if {[expr {$type & $mode}] == $mode} {
										set ::buttons::groupnum $g
										break
									}
								}
							}
						}

						if {[info exists ::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts)] && \
								$::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts) > 0} {
							set defaultbutnum [expr {$::tr::sprm(8) / 1024}]
							if {[info exists ::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,forceselectbutn)] && \
									$::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,forceselectbutn) > 0 && \
									$::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,forceselectbutn) <= \
									$::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts)} {
								set defaultbutnum $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,forceselectbutn)
							}

							# don't process BOVs if the default button is an hidden button with a NOP command
							# and there is no still time for the current cell
							set stilltime [get_int ::pgcs($cur_vts,$cur_pgc,cellpbtable) \
									[expr {(($::tr::current(Cell)-1)*24)+2}] 1]
							set index pgc,$cur_vts,$vobid,$cellid,$::buttons::groupnum,$defaultbutnum
							if {[info exists ::buttons::butsinfo($index,coords)]} {
								set coords $::buttons::butsinfo($index,coords)
							} else {
								set coords {0 0 - 0 0}
							}
							if { (	[expr {[lindex $coords 3] - [lindex $coords 0]}] > 9 && \
									[expr {[lindex $coords 4] - [lindex $coords 1]}] > 9 ) || \
									$::buttons::butsinfo($index,command) != {00 00 00 00 00 00 00 00} \
									|| $stilltime > 1} {
								logprint always "        $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts) BOVs detected."
								snapshot_state menu
								set but [::menubuttons::display_BOVs $cur_vts $cur_pgc $::tr::sprm(8) trace $::tr::current(Cell)]
								if {$::tr::wait_step < 0 || $::menubuttons::menubuttons_done == -1} {
									set ::tr::now_playing false
									return
								}
							} else {
								set but 0
								set ::menubuttons::menubuttons_done 999
								logprint always "        $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts) BOVs detected.  BOVs are hidden.  Skipped."
							}
						} else {
							set but 0
							set ::menubuttons::menubuttons_done 999
						}
					}
					#puts "button = $but"
					if {$but > 0} {
						set cmd $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,$::buttons::groupnum,$but,command)
						logprint always "        Selected BOV: $but"
						logprint cmd "  -- BOV : [::disasm::disasm $cmd]"
						set ::tr::disasmbled [::disasm::disasm $cmd]
						if {$cmd != {00 00 00 00 00 00 00 00}} {
							if {[check_condition true false]} {
								set nexteditorline [trace_disasm_button_cmd $cmd]
								::main::set_listbox_selection $nexteditorline

								set ::tr::now_playing false
								trace_main_window_state disabled
								pause2 0.3	;# delay needed to allow the windows to refresh
								return
							} else {
								pause2 0.3	;# delay needed to allow the windows to refresh
							}
						} else {
							pause2 0.3	;# delay needed to allow the windows to refresh
						}
					} else {
						if {$::menubuttons::menubuttons_done==0} {
							logprint always "        End of cell playback reached. No BOV selected."
						}
					}
				}
				set cmds $::pgcs($cur_vts,$cur_pgc)
			}
		}

		if {$::config(use_preview) && $::config(trace_auto_preview)} {
			if {! $::config(trace_skip_small_cells_preview) || $nstilltime > 0 || \
						[string range $bcdcelltime 0 2] != "\000\000\000"} {
				# refresh GUI before launching the preview
				update idletasks
				after 100 {set ::pause false}
				tkwait variable ::pause

				if {$::tr::delay >= 1.0} {
					::preview::preview2 [lindex $::editor_cmds_from 0] \
							[lindex $::editor_cmds_from 1] $::tr::current(Cell) \
							true $::config(trace_auto_close) $::tr::delay
				} else {
					::preview::preview2 [lindex $::editor_cmds_from 0] \
							[lindex $::editor_cmds_from 1] $::tr::current(Cell) \
							true $::config(trace_auto_close) 0
				}
			}
		}

		# Process cell command, if any
		set cellcmdnum [get_int cellpbtable [expr {(($::tr::current(Cell) - 1) * 24) + 3}] 1]
		if {$cellcmdnum > 0} {
			wait_for_step
			if {$::tr::abort_current_step} {
				set ::tr::now_playing false
				return
			}
			set cellcmds [get_type_cmds cell cmds]
			set numcellcmds [llength $cellcmds]
			logprint cmd "        Cell $::tr::current(Cell): Executing cell command $cellcmdnum."
			if {$::tr::current(Cell) < $cur_numcells} {
				set seamlessflag [expr ([get_int cellpbtable [expr {$::tr::current(Cell) * 24}] 1] & 8) / 8]
			} else {
				set seamlessflag 0
			}
			if {$seamlessflag} {
				logprint cmd "  -- cell cmd $cellcmdnum : WARNING: Seamless flag set on the next cell: cell command NOT executed!"
				bell
				set ::tr::wait_step 0
			} else {
				if {$cellcmdnum <= $numcellcmds} {
					trace_main_window_state disabled
					set linenum [expr {[num_type_cmds pre cmds]+[num_type_cmds post cmds]+$cellcmdnum+2}]
					::main::set_listbox_selection $linenum
					set ::tr::now_playing false
					return
				} else {
					logprint cmd "  -- cell cmd $cellcmdnum : WARNING: cell command not defined!"
					bell
					set ::tr::wait_step 0
				}
			}
		}
		incr ::tr::current(Cell)
	} else {	;# if current Cell > number of cells in PGC
		if {$cur_numcells == 0} {
			if {$::tr::current(Cell) > 1} {
				set msg "####### WARNING: Trying to play cell $::tr::current(Cell) of a dummy PGC!"
				set ::tr::wait_step 0
				tk_messageBox -title "Trace" -icon warning -message $msg -parent .
				wait_for_step
			} else {
				set msg "******* Dummy PGC: no cell to play."
			}
			logprint always $msg
		} else {
			logprint always "******* End playback of PGC $cur_pgc."
			set ::tr::current(Cell) 1
		}
		# current cell out of range:
		# assumes the playback has ended.
		set ::tr::now_playing false
		trace_main_window_state disabled
		if {[info exists ::tr::breakpoints([lindex $::editor_cmds_from 0],[lindex $::editor_cmds_from 1],prepost)]} {
			set ::tr::wait_step 0
		}
		if {$::tr::break_at_all_menus && $cur_type=="menu"} {
			set ::tr::wait_step 0
		}
		if {$::tr::break_at_all_titles && $cur_type=="pgc"} {
			set ::tr::wait_step 0
		}
		incr idx
		::main::set_listbox_selection $idx
	}
}


#########################################
# Disassemble a button command,
# set the GPRMs and SPRMs accordingly,
# and select the destination PGC and line.
proc trace_disasm_button_cmd {hex} {
	#set hex [string range $hex 1 23]
	set nexteditorline -100
	set editorlinenum  [lindex [::main::get_selection_indexes] 0]

	set cmdtype [::disasm::get_nibble $hex 0 0]

	switch $cmdtype {
		0 {
			# flow control commands
			switch [::disasm::get_nibble $hex 1 1] {
				0 {
					# NOP
					# TODO: What is the correct behavior?
					logprint always "####### NOP command in button is normally ignored!"
					set nexteditorline [expr {$editorlinenum + 1}]
				}
				1 {
					# Goto
					logprint always "####### Goto command in button is illegal!"
					bell
				}
				2 {
					# Break
					logprint always "####### Break command in button is illegal!"
					bell
				}
				3 {
					# SetTmpPML
					logprint always "####### SetTmpPML command in button is illegal!"
					bell
				}
			}
		}
		2 {
			# link commands
			set k [::disasm::get_nibble $hex 1 1]
			if {$k == 1} {
				set ll [::disasm::get_byte $hex 7]
			} else {
				set ll 0
			}
			if {[::disasm::get_nibble $hex 1 1] != 4} {
				set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
				if {$hl_button > 0} {
					set ::tr::sprm(8) $hl_button
				}
			}
			set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline -200]
			#init_pgc_playback
		}
		3 {
			# jump/call commands
			set jumpfound false
			switch [::disasm::get_nibble $hex 1 1] {
				1 {
					# Exit
					#set nexteditorline $editorlinenum	;# TODO: where to go?
					logprint always "------> Exiting!  DVD Playback end."
					reset_log_highlight
					set ::tr::wait_step 0
					set jumpfound true
				}
				2 {
					# JumpTT
					set title [::disasm::get_byte_val $hex 5 1 99]
					set jumpfound false
					if {$title>0 && $title<100} {
#						for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
#							for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
#								if {[info exists ::pgcs($vts,$pgc,Title)] && \
#										$::pgcs($vts,$pgc,Title) == $title} {
#									set targetvts $vts
#									set targetpgc $pgc
#									set jumpfound true
#									break
#								}
#							}
#							if {$jumpfound} break
#						}
						if {[info exists ::pgcs(Title,$title)]} {
							set vts [lindex $::pgcs(Title,$title) 0]
							set pgc [lindex $::pgcs(Title,$title) 1]
							if {[info exists ::pgcs($vts,$pgc)]} {
								set targetvts $vts
								set targetpgc $pgc
								set jumpfound true
							}
						}
					}
					if {$jumpfound} {
						display_VTST_PGC $targetvts $targetpgc 1
						init_pgc_playback
						set nexteditorline 0
					}
				}
				3 {
					# JumpVTS_TT
					set ttn [::disasm::get_byte_val $hex 5 1 99]
					set vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
					if {[info exists ::ttns($vts,$ttn)]} {
						set pgc $::ttns($vts,$ttn)
						if {![display_VTST_PGC $vts $pgc 1]} {
							init_pgc_playback
							set nexteditorline 0
							set jumpfound true
						}
					}
				}
				5 {
					# JumpVTS_PTT
					set pgclist [split [lindex $::editor_cmds_from 1] ","]
					set vts [lindex $pgclist 0]
					set ttn [::disasm::get_byte_val $hex 5 1 99]
					set chapter [expr {[::disasm::get_word $hex 2 false] & 0X03FF}]
					set jumpfound false
					for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
						if {[info exists ::pgcs($vts,$p,TTN)] && $::pgcs($vts,$p,TTN) == $ttn} {
							foreach j $::pgcs($vts,$p,pttlist) {
								if {[lindex $j 1] == $chapter} {
									set jumpfound true
									set pgc $p
									break
								}
							}
						}
						if {$jumpfound} break
					}
					if {$jumpfound} {
						if {![display_VTST_PGC $vts $pgc $chapter]} {
							set cell [chapter2cell $vts $pgc $chapter]
							if {$cell != -1} {
								set cmds $::pgcs($vts,$pgc)
								set nexteditorline 0
								init_pgc_playback
								set ::tr::current(Cell) $cell
								set ::tr::current(PTT) $chapter
								set ::tr::display(PTT) $chapter
								set ::tr::now_playing false
							} else {
								set jumpfound false
							}
						}
					}
				}
				6 {
					# JumpSS *
					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# JumpSS-FP
							display_VTST_PGC 0 0 0
							set nexteditorline 0
							init_pgc_playback
							set jumpfound true
						}
						4 {
							# JumpSS-VMGM-menu
							if {[info exists ::menus(0,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
										if {![display_menu_PGC 0 $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						8 {
							# JumpSS-VTSM
							set ttn [::disasm::get_byte_val $hex 3 0 99]
							if {$ttn > 0} {
								set ::tr::sprm(5) $ttn
							}
							set vts [::disasm::get_byte_val $hex 4 0 99]
							if {$vts == 0} {	;# VTS=0 means Last visited VTS, or current VTS
								set vts $::tr::last_visited_vts
							}
							set menutype [::disasm::get_nibble $hex 5 1]
							if {[info exists ::menus($vts,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus($vts,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus($vts,1,$pgc,mask) & 0x7F}] == $menutype} {
										if {![display_menu_PGC $vts $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						12 {
							# JumpSS-VMGM-pgc
							if {![display_menu_PGC 0 [::disasm::get_word_val $hex 2 1 999]]} {
								init_pgc_playback
								set nexteditorline 0
								set jumpfound true
							}
						}
					}
				}
				8 {
					# CallSS *
					set ::tr::rsm_vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
					set ::tr::rsm_pgc [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
					set ::tr::rsm_cell [::disasm::get_byte_val $hex 4 0 255]
					if {$::tr::rsm_cell == 0} {
						set ::tr::rsm_cell [expr {$::tr::current(Cell) -1}]
					}
					switch [::disasm::get_nibble $hex 5 0] {
						0 {
							# CallSS-FP
							display_VTST_PGC 0 0 0
							set nexteditorline 0
							init_pgc_playback
							set jumpfound true
						}
						4 {
							# CallSS-VMGM-menu
							if {[info exists ::menus(0,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus(0,1,$pgc,mask) & 0x7F}] == 2} {
										if {![display_menu_PGC 0 $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						8 {
							# CallSS-VTSM
							set ttn [::disasm::get_byte_val $hex 3 0 99]
							if {$ttn > 0} {
								set ::tr::sprm(5) $ttn
							}
							set vts [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
#							set vts [::disasm::get_byte_val $hex 4 0 99]
#							if {$vts == 0} {	;# VTS=0 means Last visited VTS, or current VTS
#								set vts $::tr::last_visited_vts
#							}
							set menutype [::disasm::get_nibble $hex 5 1]
							if {[info exists ::menus($vts,1,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus($vts,1,numpgcs)} {incr pgc} {
									if {[expr {$::menus($vts,1,$pgc,mask) & 0x7F}] == $menutype} {
										if {![display_menu_PGC $vts $pgc]} {
											init_pgc_playback
											set nexteditorline 0
											set jumpfound true
										}
										break
									}
								}
							}
						}
						12 {
							# CallSS-VMGM-pgc
							if {![display_menu_PGC 0 [::disasm::get_word_val $hex 2 1 999]]} {
								init_pgc_playback
								set nexteditorline 0
								set jumpfound true
							}
						}
					}
				}
			}
			if {! $jumpfound} {
				bell
				logprint always "####### Target not found!"
				reset_log_highlight
				#set nexteditorline $editorlinenum	;# TODO: where to go?
				set ::tr::wait_step 0
			}
		}
		4 -
		5 {
			# Set SPRM commands

			switch [::disasm::get_nibble $hex 0 1] {
				1 {
					# SetSTN
					set audio [::disasm::get_byte $hex 3]
					if {[expr {$audio & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(1) [expr {$audio & 0x0F}]
						} else {
							set ::tr::sprm(1) $::tr::gprm([expr {$audio & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(Audio)  \
									"[::streams::get_pgc_audio_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(1)]"
						}
					}
					set subpic [::disasm::get_byte $hex 4]
					if {[expr {$subpic & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(2) [expr {$subpic & 0x7F}]
						} else {
							set ::tr::sprm(2) $::tr::gprm([expr {$subpic & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(SubPic) \
									"[::streams::get_pgc_subpic_descr pgc [lindex $::editor_cmds_from 1] $::tr::sprm(2)]"
						}
					}
					set angle [::disasm::get_byte $hex 5]
					if {[expr {$angle & 0X80}] > 0} {
						if {$cmdtype == 5} {
							set ::tr::sprm(3) [expr {$angle & 0x0F}]
						} else {
							set ::tr::sprm(3) $::tr::gprm([expr {$angle & 0x0F}])
						}
						if {[lindex $::editor_cmds_from 0] == "pgc"} {
							set ::tr::display(Angle)  $::tr::sprm(3)
						}
					}
				}
				2 {
					# SetNVTMR
					if {$cmdtype == 5} {
						set ::tr::sprm(9) [::disasm::get_word $hex 2 false]
					} else {
						set reg [::disasm::get_byte $hex 3]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(9) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(9) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					set ::tr::sprm(10) [::disasm::get_word_val $hex 4 1 999]
					if {$::tr::sprm(9) > 0} {
						set txt "The Navigation timer has been set to $::tr::sprm(9) seconds,"
						set txt "$txt\nand the Jump for Nav Timer to PGC number $::tr::sprm(10)."
						set txt "$txt\n\nThis jump cannot be simulated with Trace..."	;# TODO:
						tk_messageBox -title "Trace" -icon warning -message $txt -parent .
					}
				}
				3 {
					# SetGPRMMD
					set gprm [::disasm::get_nibble $hex 5 1]
					if {$cmdtype == 5} {
						set ::tr::gprm($gprm) [::disasm::get_word $hex 2 false]
					} else {
						set reg [::disasm::get_byte $hex 3]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::gprm($gprm) $::tr::gprm($reg)
						} else {
							set ::tr::gprm($gprm) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					set mode [expr {[::disasm::get_nibble $hex 5 0] & 8}]
					if {$mode == 8} {
						set txt "The GPRM($gprm) has been set to $::tr::gprm($gprm) in counter mode."
						set txt "$txt\n\nThe counter mode cannot be simulated with Trace..."	;# TODO:
						tk_messageBox -title "Trace" -icon warning -message $txt -parent .
					}
				}
				4 {
					# SetAMXMD
					set txt "${iftxt}(SetAMXMD) Set Karaoke Audio Mixing Mode =(mov)"
					if {$cmdtype == 5} {
						set ::tr::sprm(11) [::disasm::get_word_val $hex 4 0 7196]
					} else {
						set reg [::disasm::get_byte $hex 5]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(11) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(11) $::tr::sprm([expr {$reg-0x80}])
						}
					}
				}
				6 {
					# SetHL_BTN
					if {$cmdtype == 5} {
						set but [expr {([::disasm::get_byte $hex 4] / 4) * 1024}]
						set ::tr::sprm(8) $but
					} else {
						set reg [::disasm::get_byte $hex 5]
						if {[expr {$reg & 0x80}] == 0} {
							set ::tr::sprm(8) $::tr::gprm($reg)
						} else {
							set ::tr::sprm(8) $::tr::sprm([expr {$reg-0x80}])
						}
					}
					# TODO: no button here because it is already set!
				}
			}

			# Set + Link
			# (ces commandes ne peuvent pas avoir en meme temps un IF et un Link)
			set k [::disasm::get_nibble $hex 1 1]
			if {$k > 0} {
				if {$k == 1} {
					set ll [::disasm::get_byte $hex 7]
				} else {
					set ll 0
				}
				if {$k != 4} {
					set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
					if {$hl_button > 0} {
						set ::tr::sprm(8) $hl_button
					}
				}
				set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline -200]
				#init_pgc_playback
			}
		}
		6 -
		7 {
			# Set command
			if {[expr {$cmdtype & 1}]} {
				set val [::disasm::get_word $hex 4 false]
			} else {
				set reg [::disasm::get_byte $hex 5]
				if {[expr {$reg & 0x80}] == 0} {
					set val $::tr::gprm($reg)
				} else {
					set val $::tr::sprm([expr {$reg & 0x1F}])
				}
			}
			set gprm1 [::disasm::get_byte $hex 3]
			set op1 $::tr::gprm($gprm1)
			set operand [::disasm::get_nibble $hex 0 1]
			if {$operand == 2} {
				# SWAP operation
				set ::tr::gprm($gprm1) $val
				if {[expr {$reg & 0x80}] == 0} {
					set ::tr::gprm($reg) $op1
				} else {
					set ::tr::sprm([expr {$reg & 0x1F}]) $op1
				}
			} else {
				set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
			}

			# Link part
			# Set ne peut pas avoir en meme temps un IF et un Link
			set k [::disasm::get_nibble $hex 1 1]
			if {$k > 0} {
				if {$k == 1} {
					set ll [::disasm::get_byte $hex 7]
				} else {
					set ll 0
				}
				if {$k != 4} {
					set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
					if {$hl_button > 0} {
						set ::tr::sprm(8) $hl_button
					}
				}
				set nexteditorline [trace_disasm_link $hex $k $ll $nexteditorline -200]
				#init_pgc_playback
			}
		}
		8 -
		9 {
			# SetCLnk
			# Set part
			if {[expr {$cmdtype & 1}]} {
				set val [::disasm::get_word $hex 2 false]
			} else {
				set reg [::disasm::get_byte $hex 3]
				if {[expr {$reg & 0x80}] == 0} {
					set val $::tr::gprm($reg)
				} else {
					set val $::tr::sprm([expr {$reg & 0x1F}])
				}
			}
			set gprm1 [::disasm::get_nibble $hex 1 1]
			set op1 $::tr::gprm($gprm1)
			set operand [::disasm::get_nibble $hex 0 1]
			if {$operand == 2} {
				# SWAP operation
				set ::tr::gprm($gprm1) $val
				if {[expr {$reg & 0x80}] == 0} {
					set ::tr::gprm($reg) $op1
				} else {
					set ::tr::sprm([expr {$reg & 0x1F}]) $op1
				}
			} else {
				set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
			}

			# if then link part
			if {[check_condition true false]} {
				set ll [::disasm::get_byte $hex 7]
				set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
				if {$hl_button > 0} {
					set ::tr::sprm(8) $hl_button
				}
				set nexteditorline [trace_disasm_link $hex 1 $ll $nexteditorline -200]
				#init_pgc_playback
			}
		}
		10 -
		11 -
		12 -
		13 {
			# CSetCLnk and CSetLnk
			# CSetCLnk don't even enter here if the condition is not true

			# set part: CSetLnk performs the set conditionally
			if {$cmdtype < 12 || [check_condition false false]} {
				if {[expr {$cmdtype & 1}]} {
					set val [::disasm::get_word $hex 2 false]
				} else {
					set reg [::disasm::get_byte $hex 2]
					if {[expr {$reg & 0x80}] == 0} {
						set val $::tr::gprm($reg)
					} else {
						set val $::tr::sprm([expr {$reg & 0x1F}])
					}
				}
				set gprm1 [::disasm::get_nibble $hex 1 1]
				set op1 $::tr::gprm($gprm1)
				set operand [::disasm::get_nibble $hex 0 1]
				if {$operand == 2} {
					# SWAP operation
					set ::tr::gprm($gprm1) $val
					if {[expr {$reg & 0x80}] == 0} {
						set ::tr::gprm($reg) $op1
					} else {
						set ::tr::sprm([expr {$reg & 0x1F}]) $op1
					}
				} else {
					set ::tr::gprm($gprm1) [operation_result $op1 $operand $val]
				}
			}

			# Link part: CSetLnk performs the link unconditionally
			set ll [::disasm::get_byte $hex 7]
			set hl_button [expr {([::disasm::get_byte $hex 6] & 0XFC) * 0x100}]
			if {$hl_button > 0} {
				set ::tr::sprm(8) $hl_button
			}
			set nexteditorline [trace_disasm_link $hex 1 $ll $nexteditorline -200]
			#init_pgc_playback
		}
	}
	# select the line to be executed next
	if {$nexteditorline != -1} {
		::main::set_listbox_selection $nexteditorline
	}
	return $nexteditorline
}

###################### menu procedures #####################
proc log_gprms_now {} {
	set freegprms [::macros::find_unused_gprms true]
	reset_log_highlight
	logprint always "+++++++ Log: current GPRMs dump"
	for {set i 0} {$i < 16} {incr i} {
		if {[lsearch $freegprms $i] == -1} {
			set val $::tr::gprm($i)
			set txt "\t-> gprm($i) = $val (0x[format {%04X} $val])"
			if {[::streams::valid_langcode [d2ascii $val]]} {
				set txt "${txt} = \"[d2ascii $val]\""
			}
			set but [expr {$val/1024}]
			if {$but > 0 && $but <= 36 && [expr {$val%1024}] == 0} {
				set txt "${txt} = button [expr {$val/1024}]"
			}
			logprint always $txt
		} else {
			logprint always "\t-> gprm($i) : not used"
		}
	}
	reset_log_highlight
}
proc log_sprms_now {} {
	reset_log_highlight
	logprint always "+++++++ Log: current SPRMs dump"
	for {set i 0} {$i < 24} {incr i} {
		set val $::tr::sprm($i)
		set txt "\t-> sprm($i:[lindex $::disasm::SPRMlist $i]) = $val (0x[format {%04X} $val])"
		if {[::streams::valid_langcode [d2ascii $val]]} {
			set txt "${txt} = \"[d2ascii $val]\""
		}
		set but [expr {$val/1024}]
		if {$but > 0 && $but <= 36 && [expr {$val%1024}] == 0} {
			set txt "${txt} = button [expr {$val/1024}]"
		}
		logprint always $txt
	}
	reset_log_highlight
}
proc snapshot_gprms {set} {
	set ::tr::snapshot_bank($set) 1
	set numchanged 0
	for {set i 0} {$i < 16} {incr i} {
		set ::tr::backup_gprm($set,$i) $::tr::gprm($i)
		set ::tr::backup_gprm_changed($set,$i) $::tr::gprm_changed($i)
		if {$::tr::gprm_changed($i)} {
			incr numchanged
		}
	}
	if {$set == "gprm"} {
		set ::tr::bank_pgc_label(gprm) "$numchanged GPRMs changed."
		.mf.cf.trace.fl.mnu.state.m_state.rem entryconfigure "Only GPRMs*" -label "Only GPRMs:  $::tr::bank_pgc_label(gprm)"
		.mf.cf.trace.fl.mnu.state.m_state.res entryconfigure "Only GPRMs*" -label "Only GPRMs:  $::tr::bank_pgc_label(gprm)"
	}
}
proc restore_gprms {set} {
	if {[info exists ::tr::backup_gprm($set,0)]} {
		set ::tr::snapshot_bank($set) 1
	} else {
		set ::tr::snapshot_bank($set) 0
		bell
		return
	}
	for {set i 0} {$i < 16} {incr i} {
		if {$::tr::gprm_changed($i) || $::tr::gprm($i) != $::tr::backup_gprm($set,$i)} {
			set ::tr::gprm($i) $::tr::backup_gprm($set,$i)
		}
		set ::tr::gprm_changed($i) $::tr::backup_gprm_changed($set,$i)
		if {$::tr::view_gprms} {
			if {$::tr::gprm_changed($i)} {
				.trinfo.f1.top.top.f.e$i configure -foreground black
			} else {
				.trinfo.f1.top.top.f.e$i configure -foreground $::config(disabledforeground)
			}
		}
	}
}
proc snapshot_sprms {set} {
	set ::tr::snapshot_bank($set) 1
	set numchanged 0
	for {set i 0} {$i < 24} {incr i} {
		set ::tr::backup_sprm($set,$i) $::tr::sprm($i)
		set ::tr::backup_sprm_changed($set,$i) $::tr::sprm_changed($i)
		if {$::tr::sprm_changed($i)} {
			incr numchanged
		}
	}
	if {$set == "sprm"} {
		set ::tr::bank_pgc_label(sprm) "$numchanged SPRMs changed."
		.mf.cf.trace.fl.mnu.state.m_state.rem entryconfigure "Only SPRMs*" -label "Only SPRMs:  $::tr::bank_pgc_label(sprm)"
		.mf.cf.trace.fl.mnu.state.m_state.res entryconfigure "Only SPRMs*" -label "Only SPRMs:  $::tr::bank_pgc_label(sprm)"
	}
}
proc restore_sprms {set} {
	if {[info exists ::tr::backup_sprm($set,0)]} {
		set ::tr::snapshot_bank($set) 1
	} else {
		set ::tr::snapshot_bank($set) 0
		bell
		return
	}
	for {set i 0} {$i < 24} {incr i} {
		if {$::tr::sprm_changed($i) || $::tr::sprm($i) != $::tr::backup_sprm($set,$i)} {
			set ::tr::sprm($i) $::tr::backup_sprm($set,$i)
		}
		set ::tr::sprm_changed($i) $::tr::backup_sprm_changed($set,$i)
		if {$::tr::view_sprms} {
			if {$::tr::sprm_changed($i)} {
				.trinfo.f2.f2.f.e$i configure -foreground black
			} else {
				.trinfo.f2.f2.f.e$i configure -foreground $::config(disabledforeground)
			}
		}
	}
}
proc snapshot_pos {set} {
	set ::tr::snapshot_bank($set) 1
	set ::tr::backup_pos($set,pgc) $::editor_cmds_from
	set absline [lindex [::main::get_selection_indexes] 0]
	if {$absline == "-"} {
		bell
		return
	}
	set curcmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set ::tr::backup_pos($set,line) [list [get_cmd_type curcmds $absline] [get_cmd_linenum curcmds $absline]]
	set ::tr::backup_pos($set,cmd) [lindex $curcmds $absline]

	set ::tr::backup_current($set,Title)	$::tr::current(Title)
	set ::tr::backup_current($set,TTN)		$::tr::current(TTN)
	set ::tr::backup_current($set,LU)		$::tr::current(LU)
	set ::tr::backup_current($set,PGC)		$::tr::current(PGC)
	set ::tr::backup_current($set,PTT)		$::tr::current(PTT)
	set ::tr::backup_current($set,PG)		$::tr::current(PG)
	set ::tr::backup_current($set,Cell)		$::tr::current(Cell)
	set ::tr::backup_last_visited_vts($set)	$::tr::last_visited_vts
	set ::tr::backup_now_playing($set)		$::tr::now_playing
	set ::tr::backup_display($set,Title)	$::tr::display(Title)
	set ::tr::backup_display($set,PTT)		$::tr::display(PTT)
	set ::tr::backup_display($set,Audio)	$::tr::display(Audio)
	set ::tr::backup_display($set,SubPic)	$::tr::display(SubPic)
	set ::tr::backup_display($set,Angle)	$::tr::display(Angle)
	if {[info exists ::tr::rsm_vts]} {
		set ::tr::backup_rsm_vts($set)			$::tr::rsm_vts
		set ::tr::backup_rsm_pgc($set)			$::tr::rsm_pgc
		set ::tr::backup_rsm_cell($set)			$::tr::rsm_cell
	}

	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	if {[lindex $::editor_cmds_from 0] == "menu"} {
		set pgclabel [::main::pgc_label {} menu $vts [lindex [split [lindex $::editor_cmds_from 1] ,] 1] $::tr::current(PGC)]
	} else {
		set pgclabel [::main::pgc_label {} pgc $vts $::tr::current(PGC)]
	}
	if {[string range [lindex $::tr::backup_pos($set,line) 0] 0 0] == "*"} {
		set pgclabel "$pgclabel  (**[lindex $::tr::backup_pos($set,line) 0])"
	} else {
		set pgclabel "$pgclabel  ([lindex $::tr::backup_pos($set,line) 0] [lindex $::tr::backup_pos($set,line) 1])"
	}
	set ::tr::bank_pgc_label($set) $pgclabel
	set w .mf.cf.trace
	if {[string is integer $set]} {
		$w.fl.mnu.state.m_state.rem entryconfigure "User bank $set*" -label "User bank $set:  $pgclabel"
		$w.fl.mnu.state.m_state.res entryconfigure "User bank $set*" -label "User bank $set:  $pgclabel"
	} else {
		switch $set {
			pgc			{$w.fl.mnu.state.m_state.auto entryconfigure "Current PGC*"			-label "Current PGC:  $pgclabel"}
			domain		{$w.fl.mnu.state.m_state.auto entryconfigure "Current Domain*"		-label "Current Domain:  $pgclabel"}
			pb			{$w.fl.mnu.state.m_state.auto entryconfigure "Last Playback*"		-label "Last Playback:  $pgclabel"}
			menu		{$w.fl.mnu.state.m_state.auto entryconfigure "Last Menu*"			-label "Last Menu:  $pgclabel"}
			auto		{$w.fl.mnu.state.m_state.auto entryconfigure "Trace Mode closed*"	-label "Trace Mode closed:  $pgclabel"}
			pos			{
				$w.fl.mnu.state.m_state.rem entryconfigure "Only Position*"		-label "Only Position:  $pgclabel"
				$w.fl.mnu.state.m_state.res entryconfigure "Only Position*"		-label "Only Position:  $pgclabel"
			}
			default		{}
		}
	}
}
proc restore_pos {set} {
	if {[info exists ::tr::backup_pos($set,pgc)]} {
		set ::tr::snapshot_bank($set) 1
	} else {
		set ::tr::snapshot_bank($set) 0
		bell
		return
	}

	reset_log_highlight

	set pgctype [lindex $::tr::backup_pos($set,pgc) 0]
	set pgcidx  [lindex $::tr::backup_pos($set,pgc) 1]
	set pgcidxs [split $pgcidx ,]
	set cmdtype [lindex $::tr::backup_pos($set,line) 0]
	set cmdline [lindex $::tr::backup_pos($set,line) 1]

	set curcmds [subst \$::${pgctype}s($pgcidx)]
	set nprecmds  [num_type_cmds pre  curcmds]
	set npostcmds [num_type_cmds post curcmds]
	set ncellcmds [num_type_cmds cell curcmds]

	switch $cmdtype {
		"pre" {
			set cmdtxt "pre command $cmdline"
			set absline $cmdline
			if {$cmdline > $nprecmds} {
				set cmdtxt "INVALID pre command $cmdline!"
				set absline $nprecmds
				bell
			}
		}
		"post" {
			set cmdtxt "post command $cmdline"
			set absline [expr $nprecmds + 1 + $cmdline]
			if {$cmdline > $npostcmds} {
				set cmdtxt "INVALID post command $cmdline!"
				set absline [expr $nprecmds + 1 + $npostcmds]
				bell
			}
		}
		"cell" {
			set cmdtxt "cell command $cmdline"
			set absline [expr $nprecmds + $npostcmds + 2 + $cmdline]
			if {$cmdline > $nprecmds} {
				set absline $nprecmds
				bell
			}
			if {$cmdline > $ncellcmds} {
				set cmdtxt "INVALID cell command $cmdline!"
				set absline [expr $nprecmds + $npostcmds + 2 + $ncellcmds]
				bell
			}
		}
		"* pre" {
			set cmdtxt "entry point."
			set absline 0
		}
		"* post" {
			set cmdtxt "playback point."
			set absline [expr $nprecmds + 1]
		}
		"* cell" {
			set cmdtxt "exit point."
			set absline [expr $nprecmds + $npostcmds + 2]
		}
	}
	if {$pgctype == "menu"} {
		set rc [catch {::main::display_pgc $pgctype [lindex $pgcidxs 0] [lindex $pgcidxs 1] [lindex $pgcidxs 2]}]
		set label [::main::pgc_label {} $pgctype [lindex $pgcidxs 0] [lindex $pgcidxs 1] [lindex $pgcidxs 2]]
	} else {
		set rc [catch {::main::display_pgc $pgctype [lindex $pgcidxs 0] [lindex $pgcidxs 1]}]
		set label [::main::pgc_label {} $pgctype [lindex $pgcidxs 0] [lindex $pgcidxs 1]]
	}
	if {$absline == "-"} {
		set rc 1
	}
	if {$rc} {
		logprint always "+++++++ WARNING: PGC $label not found!  Cannot restore position."
		bell
	} else {
		logprint always "+++++++ Returning to $label, $cmdtxt"
		::main::init_tree_menu_no_move ;# clear PGC played highlights
		set rc [catch {::main::set_listbox_selection $absline}]
		if {! $rc} {
			set curcmd [lindex $curcmds $absline]
			if {$curcmd != $::tr::backup_pos($set,cmd)} {
				set rc 1
			}
		}
		if {$rc} {
			bell
			logprint always "+++++++ WARNING: Current command has changed!"
		}

		set ::tr::current(Title)	$::tr::backup_current($set,Title)
		set ::tr::current(TTN)		$::tr::backup_current($set,TTN)
		set ::tr::current(LU)		$::tr::backup_current($set,LU)
		set ::tr::current(PGC)		$::tr::backup_current($set,PGC)
		set ::tr::current(PTT)		$::tr::backup_current($set,PTT)
		set ::tr::current(PG)		$::tr::backup_current($set,PG)
		set ::tr::current(Cell)		$::tr::backup_current($set,Cell)
		set ::tr::last_visited_vts	$::tr::backup_last_visited_vts($set)
		set ::tr::now_playing		$::tr::backup_now_playing($set)
		set ::tr::display(Title)	$::tr::backup_display($set,Title)
		set ::tr::display(PTT)		$::tr::backup_display($set,PTT)
		set ::tr::display(Audio)	$::tr::backup_display($set,Audio)
		set ::tr::display(SubPic)	$::tr::backup_display($set,SubPic)
		set ::tr::display(Angle)	$::tr::backup_display($set,Angle)
		if {[info exists ::tr::backup_rsm_vts($set)]} {
			set ::tr::rsm_vts			$::tr::backup_rsm_vts($set)
			set ::tr::rsm_pgc			$::tr::backup_rsm_pgc($set)
			set ::tr::rsm_cell			$::tr::backup_rsm_cell($set)
		}
		# set PUOs for current PGC
		set header [subst \$::${pgctype}s($pgcidx,header)]
		trace_main_window_state_puos [get_int header 8 4]

		if {! $::tr::now_playing} {
			trace_main_window_state disabled
		}
	}
}
proc snapshot_state {set} {
	set ::tr::snapshot_bank($set) 1
	snapshot_pos   $set
	snapshot_gprms $set
	snapshot_sprms $set
}
proc restore_state {set} {
	if {[info exists ::tr::backup_pos($set,pgc)]} {
		#blocking_msg "Restoring Trace state..."
		set ::tr::snapshot_bank($set) 1
		restore_pos   $set
		restore_gprms $set
		restore_sprms $set
		reset_log_highlight
		#close_blocking_msg
	} else {
		set ::tr::snapshot_bank($set) 0
		bell
	}
}
proc clear_all_state_banks {} {
	foreach set [array names ::tr::snapshot_bank] {
		set ::tr::snapshot_bank($set) 0
		unset -nocomplain ::tr::backup_pos($set,pgc)	;# reset trace positions
		unset -nocomplain ::tr::bank_pgc_label($set)	;# restore original labels
	}
	::tr::update_state_menu_labels
}

proc snapshot_copy {src dest} {
	if {! [info exists ::tr::snapshot_bank($src)] || !$::tr::snapshot_bank($src)} {
		return
	}
	set ::tr::snapshot_bank($dest) 1
	for {set i 0} {$i < 16} {incr i} {
		set ::tr::backup_gprm($dest,$i) $::tr::backup_gprm($src,$i)
		set ::tr::backup_gprm_changed($dest,$i) $::tr::backup_gprm_changed($src,$i)
	}
	for {set i 0} {$i < 24} {incr i} {
		set ::tr::backup_sprm($dest,$i) $::tr::backup_sprm($src,$i)
		set ::tr::backup_sprm_changed($dest,$i) $::tr::backup_sprm_changed($src,$i)
	}
	set ::tr::backup_pos($dest,pgc)  $::tr::backup_pos($src,pgc)
	set ::tr::backup_pos($dest,line) $::tr::backup_pos($src,line)
	set ::tr::backup_pos($dest,cmd)  $::tr::backup_pos($src,cmd)

	set ::tr::backup_current($dest,Title)	$::tr::backup_current($src,Title)
	set ::tr::backup_current($dest,TTN)		$::tr::backup_current($src,TTN)
	set ::tr::backup_current($dest,LU)		$::tr::backup_current($src,LU)
	set ::tr::backup_current($dest,PGC)		$::tr::backup_current($src,PGC)
	set ::tr::backup_current($dest,PTT)		$::tr::backup_current($src,PTT)
	set ::tr::backup_current($dest,PG)		$::tr::backup_current($src,PG)
	set ::tr::backup_current($dest,Cell)	$::tr::backup_current($src,Cell)
	set ::tr::backup_last_visited_vts($dest) $::tr::backup_last_visited_vts($src)
	set ::tr::backup_now_playing($dest)		$::tr::backup_now_playing($src)
	set ::tr::backup_display($dest,Title)	$::tr::backup_display($src,Title)
	set ::tr::backup_display($dest,PTT)		$::tr::backup_display($src,PTT)
	set ::tr::backup_display($dest,Audio)	$::tr::backup_display($src,Audio)
	set ::tr::backup_display($dest,SubPic)	$::tr::backup_display($src,SubPic)
	set ::tr::backup_display($dest,Angle)	$::tr::backup_display($src,Angle)
	if {[info exists ::tr::backup_rsm_vts($src)]} {
		set ::tr::backup_rsm_vts($dest)		$::tr::backup_rsm_vts($src)
		set ::tr::backup_rsm_pgc($dest)		$::tr::backup_rsm_pgc($src)
		set ::tr::backup_rsm_cell($dest)	$::tr::backup_rsm_cell($src)
	}
	# update label in trace menus
	if {[info exists ::tr::bank_pgc_label($src)]} {
		set ::tr::bank_pgc_label($dest) $::tr::bank_pgc_label($src)
		set w .mf.cf.trace
		switch $dest {
			last_pgc	{
				$w.fl.mnu.state.m_state.auto entryconfigure "Last PGC*" \
						-label "Last PGC:  $::tr::bank_pgc_label($dest)"
			}
			last_domain	{
				$w.fl.mnu.state.m_state.auto entryconfigure "Last Domain*" \
						-label "Last Domain:  $::tr::bank_pgc_label($dest)"
			}
		}
	}
}

proc clear_changed_flags {} {
	for {set g 0} {$g < 16} {incr g} {
		if {$::tr::view_gprms} {
			.trinfo.f1.top.top.f.e$g configure -foreground $::config(disabledforeground)
		}
		set ::tr::gprm_changed($g) 0
	}
	for {set s 0} {$s < 24} {incr s} {
		if {$::tr::view_sprms} {
			.trinfo.f2.f2.f.e$s configure -foreground $::config(disabledforeground)
		}
		set ::tr::sprm_changed($s) 0
	}
}
proc copy_regs_as_cmds {mode destination} {

	set cmds {}
	set cmd {71 00 00 00 00 00 00 00}

	set freegprms [::macros::find_unused_gprms true]
	for {set i 0} {$i < 16} {incr i} {
		set useit true
		if {$mode == "used" && [lsearch $freegprms $i] >= 0} {set useit false}
		if {$mode == "changed" && $::tr::gprm_changed($i)==0} {set useit false}
		if {$useit} {
			lset cmd 3 [format %02X $i]
			lset cmd 4 [format %02X [expr {$::tr::gprm($i) / 0x100}]]
			lset cmd 5 [format %02X [expr {$::tr::gprm($i) & 0xFF}]]
			lappend cmds $cmd
		}
	}

	# SetSTN: SPRMs 1, 2, 3
	set cmd {51 00 00 00 00 00 00 00}
	if {$mode == "changed"} {
		set useit false
		if {$::tr::sprm_changed(1)>0} {
			lset cmd 3 [format %02X [expr {$::tr::sprm(1) | 0x80}]]
			set useit true
		}
		if {$::tr::sprm_changed(2)>0} {
			lset cmd 4 [format %02X [expr {$::tr::sprm(2) | 0x80}]]
			set useit true
		}
		if {$::tr::sprm_changed(3)>0} {
			lset cmd 5 [format %02X [expr {$::tr::sprm(3) | 0x80}]]
			set useit true
		}
		if {$useit} {
			lappend cmds $cmd
		}
	} else {
		lset cmd 3 [format %02X [expr {$::tr::sprm(1) | 0x80}]]
		lset cmd 4 [format %02X [expr {$::tr::sprm(2) | 0x80}]]
		lset cmd 5 [format %02X [expr {$::tr::sprm(3) | 0x80}]]
		lappend cmds $cmd
	}

	# SetHL_BTN: SPRM 8
	set useit true
	set val $::tr::sprm(8)
	if {$val < 1024} {set useit false}
	if {$mode == "changed" && $::tr::sprm_changed(8)==0} {set useit false}
	if {$useit} {
		set cmd {56 00 00 00 04 00 00 00}
		lset cmd 4 [format %02X [expr {$val / 256}]]
		lappend cmds $cmd
	}

	if {$destination == "clip"} {
		clipboard clear
		clipboard append -- $cmds
	}
	if {$destination == "scratch"} {
		foreach c $cmds {
			lappend ::scratchbook_buffer $c
		}
	}
	if {$destination != "none" && [llength $cmds] == 0} {
		show_msg "No $mode registers found."
		bell
	} else {
		show_msg ""
	}
	return $cmds
}
proc toggle_breakpoint {type} {
	if {![info exists ::dvddir]} {bell; return}
	set cur_type [lindex $::editor_cmds_from 0]
	set index    [lindex $::editor_cmds_from 1]
	if {[info exists ::tr::breakpoints($cur_type,$index,$type)]} {
		unset ::tr::breakpoints($cur_type,$index,$type)
	} else {
		set ::tr::breakpoints($cur_type,$index,$type) 1
	}
	::main::highlight_trace_breakpoints
}
proc clear_all_breakpoints {} {
	if {[info exists ::tr::breakpoints]} {
		unset ::tr::breakpoints
	}
	::main::highlight_trace_breakpoints
}

proc display_VTST_PGC {vts pgc ptt} {
	if {$::pgcs(numvts) < $vts} {return 1}
	if {$::pgcs($vts,numpgcs) < $pgc} {return 1}
	if {$vts > 0} {
		if {[info exists ::pgcs($vts,$pgc,Title)] && $::pgcs($vts,$pgc,Title) != ""} {
			set ::tr::sprm(4) $::pgcs($vts,$pgc,Title)
			set ::tr::current(Title) $::pgcs($vts,$pgc,Title)
			set ::tr::display(Title) $::tr::current(Title)
		} else {
			if {[info exists ::pgcs($vts,$pgc,TTN)]} {
				for {set p 1} {$p <= $::pgcs($vts,numpgcs)} {incr p} {
					if {$::pgcs($vts,$p,TTN) == $::pgcs($vts,$pgc,TTN) && \
							[info exists ::pgcs($vts,$p,Title)] && \
							$::pgcs($vts,$p,Title) != ""} {
						set ::tr::sprm(4) $::pgcs($vts,$p,Title)
						set ::tr::current(Title) $::pgcs($vts,$p,Title)
						set ::tr::display(Title) $::tr::current(Title)
						break
					}
				}
			}
		}
		if {[info exists ::pgcs($vts,$pgc,TTN)]} {
			set ::tr::sprm(5) $::pgcs($vts,$pgc,TTN)
		}
		set ::tr::sprm(6) $pgc
		if {$ptt > 0} {
			set ::tr::sprm(7) $ptt
			if {$::tr::playerconfig(sprm7holdsPG) && $ptt > 0} {
				set numcells $::pgcs($vts,$pgc,numcells)
				set pttlist  $::pgcs($vts,$pgc,pttlist)
				set numptts  [llength $pttlist]
				foreach pair $pttlist {
					if {[lindex $pair 1] == $ptt} {
						set ::tr::sprm(7) [lindex $pair 0]
					}
				}
			}
		}
		set ::tr::sprm(14) [computeCurrentVideoMode pgc $vts]
	}
	::main::display_pgc pgc $vts $pgc
	set ::tr::pgclabel [$::widget(main_cmd_lab) cget -text]
	return 0
}

proc display_menu_PGC {vts pgc} {
	if {$::menus(numvts) < $vts} {return 1}
	# find most appropriate LU
	set lu 1
	set langcode [d2c $::tr::sprm(0) 2]
	if {$::menus($vts,numlus) > 0} {
		for {set l 1} {$l <= $::menus($vts,numlus)} {incr l} {
			if {$::menus($vts,$l,lang) == $langcode} {
				set lu $l
				break
			}
		}
	} else {
		return 1
	}
	if {$::menus($vts,$lu,numpgcs) < $pgc} {
		return 1
	}
	set ::tr::current(LU) $lu

	::main::display_pgc menu $vts $lu $pgc
	set ::tr::pgclabel [$::widget(main_cmd_lab) cget -text]
	return 0
}
proc computeCurrentVideoMode {type vts} {
	if {$type == "pgc"} {
		set vattr [get_int ::raw_sectors($vts,VTSI_MAT) 0X200 2]
	} else {
		if {$vts == 0} {
			set vattr [get_int ::raw_sectors(0,VMGM_MAT) 0X100 2]
		} else {
			set vattr [get_int ::raw_sectors($vts,VTSI_MAT) 0X100 2]
		}
	}
	if {[expr {$::tr::playerconfig(14) & 0x0C00}] == 0x0C00} {
		# 16:9 TV: always return 0x0C00: 16:9
		return [expr {(0x0C00)}]
	} else {	;# 4:3 TV
		if {[expr {$vattr & 0X0C00}] == 0} {
			# 4:3 Video: return 'normal'
			return 0
		} else {
			# 16:9 Video: check automatic video mode
			switch [expr {($vattr & 0X0300) / 0X100}] {
				1 {	;# Automatic Pan&Scan
					return 1
				}
				2 {	;# Automatic Letterbox
					return 2
				}
				default {	;# none or both: return preferred mode
					return [expr {$::tr::playerconfig(14) & 0x0300}]
				}
			}
		}
	}
}

}


::tr::init_trace_config
